@@ -23,7 +23,6 @@ from __future__ import division, absolute_import, print_function, unicode_litera
 #from awlsim.common.cython_support cimport * #@cy
 from awlsim.common.compat import *
 
-import sys
 import re
 
 from awlsim.common.enumeration import *
@@ -31,11 +30,11 @@ from awlsim.common.project import *
 from awlsim.common.refmanager import *
 from awlsim.common.namevalidation import *
 from awlsim.common.sources import AwlSource
+from awlsim.common.exceptions import *
+from awlsim.common.util import *
 
-from awlsim.core.util import *
 from awlsim.core.datatypes import *
-from awlsim.core.memory import * #@nocy
-#from awlsim.core.memory cimport * #@cy
+from awlsim.core.memory import * #+cimport
 from awlsim.core.identifier import *
 
 
@@ -59,7 +58,7 @@ class RawAwlInsn(object):
 		self.name = None
 		self.ops = []
 
-	def __repr__(self):
+	def __repr__(self): #@nocov
 		ret = []
 		if self.hasLabel():
 			ret.append(self.getLabel() + ':  ')
@@ -224,7 +223,7 @@ class RawAwlDataInit(object):
 	def getIdentString(self):
 		return ".".join(str(ident) for ident in self.identChain)
 
-	def __repr__(self):
+	def __repr__(self): #@nocov
 		return self.getIdentString() + " := " + str(self.valueTokens)
 
 class RawAwlDataField(object):
@@ -275,7 +274,8 @@ class RawAwlDataField(object):
 					return field.getChild(identChain[1:])
 				return field
 		return None
-	def __repr__(self):
+
+	def __repr__(self): #@nocov
 		return self.getIdentString()
 
 class RawAwlDB(RawAwlBlock):
@@ -548,8 +548,8 @@ class AwlParser(object):
 				   (c == '(' and not t.haveLabelToken()):
 					# Parenthesis begin
 					t.inParens = True
-					t.addCharacter(c)
 					t.finishCurToken()
+					t.addToken(c)
 					cont(); continue
 				if t.inParens and c == ')':
 					# Parenthesis end
@@ -557,7 +557,7 @@ class AwlParser(object):
 					t.finishCurToken()
 					t.addToken(c)
 					cont(); continue
-				if ((self.__inAnyHeaderOrGlobal() or self.__inVariableSection()) and\
+				if ((self.__inAnyHeaderOrGlobal() or self.__inVariableSection() or t.inParens) and\
 				    c in {'=', ':', '{', '}', '.'}) or\
 				   c in {',', '[', ']'} or\
 				   (c == '=' and len(t.tokens) == 1 and not t.curToken):
