@@ -25,10 +25,11 @@ from awlsim.common.compat import *
 
 from awlsim.common.cpuconfig import *
 from awlsim.common.datatypehelpers import * #+cimport
+from awlsim.common.exceptions import *
+from awlsim.common.util import *
 
 from awlsim.core.instructions.all_insns import * #+cimport
 from awlsim.core.datastructure import * #+cimport
-from awlsim.core.util import *
 from awlsim.core.offset import * #+cimport
 from awlsim.core.operatortypes import * #+cimport
 from awlsim.core.operators import * #+cimport
@@ -203,6 +204,8 @@ class AwlTranslator(object):
 	# dataType: The AwlDataType for this field.
 	# initBytes: The initialization bytes for this field, or None.
 	def rawFieldTranslate(self, rawField):
+#@cy		cdef AwlMemoryObject memObj
+
 		# Get the field name
 		name = rawField.getIdentString()
 		# Get the field data type
@@ -230,7 +233,6 @@ class AwlTranslator(object):
 			# Translate the initialization values and
 			# put them into a bytearray.
 			initMem = AwlMemory(intDivRoundUp(dataType.width, 8))
-			initBytes = initMem.dataBytes
 			AwlDataType = _getAwlDataTypeClass()
 			if dataType.type == AwlDataType.TYPE_ARRAY:
 				for rawDataInit in rawField.defaultInits:
@@ -240,29 +242,44 @@ class AwlTranslator(object):
 					offset = make_AwlOffset_fromLongBitOffset(
 							linArrayIndex *
 							dataType.arrayElementType.width)
+					memObj = make_AwlMemoryObject_fromGeneric(
+							value,
+							dataType.arrayElementType.width)
 					try:
-						initMem.store(offset, dataType.arrayElementType.width,
-							      value)
+						initMem.store(offset, memObj)
 					except AwlSimError as e:
 						raise AwlSimError("Data field '%s' initialization "
 							"is out of range." % str(rawField))
 			else:
 				assert(len(rawField.defaultInits) == 1)
 				value = dataType.parseMatchingImmediate(rawField.defaultInits[0].valueTokens)
+				memObj = make_AwlMemoryObject_fromGeneric(
+						value, dataType.width)
 				try:
-					initMem.store(make_AwlOffset(0, 0), dataType.width, value)
+					initMem.store(make_AwlOffset(0, 0), memObj)
 				except AwlSimError as e:
 					raise AwlSimError("Data field '%s' initialization "
 						"is out of range." % str(rawField))
+			initBytes = initMem.getDataBytes()
 		else:
 			initBytes = None
 		return name, dataType, initBytes
 
 	# Initialize a DB (global or instance) data field from a raw data-init.
 	def __initDBField(self, db, dataType, rawDataInit):
+#@cy		cdef AwlStructInstance structInstance
+#@cy		cdef AwlMemoryObject memObj
+
 		fieldName = rawDataInit.getIdentString()
 		value = dataType.parseMatchingImmediate(rawDataInit.valueTokens)
-		db.structInstance.setFieldDataByName(fieldName, value)
+		if dataType.type == dataType.TYPE_ARRAY:
+			# This is a single-array-element initializer.
+			width = dataType.arrayElementType.width
+		else:
+			width = dataType.width
+		memObj = make_AwlMemoryObject_fromGeneric(value, width)
+		structInstance = db.structInstance
+		structInstance.setFieldDataByName(fieldName, memObj)
 
 	# Create a DB data field.
 	def __createDBField(self, db, rawDataField):
@@ -298,6 +315,8 @@ class AwlTranslator(object):
 		return db
 
 	def __translateInstanceDB(self, rawDB):
+#@cy		cdef S7CPU cpu
+
 		if rawDB.fields:
 			raise AwlSimError("DB %d is an "
 				"instance DB, but it also "
@@ -321,12 +340,12 @@ class AwlTranslator(object):
 			isSFB = sym.type.type == AwlDataType.TYPE_SFB_X
 
 		# Get the FB/SFB code block
-		try:
-			if isSFB:
-				fb = self.cpu.sfbs[fbNumber]
-			else:
-				fb = self.cpu.fbs[fbNumber]
-		except KeyError:
+		cpu = self.cpu
+		if isSFB:
+			fb = cpu.getSFB(fbNumber)
+		else:
+			fb = cpu.getFB(fbNumber)
+		if not fb:
 			raise AwlSimError("Instance DB %d references %s, "
 				"but %s does not exist." %\
 				(rawDB.index, fbStr, fbStr))
@@ -504,6 +523,8 @@ class AwlSymResolver(object):
 	# to array variables are supported.
 	def resolveNamedLocal(self, block, insn, oper,
 			      pointer=False, allowWholeArrayAccess=False):
+#@cy		cdef S7CPU cpu
+
 		# Check whether we need to do something.
 		# Otherwise just return the source operator.
 		if pointer:
@@ -514,6 +535,7 @@ class AwlSymResolver(object):
 				return oper
 
 		AwlDataType = _getAwlDataTypeClass()
+		cpu = self.cpu
 
 		# Walk the ident chain to accumulate the sub-offsets
 		# for the ARRAY accesses.
@@ -546,9 +568,8 @@ class AwlSymResolver(object):
 			# Assign the struct to the UDT data type, if
 			# not already done so.
 			if dataType.type == AwlDataType.TYPE_UDT_X:
-				try:
-					udt = self.cpu.udts[dataType.index]
-				except KeyError as e:
+				udt = cpu.getUDT(dataType.index)
+				if not udt:
 					raise AwlSimError("UDT %d not found on CPU" %\
 						dataType.index)
 				assert(dataType._struct is None or
@@ -666,10 +687,12 @@ class AwlSymResolver(object):
 
 	# Get offset and width of a DB field.
 	def __dbVarToOffset(self, dbNumber, identChain, allowWholeArrayAccess=True):
+#@cy		cdef S7CPU cpu
+
 		# Get the DB
-		try:
-			db = self.cpu.dbs[dbNumber]
-		except KeyError as e:
+		cpu = self.cpu
+		db = cpu.getDB(dbNumber)
+		if not db:
 			raise AwlSimError("DB %d specified in fully qualified "
 				"operator does not exist." % dbNumber)
 
