@@ -2,7 +2,7 @@
 #
 # AWL simulator - project
 #
-# Copyright 2014-2018 Michael Buesch <m@bues.ch>
+# Copyright 2014-2020 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -41,6 +41,8 @@ import sys
 
 
 __all__ = [
+	"GuiCpuStateWindowSettings",
+	"GuiCpuStateViewSettings",
 	"GuiSettings",
 	"CoreLinkSettings",
 	"HwmodSettings",
@@ -48,15 +50,98 @@ __all__ = [
 ]
 
 
+class GuiCpuStateWindowSettingsFactory(XmlFactory):
+	def parser_open(self, tag=None):
+		cpuWinSettings = self.cpuWinSettings
+
+		if tag:
+			x = tag.getAttrInt("x", 0)
+			y = tag.getAttrInt("y", 0)
+			winType = tag.getAttrInt("type")
+			winConfig = tag.getAttr("config", "")
+			cpuWinSettings.setPosX(x)
+			cpuWinSettings.setPosY(y)
+			cpuWinSettings.setWindowType(winType)
+			cpuWinSettings.setWindowConfig(winConfig)
+		XmlFactory.parser_open(self, tag)
+
+	def parser_beginTag(self, tag):
+		XmlFactory.parser_beginTag(self, tag)
+
+	def parser_endTag(self, tag):
+		if tag.name == "cpu_view_window":
+			self.parser_finish()
+			return
+		XmlFactory.parser_endTag(self, tag)
+
+	def composer_getTags(self):
+		cpuWinSettings = self.cpuWinSettings
+
+		return [
+			self.Tag(name="cpu_view_window",
+				 attrs={
+					"x" : str(int(cpuWinSettings.getPosX())),
+					"y" : str(int(cpuWinSettings.getPosY())),
+					"type" : str(cpuWinSettings.getWindowType()),
+					"config" : str(cpuWinSettings.getWindowConfig()),
+				 },
+				 emitEmptyAttrs=True),
+		]
+
+class GuiCpuStateViewSettingsFactory(XmlFactory):
+	def parser_open(self, tag=None):
+		XmlFactory.parser_open(self, tag)
+
+	def parser_beginTag(self, tag):
+		cpuViewSettings = self.cpuViewSettings
+
+		if tag.name == "cpu_view_window":
+			cpuWinSettings = GuiCpuStateWindowSettings()
+			self.parser_switchTo(cpuWinSettings.factory(
+				cpuWinSettings=cpuWinSettings))
+			lst = cpuViewSettings.getCpuStateWindowSettingsList()
+			lst.append(cpuWinSettings)
+			cpuViewSettings.setCpuStateWindowSettingsList(lst)
+			return
+		XmlFactory.parser_beginTag(self, tag)
+
+	def parser_endTag(self, tag):
+		if tag.name == "cpu_view":
+			self.parser_finish()
+			return
+		XmlFactory.parser_endTag(self, tag)
+
+	def composer_getTags(self):
+		cpuViewSettings = self.cpuViewSettings
+
+		childTags = []
+		for cpuWinSettings in cpuViewSettings.getCpuStateWindowSettingsList():
+			childTags.extend(cpuWinSettings.factory(
+				cpuWinSettings=cpuWinSettings).composer_getTags())
+
+		return [
+			self.Tag(name="cpu_view",
+				 comment="\nCPU state view layout",
+				 attrs={
+				 },
+				 tags=childTags),
+		]
+
 class GuiSettingsFactory(XmlFactory):
 	def parser_open(self, tag=None):
+		self.inCore = False
 		self.inEditor = False
 		XmlFactory.parser_open(self, tag)
 
 	def parser_beginTag(self, tag):
 		guiSettings = self.guiSettings
 		if not self.inEditor:
-			if tag.name == "editor":
+			if tag.name == "core":
+				preDownloadValidation = tag.getAttrBool("pre_download_validation", True)
+				guiSettings.setPreDownloadValidationEn(preDownloadValidation)
+				self.inCore = True
+				return
+			elif tag.name == "editor":
 				autoIndent = tag.getAttrBool("autoindent", True)
 				pasteAutoIndent = tag.getAttrBool("paste_autoindent", True)
 				validation = tag.getAttrBool("validation", True)
@@ -68,6 +153,14 @@ class GuiSettingsFactory(XmlFactory):
 					guiSettings.setEditorFont(font)
 				self.inEditor = True
 				return
+			elif tag.name == "cpu_view":
+				cpuViewSettings = GuiCpuStateViewSettings()
+				self.parser_switchTo(cpuViewSettings.factory(
+					cpuViewSettings=cpuViewSettings))
+				lst = guiSettings.getCpuStateViewSettingsList()
+				lst.append(cpuViewSettings)
+				guiSettings.setCpuStateViewSettingsList(lst)
+				return
 		XmlFactory.parser_beginTag(self, tag)
 
 	def parser_endTag(self, tag):
@@ -75,6 +168,10 @@ class GuiSettingsFactory(XmlFactory):
 			if tag.name == "editor":
 				self.inEditor = False
 				return
+		elif self.inCore:
+			if tag.name == "core":
+				self.inCore = False
+				return
 		else:
 			if tag.name == "gui":
 				self.parser_finish()
@@ -86,7 +183,14 @@ class GuiSettingsFactory(XmlFactory):
 
 		childTags = []
 
+		childTags.append(self.Tag(name="core",
+					  comment="Core interaction settings",
+					  attrs={
+			"pre_download_validation" : str(int(guiSettings.getPreDownloadValidationEn())),
+		}))
+
 		childTags.append(self.Tag(name="editor",
+					  comment="\nAWL editor settings",
 					  attrs={
 			"autoindent"		: str(int(guiSettings.getEditorAutoIndentEn())),
 			"paste_autoindent"	: str(int(guiSettings.getEditorPasteIndentEn())),
@@ -94,6 +198,10 @@ class GuiSettingsFactory(XmlFactory):
 			"font"			: str(guiSettings.getEditorFont()),
 		}))
 
+		for cpuViewSettings in guiSettings.getCpuStateViewSettingsList():
+			childTags.extend(cpuViewSettings.factory(
+				cpuViewSettings=cpuViewSettings).composer_getTags())
+
 		tags = [
 			self.Tag(name="gui",
 				 comment="\nGraphical user interface configuration",
@@ -323,6 +431,8 @@ class ProjectFactory(XmlFactory):
 								   S7CPUConfig.DEFAULT_MNEMONICS)
 					cycleTimeLimitUs = tag.getAttrInt("cycle_time_limit_us",
 							S7CPUConfig.DEFAULT_CYCLETIMELIMIT_US)
+					cycleTimeTargetUs = tag.getAttrInt("cycle_time_target_us",
+							S7CPUConfig.DEFAULT_CYCLETIMETARGET_US)
 					runTimeLimitUs = tag.getAttrInt("run_time_limit_us",
 							S7CPUConfig.DEFAULT_RUNTIMELIMIT_US)
 					obStartEn = tag.getAttrBool("ob_startinfo_enable",
@@ -333,6 +443,7 @@ class ProjectFactory(XmlFactory):
 					conf.setClockMemByte(clockMem)
 					conf.setConfiguredMnemonics(mnemonics)
 					conf.setCycleTimeLimitUs(cycleTimeLimitUs)
+					conf.setCycleTimeTargetUs(cycleTimeTargetUs)
 					conf.setRunTimeLimitUs(runTimeLimitUs)
 					conf.setExtInsnsEn(extInsnsEn)
 					conf.setOBStartinfoEn(obStartEn)
@@ -413,7 +524,7 @@ class ProjectFactory(XmlFactory):
 		else:
 			if tag.name == "awlsim_project":
 				version = tag.getAttrInt("format_version")
-				if version != self.FILE_FORMAT_VERSION:
+				if version != self.FILE_FORMAT_VERSION: #@nocov
 					raise self.Error("Unsupported .awlpro format version. "
 						"Got %d, but expected %d." % (
 						version, self.FILE_FORMAT_VERSION))
@@ -424,14 +535,14 @@ class ProjectFactory(XmlFactory):
 						createDate = datetime.datetime.strptime(
 							createDate, project.DATETIME_FMT)
 						project.setCreateDate(createDate)
-				except (ValueError, TypeError) as e:
+				except (ValueError, TypeError) as e: #@nocov
 					pass
 				try:
 					if modifyDate:
 						modifyDate = datetime.datetime.strptime(
 							modifyDate, project.DATETIME_FMT)
 						project.setModifyDate(modifyDate)
-				except (ValueError, TypeError) as e:
+				except (ValueError, TypeError) as e: #@nocov
 					pass
 				self.inProject = True
 				return
@@ -508,6 +619,7 @@ class ProjectFactory(XmlFactory):
 					"clock_memory_byte"	: str(int(conf.clockMemByte)),
 					"mnemonics"		: str(int(conf.getConfiguredMnemonics())),
 					"cycle_time_limit_us"	: str(int(conf.cycleTimeLimitUs)),
+					"cycle_time_target_us"	: str(int(conf.cycleTimeTargetUs)),
 					"run_time_limit_us"	: str(int(conf.runTimeLimitUs)),
 				 })
 		]
@@ -594,6 +706,56 @@ class ProjectFactory(XmlFactory):
 
 		return tags
 
+class GuiCpuStateWindowSettings(object):
+	factory = GuiCpuStateWindowSettingsFactory
+
+	def __init__(self,
+		     posX=0,
+		     posY=0,
+		     windowType=0,
+		     windowConfig=""):
+		self.setPosX(posX)
+		self.setPosY(posY)
+		self.setWindowType(windowType)
+		self.setWindowConfig(windowConfig)
+
+	def setPosX(self, posX):
+		self.posX = posX
+
+	def getPosX(self):
+		return self.posX
+
+	def setPosY(self, posY):
+		self.posY = posY
+
+	def getPosY(self):
+		return self.posY
+
+	def setWindowType(self, windowType):
+		self.windowType = windowType or 0
+
+	def getWindowType(self):
+		return self.windowType
+
+	def setWindowConfig(self, windowConfig):
+		self.windowConfig = windowConfig or ""
+
+	def getWindowConfig(self):
+		return self.windowConfig
+
+class GuiCpuStateViewSettings(object):
+	factory = GuiCpuStateViewSettingsFactory
+
+	def __init__(self,
+		     cpuStateWindowSettingsList=None):
+		self.setCpuStateWindowSettingsList(cpuStateWindowSettingsList)
+
+	def setCpuStateWindowSettingsList(self, cpuStateWindowSettingsList):
+		self.cpuStateWindowSettingsList = cpuStateWindowSettingsList or []
+
+	def getCpuStateWindowSettingsList(self):
+		return self.cpuStateWindowSettingsList
+
 class GuiSettings(object):
 	factory	= GuiSettingsFactory
 
@@ -601,11 +763,15 @@ class GuiSettings(object):
 		     editorAutoIndentEn=True,
 		     editorPasteIndentEn=True,
 		     editorValidationEn=True,
-		     editorFont=""):
+		     editorFont="",
+		     cpuStateViewSettingsList=None,
+		     preDownloadValidationEn=True):
 		self.setEditorAutoIndentEn(editorAutoIndentEn)
 		self.setEditorPasteIndentEn(editorPasteIndentEn)
 		self.setEditorValidationEn(editorValidationEn)
 		self.setEditorFont(editorFont)
+		self.setCpuStateViewSettingsList(cpuStateViewSettingsList)
+		self.setPreDownloadValidationEn(preDownloadValidationEn)
 
 	def setEditorAutoIndentEn(self, editorAutoIndentEn):
 		self.editorAutoIndentEn = bool(editorAutoIndentEn)
@@ -631,10 +797,22 @@ class GuiSettings(object):
 	def getEditorFont(self):
 		return self.editorFont
 
+	def setCpuStateViewSettingsList(self, cpuStateViewSettingsList):
+		self.cpuStateViewSettingsList = cpuStateViewSettingsList or []
+
+	def getCpuStateViewSettingsList(self):
+		return self.cpuStateViewSettingsList
+
+	def setPreDownloadValidationEn(self, preDownloadValidationEn):
+		self.preDownloadValidationEn = bool(preDownloadValidationEn)
+
+	def getPreDownloadValidationEn(self):
+		return self.preDownloadValidationEn
+
 class CoreLinkSettings(object):
 	factory			= CoreLinkSettingsFactory
 
-	DEFAULT_INTERPRETERS	= "pypy3; pypy; $CURRENT; python3; python2; python; py"
+	DEFAULT_INTERPRETERS	= "pypy3; $CURRENT; python3; py"
 	DEFAULT_INTERP		= "$DEFAULT"
 
 	SPAWN_PORT_BASE		= 4151 + 32
@@ -990,6 +1168,15 @@ class Project(object):
 		project.projectDir = None
 		return project
 
+	@classmethod
+	def fromDataBytes(cls, dataBytes, projectFile=""):
+		try:
+			return cls.fromText(dataBytes.decode(cls.ENCODING), projectFile)
+		except UnicodeError as e:
+			raise AwlSimError("Project data: Failed to %s decode "
+				"project data: %s" % (
+				cls.ENCODING, str(e)))
+
 	@classmethod
 	def fromFile(cls, filename):
 		try:
@@ -1011,8 +1198,24 @@ class Project(object):
 			awlSrc = AwlSource.fromFile(name=filename,
 						    filepath=filename,
 						    compatReEncode=True)
-			project = Project(projectFile = None,
-					  awlSources = [ awlSrc, ])
+			project = Project(projectFile=None,
+					  awlSources=[ awlSrc, ])
+		return project
+
+	@classmethod
+	def fromProjectOrRawAwlData(cls, dataBytes):
+		"""Read project data (.awlpro) or raw AWL data (.awl)
+		and return a Project()."""
+
+		if Project.dataIsProject(dataBytes):
+			project = Project.fromDataBytes(dataBytes)
+		else:
+			# Make a fake project
+			awlSrc = AwlSource.fromBytes(name=None,
+						     sourceBytes=dataBytes,
+						     compatReEncode=True)
+			project = Project(projectFile=None,
+					  awlSources=[ awlSrc, ])
 		return project
 
 	def toText(self, projectFile=None):
@@ -1052,7 +1255,7 @@ class Project(object):
 		for symSrc in self.symTabSources:
 			symSrc.writeFileBacking(compatReEncode=True)
 
-	def __repr__(self):
+	def __repr__(self): #@nocov
 		if self.projectFile:
 			return 'Project("%s")' % self.projectFile
 		return "Project(None)"
