@@ -27,14 +27,14 @@ from awlsim.common.refmanager import *
 from awlsim.common.blockinfo import *
 from awlsim.common.wordpacker import *
 from awlsim.common.exceptions import *
+from awlsim.common.util import *
 
 from awlsim.core.blockinterface import *
-from awlsim.core.labels import *
+from awlsim.core.labels import * #+cimport
 from awlsim.core.datatypes import *
 from awlsim.core.memory import * #+cimport
 from awlsim.core.operatortypes import * #+cimport
 from awlsim.core.operators import * #+cimport
-from awlsim.core.util import *
 from awlsim.core.offset import * #+cimport
 from awlsim.core.lstack import * #+cimport
 
@@ -139,6 +139,7 @@ class CodeBlock(Block): #+cdef
 		self.insns = insns
 		self.nrInsns = len(insns)
 		self.labels = None
+		self.nrLabels = 0
 		self.interface = interface
 		self.tempAllocation = 0		# The number of allocated TEMP bytes
 		self.resolveLabels()
@@ -146,8 +147,10 @@ class CodeBlock(Block): #+cdef
 	def resolveLabels(self):
 		if self.insns:
 			self.labels = AwlLabel.resolveLabels(self.insns)
+			self.nrLabels = len(self.labels)
 		else:
 			self.labels = None
+			self.nrLabels = 0
 
 	def resolveSymbols(self):
 		pass
