@@ -2,7 +2,7 @@
 #
 # AWL simulator - data structs
 #
-# Copyright 2013-2015 Michael Buesch <m@bues.ch>
+# Copyright 2013-2018 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -25,12 +25,14 @@ from awlsim.common.compat import *
 
 from awlsim.common.datatypehelpers import * #+cimport
 from awlsim.common.exceptions import *
+from awlsim.common.util import *
 
-from awlsim.core.util import *
 from awlsim.core.identifier import *
 from awlsim.core.memory import * #+cimport
 from awlsim.core.offset import * #+cimport
 
+#from awlsim.core.cpu cimport * #@cy
+
 
 __all__ = [
 	"AwlStructField",
@@ -85,7 +87,7 @@ class AwlStructField(object): #+cdef
 		self.callByRef = bool(dataType.type in AwlDataType.callByRefTypes)
 
 		if self.initBytes is not None:
-			assert(len(self.initBytes) == self.byteSize)
+			assert(len_u32(self.initBytes) == self.byteSize)
 
 	def __eq__(self, other): #@nocy
 #@cy	cdef __eq(self, other):
@@ -114,7 +116,7 @@ class AwlStructField(object): #+cdef
 			return self.override.getFinalOverride()
 		return self
 
-	def __repr__(self):
+	def __repr__(self): #@nocov
 		s = "AwlStructField(\"%s\", P#%s, %s (%d bit), init=%s)" %\
 			(str(self.name),
 			 str(self.offset),
@@ -224,16 +226,17 @@ class AwlStruct(object): #+cdef
 		return baseField
 
 	def addField(self, cpu, name, dataType, initBytes=None):
+#@cy		cdef S7CPU _cpu
+
 		from awlsim.core.datatypes import AwlDataType
 
-		initMem = AwlMemory()
-		initMem.dataBytes = initBytes
+		_cpu = cpu
+		initMem = AwlMemory(initBytes)
 
 		if dataType.type == dataType.TYPE_UDT_X:
 			# Add an UDT.
-			try:
-				udt = cpu.udts[dataType.index]
-			except KeyError:
+			udt = _cpu.getUDT(dataType.index)
+			if not udt: #@nocov
 				assert(0) # Should never happen
 			assert(not initBytes)
 			# Assign the struct to the UDT data type, if
@@ -285,20 +288,20 @@ class AwlStruct(object): #+cdef
 					if not initBytes:
 						raise ValueError
 					fieldInitMem = AwlMemory(intDivRoundUp(childType.width, 8))
-					fieldInitData = fieldInitMem.dataBytes
-					fieldInitMem.store(make_AwlOffset(0, 0), childType.width,
-							    initMem.fetch(initOffset,
-									  childType.width))
+					fieldInitMem.store(make_AwlOffset(0, 0),
+							   initMem.fetch(initOffset,
+									 childType.width))
+					fieldInitData = fieldInitMem.getDataBytes()
 				except (AwlSimError, ValueError) as e:
 					fieldInitData = None
-				self.addField(cpu, str(childIdent), childType,
+				self.addField(_cpu, str(childIdent), childType,
 					      fieldInitData)
 				initOffset += make_AwlOffset_fromLongBitOffset(childType.width)
 				childIdent.advanceToNextArrayElement(dataType.arrayDimensions)
 				if childType.width > 8 and\
 				   intDivRoundUp(childType.width, 8) % 2 != 0:
 					# Align each element to 2-byte-boundary, if word or bigger.
-					self.addField(cpu, None, AwlDataType.makeByName("BYTE"))
+					self.addField(_cpu, None, AwlDataType.makeByName("BYTE"))
 			# Add a zero-length array-end guard field,
 			# to enforce alignment of following fields.
 			self.addDummyField()
@@ -343,7 +346,7 @@ class AwlStruct(object): #+cdef
 		except KeyError:
 			raise AwlSimError("Data structure field '%s' not found" % name)
 
-	def __repr__(self):
+	def __repr__(self): #@nocov
 		return "\n".join(str(field) for field in self.fields)
 
 class AwlStructInstance(object): #+cdef
@@ -366,30 +369,32 @@ class AwlStructInstance(object): #+cdef
 			if not field.initBytes:
 				continue
 			try:
-				self.memory.store(field.offset, field.bitSize,
-						  field.initBytes)
+				self.memory.store(field.offset,
+					make_AwlMemoryObject_fromBytes(field.initBytes,
+								       field.bitSize))
 			except AwlSimError as e:
 				raise AwlSimError("Data structure field '%s' "
 					"initialization is out of range." %\
 					str(field))
 
 	def getFieldData(self, field, baseOffset): #@nocy
-#@cy	cdef object getFieldData(self, AwlStructField field, AwlOffset baseOffset):
+#@cy	cdef AwlMemoryObject getFieldData(self, AwlStructField field, AwlOffset baseOffset) except NULL:
 		if baseOffset is None:
 			return self.memory.fetch(field.offset, field.bitSize)
-		return self.memory.fetch(baseOffset + field.offset, field.bitSize)
+		return self.memory.fetch(baseOffset.add(field.offset), field.bitSize)
 
-	def setFieldData(self, field, value, baseOffset): #@nocy
-#@cy	cdef setFieldData(self, AwlStructField field, object value, AwlOffset baseOffset):
+	def setFieldData(self, field, memObj, baseOffset): #@nocy
+#@cy	cdef setFieldData(self, AwlStructField field, AwlMemoryObject memObj, AwlOffset baseOffset):
+		AwlMemoryObject_assertWidth(memObj, field.bitSize)
 		if baseOffset is None:
-			self.memory.store(field.offset,
-					  field.bitSize, value)
+			self.memory.store(field.offset, memObj)
 		else:
-			self.memory.store(baseOffset + field.offset,
-					  field.bitSize, value)
+			self.memory.store(baseOffset.add(field.offset), memObj)
 
-	def getFieldDataByName(self, name):
+	def getFieldDataByName(self, name): #@nocy
+#@cy	cdef AwlMemoryObject getFieldDataByName(self, object name) except NULL:
 		return self.getFieldData(self._struct.getField(name), None)
 
-	def setFieldDataByName(self, name, value):
-		self.setFieldData(self._struct.getField(name), value, None)
+	def setFieldDataByName(self, name, memObj): #@nocy
+#@cy	cdef setFieldDataByName(self, object name, AwlMemoryObject memObj):
+		self.setFieldData(self._struct.getField(name), memObj, None)
