@@ -2,7 +2,7 @@
 #
 # AWL simulator - instructions
 #
-# Copyright 2012-2017 Michael Buesch <m@bues.ch>
+# Copyright 2012-2019 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -35,6 +35,8 @@ class AwlInsn_AbstractCall(AwlInsn): #+cdef
 	__slots__ = ()
 
 	def staticSanityChecks(self):
+#@cy		cdef AwlParamAssign param
+
 		if self.opCount == 1:
 			# "CALL FC/SFC" or
 			# "CALL #MULTIINSTANCE" or
@@ -42,15 +44,13 @@ class AwlInsn_AbstractCall(AwlInsn): #+cdef
 			blockOper = self.op0
 
 			if blockOper.operType == AwlOperatorTypes.BLKREF_FC:
-				try:
-					codeBlock = self.cpu.fcs[blockOper.offset.byteOffset]
-				except KeyError as e:
+				codeBlock = self.cpu.getFC(blockOper.offset.byteOffset)
+				if not codeBlock:
 					raise AwlSimError("Called FC not found",
 						rawInsn=self.getRawInsn())
 			elif blockOper.operType == AwlOperatorTypes.BLKREF_SFC:
-				try:
-					codeBlock = self.cpu.sfcs[blockOper.offset.byteOffset]
-				except KeyError as e:
+				codeBlock = self.cpu.getSFC(blockOper.offset.byteOffset)
+				if not codeBlock:
 					raise AwlSimError("SFC %d not implemented, yet" %\
 						blockOper.offset.byteOffset,
 						rawInsn=self.getRawInsn())
@@ -59,9 +59,8 @@ class AwlInsn_AbstractCall(AwlInsn): #+cdef
 					raise AwlSimError("Missing DB in function "
 						"block call",
 						rawInsn=self.getRawInsn())
-				try:
-					codeBlock = self.cpu.fbs[blockOper.offset.byteOffset]
-				except KeyError as e:
+				codeBlock = self.cpu.getFB(blockOper.offset.byteOffset)
+				if not codeBlock:
 					raise AwlSimError("Called FB not found",
 						rawInsn=self.getRawInsn())
 			elif blockOper.operType == AwlOperatorTypes.BLKREF_SFB:
@@ -69,9 +68,8 @@ class AwlInsn_AbstractCall(AwlInsn): #+cdef
 					raise AwlSimError("Missing DB in system function "
 						"block call",
 						rawInsn=self.getRawInsn())
-				try:
-					codeBlock = self.cpu.sfbs[blockOper.offset.byteOffset]
-				except KeyError as e:
+				codeBlock = self.cpu.getSFB(blockOper.offset.byteOffset)
+				if not codeBlock:
 					raise AwlSimError("SFB %d not implemented, yet" %\
 						blockOper.offset.byteOffset,
 						rawInsn=self.getRawInsn())
@@ -81,9 +79,13 @@ class AwlInsn_AbstractCall(AwlInsn): #+cdef
 			elif blockOper.operType in (AwlOperatorTypes.MULTI_FB, AwlOperatorTypes.MULTI_SFB):
 				# Multi instance call (like CALL #FOO)
 				if blockOper.operType == AwlOperatorTypes.MULTI_FB:
-					codeBlock = self.cpu.fbs[blockOper.offset.fbNumber]
+					codeBlock = self.cpu.getFB(blockOper.offset.fbNumber)
+					if not codeBlock:
+						pass#TODO
 				else:
-					codeBlock = self.cpu.sfbs[blockOper.offset.fbNumber]
+					codeBlock = self.cpu.getSFB(blockOper.offset.fbNumber)
+					if not codeBlock:
+						pass#TODO
 			else:
 				raise AwlSimError("Invalid CALL operand",
 					rawInsn=self.getRawInsn())
@@ -107,9 +109,8 @@ class AwlInsn_AbstractCall(AwlInsn): #+cdef
 				raise AwlSimError("Second CALL operand is "
 					"not a DB operand.",
 					rawInsn=self.getRawInsn())
-			try:
-				db = self.cpu.dbs[dbOper.offset.byteOffset]
-			except KeyError as e:
+			db = self.cpu.getDB(dbOper.offset.byteOffset)
+			if not db:
 				raise AwlSimError("DB used in FB call not found",
 					rawInsn=self.getRawInsn())
 			if not db.isInstanceDB():
@@ -118,17 +119,15 @@ class AwlInsn_AbstractCall(AwlInsn): #+cdef
 					rawInsn=self.getRawInsn())
 
 			if blockOper.operType == AwlOperatorTypes.BLKREF_FB:
-				try:
-					fb = self.cpu.fbs[blockOper.offset.byteOffset]
-				except KeyError as e:
+				fb = self.cpu.getFB(blockOper.offset.byteOffset)
+				if not fb:
 					raise AwlSimError("Called FB not found",
 						rawInsn=self.getRawInsn())
 				# TODO check if this is an FB-DB
 				pass#TODO
 			elif blockOper.operType == AwlOperatorTypes.BLKREF_SFB:
-				try:
-					fb = self.cpu.sfbs[blockOper.offset.byteOffset]
-				except KeyError as e:
+				fb = self.cpu.getSFB(blockOper.offset.byteOffset)
+				if not fb:
 					raise AwlSimError("SFB %d not implemented, yet" %\
 						blockOper.offset.byteOffset,
 						rawInsn=self.getRawInsn())
@@ -169,39 +168,22 @@ class AwlInsn_AbstractCall(AwlInsn): #+cdef
 
 class AwlInsn_CALL(AwlInsn_AbstractCall): #+cdef
 
-	__slots__ = (
-		"run",
-	)
+	__slots__ = ()
 
 	def __init__(self, cpu, rawInsn=None, **kwargs):
 		AwlInsn_AbstractCall.__init__(self, cpu, AwlInsn.TYPE_CALL, rawInsn, **kwargs)
 		self.assertOpCount((1, 2))
 
-		if self.opCount == 1:			#@nocy
-			self.run = self.__run_CALL_FC	#@nocy
-		else:					#@nocy
-			self.run = self.__run_CALL_FB	#@nocy
-
-	def __run_CALL_FC(self): #+cdef
+	def run(self): #+cdef
 #@cy		cdef S7StatusWord s
 
-		self.cpu.run_CALL(self.op0, None, self.params, False)
+		if self.opCount == 1:
+			self.cpu.run_CALL(self.op0, None, self.params, False)
+		else:
+			self.cpu.run_CALL(self.op0, self.op1, self.params, False)
 		s = self.cpu.statusWord
 		s.OS, s.OR, s.STA, s.NER = 0, 0, 1, 0
 
-	def __run_CALL_FB(self): #+cdef
-#@cy		cdef S7StatusWord s
-
-		self.cpu.run_CALL(self.op0, self.op1, self.params, False)
-		s = self.cpu.statusWord
-		s.OS, s.OR, s.STA, s.NER = 0, 0, 1, 0
-
-#@cy	cdef run(self):
-#@cy		if self.opCount == 1:
-#@cy			self.__run_CALL_FC()
-#@cy		else:
-#@cy			self.__run_CALL_FB()
-
 class AwlInsn_CC(AwlInsn_AbstractCall): #+cdef
 
 	__slots__ = ()
