@@ -2,7 +2,7 @@
 #
 # AWL simulator - instructions
 #
-# Copyright 2012-2017 Michael Buesch <m@bues.ch>
+# Copyright 2012-2018 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -29,6 +29,8 @@ from awlsim.core.instructions.main import * #+cimport
 from awlsim.core.operatortypes import * #+cimport
 from awlsim.core.operators import * #+cimport
 
+#from libc.stdlib cimport abs #@cy
+
 
 class AwlInsn_DI_D(AwlInsn): #+cdef
 
@@ -41,22 +43,26 @@ class AwlInsn_DI_D(AwlInsn): #+cdef
 	def run(self): #+cdef
 #@cy		cdef S7StatusWord s
 #@cy		cdef int32_t accu1
+#@cy		cdef uint32_t accu1abs
 #@cy		cdef int32_t accu2
+#@cy		cdef uint32_t accu2abs
 #@cy		cdef int64_t quo
 
 		s = self.cpu.statusWord
 		accu2, accu1 = self.cpu.accu2.getSignedDWord(),\
 			       self.cpu.accu1.getSignedDWord()
 		if self.cpu.is4accu:
-			self.cpu.accu2.setDWord(self.cpu.accu3.getDWord())
-			self.cpu.accu3.setDWord(self.cpu.accu4.getDWord())
-		try:
-			quo = abs(accu2) // abs(accu1)
-			if int(accu1 < 0) ^ int(accu2 < 0):
-				quo = -quo
-		except ZeroDivisionError:
+			self.cpu.accu2.copyFrom(self.cpu.accu3)
+			self.cpu.accu3.copyFrom(self.cpu.accu4)
+		accu1abs = abs(accu1)
+		accu2abs = abs(accu2)
+		if accu1abs == 0: #+unlikely
 			s.A1, s.A0, s.OV, s.OS = 1, 1, 1, 1
 			return
+		quo = accu2abs // accu1abs
+		if int(accu1 < 0) ^ int(accu2 < 0): #@nocy
+#@cy		if <uint32_t>(accu1 < 0) ^ <uint32_t>(accu2 < 0):
+			quo = -quo
 		self.cpu.accu1.setDWord(quo)
 		if quo == 0:
 			s.A1, s.A0, s.OV = 0, 0, 0
@@ -64,5 +70,5 @@ class AwlInsn_DI_D(AwlInsn): #+cdef
 			s.A1, s.A0, s.OV = 0, 1, 0
 		else:
 			s.A1, s.A0, s.OV = 1, 0, 0
-		if quo > 0x7FFFFFFF:
+		if quo > 0x7FFFFFFF: #+unlikely
 			s.OV, s.OS = 1, 1
