@@ -2,7 +2,7 @@
 #
 # AWL simulator - instructions
 #
-# Copyright 2012-2017 Michael Buesch <m@bues.ch>
+# Copyright 2012-2018 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -30,7 +30,8 @@ from awlsim.core.instructions.main import * #+cimport
 from awlsim.core.operatortypes import * #+cimport
 from awlsim.core.operators import * #+cimport
 
-import math
+from math import log #@nocy
+#from libc.math cimport log #@cy
 
 
 class AwlInsn_LN(AwlInsn): #+cdef
@@ -42,16 +43,21 @@ class AwlInsn_LN(AwlInsn): #+cdef
 		self.assertOpCount(0)
 
 	def run(self): #+cdef
-#@cy		cdef double accu1
+#@cy		cdef double accu1Float
+#@cy		cdef uint32_t accu1DWord
 
-		accu1 = self.cpu.accu1.getPyFloat()
-		try:
-			if accu1 == 0.0:
-				raise ValueError
-			accu1 = math.log(accu1)
-		except ValueError:
-			self.cpu.accu1.setDWord(floatConst.pNaNDWord)
-			accu1 = floatConst.nNaNFloat
+		accu1DWord = self.cpu.accu1.get()
+		accu1Float = self.cpu.accu1.getPyFloat()
+		if isNaN(accu1DWord):
+			self.cpu.accu1.set(floatConst.pNaNDWord)
+			accu1Float = self.cpu.accu1.getPyFloat()
+		elif accu1DWord == floatConst.negInfDWord or\
+		     accu1DWord == floatConst.posInfDWord:
+			pass
+		elif accu1Float <= 0.0:
+			accu1Float = floatConst.nNaNFloat
+			self.cpu.accu1.set(floatConst.pNaNDWord)
 		else:
-			self.cpu.accu1.setPyFloat(accu1)
-		self.cpu.statusWord.setForFloatingPoint(accu1)
+			accu1Float = log(accu1Float)
+			self.cpu.accu1.setPyFloat(accu1Float)
+		self.cpu.statusWord.setForFloatingPoint(accu1Float)
