@@ -2,7 +2,7 @@
 #
 # AWL simulator - instructions
 #
-# Copyright 2012-2017 Michael Buesch <m@bues.ch>
+# Copyright 2012-2018 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -40,7 +40,7 @@ class AwlInsn_MI_I(AwlInsn): #+cdef
 
 	def run(self): #+cdef
 #@cy		cdef S7StatusWord s
-#@cy		cdef int32_t accu1
+#@cy		cdef uint16_t diffTrunc
 #@cy		cdef int32_t diff
 
 		s = self.cpu.statusWord
@@ -48,12 +48,13 @@ class AwlInsn_MI_I(AwlInsn): #+cdef
 		diff -= self.cpu.accu1.getSignedWord()
 		self.cpu.accu1.setWord(diff)
 		if self.cpu.is4accu:
-			self.cpu.accu2.setDWord(self.cpu.accu3.getDWord())
-			self.cpu.accu3.setDWord(self.cpu.accu4.getDWord())
-		accu1 = self.cpu.accu1.getSignedWord()
-		if accu1 == 0:
+			self.cpu.accu2.copyFrom(self.cpu.accu3)
+			self.cpu.accu3.copyFrom(self.cpu.accu4)
+		diffTrunc = diff & 0xFFFF #@nocy
+#@cy		diffTrunc = <uint16_t>diff
+		if diffTrunc == 0:
 			s.A1, s.A0, s.OV = 0, 0, 0
-		elif accu1 < 0:
+		elif diffTrunc & 0x8000: #+suffix-u
 			s.A1, s.A0, s.OV = 0, 1, 0
 		else:
 			s.A1, s.A0, s.OV = 1, 0, 0
