@@ -2,7 +2,7 @@
 #
 # AWL simulator - instructions
 #
-# Copyright 2012-2017 Michael Buesch <m@bues.ch>
+# Copyright 2012-2018 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -40,7 +40,7 @@ class AwlInsn_PL_D(AwlInsn): #+cdef
 
 	def run(self): #+cdef
 #@cy		cdef S7StatusWord s
-#@cy		cdef int32_t accu1
+#@cy		cdef uint32_t sumTrunc
 #@cy		cdef int64_t _sum
 
 		s = self.cpu.statusWord
@@ -48,15 +48,15 @@ class AwlInsn_PL_D(AwlInsn): #+cdef
 		_sum += self.cpu.accu1.getSignedDWord()
 		self.cpu.accu1.setDWord(_sum)
 		if self.cpu.is4accu:
-			self.cpu.accu2.setDWord(self.cpu.accu3.getDWord())
-			self.cpu.accu3.setDWord(self.cpu.accu4.getDWord())
-		accu1 = self.cpu.accu1.getSignedDWord()
-		if accu1 == 0:
+			self.cpu.accu2.copyFrom(self.cpu.accu3)
+			self.cpu.accu3.copyFrom(self.cpu.accu4)
+		sumTrunc = _sum & 0xFFFFFFFF #@nocy
+#@cy		sumTrunc = <uint32_t>_sum
+		if sumTrunc == 0:
 			s.A1, s.A0, s.OV = 0, 0, 0
-		elif accu1 < 0:
+		elif sumTrunc & 0x80000000: #+suffix-u
 			s.A1, s.A0, s.OV = 0, 1, 0
 		else:
 			s.A1, s.A0, s.OV = 1, 0, 0
-		if _sum > 0x7FFFFFFF or _sum < -2147483648: #@nocy
-#@cy		if _sum > 0x7FFFFFFFLL or _sum < -2147483648LL:
+		if _sum > 0x7FFFFFFF or _sum < -2147483648: #+suffix-LL
 			s.OV, s.OS = 1, 1
