@@ -2,7 +2,7 @@
 #
 # AWL simulator - instructions
 #
-# Copyright 2012-2017 Michael Buesch <m@bues.ch>
+# Copyright 2012-2018 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -43,7 +43,7 @@ class AwlInsn_SLW(AwlInsn): #+cdef
 	def run(self): #+cdef
 #@cy		cdef S7StatusWord s
 #@cy		cdef uint32_t accu1
-#@cy		cdef int64_t count
+#@cy		cdef uint32_t count
 
 		s = self.cpu.statusWord
 		accu1 = self.cpu.accu1.getWord()
@@ -53,7 +53,8 @@ class AwlInsn_SLW(AwlInsn): #+cdef
 			count = self.cpu.accu2.getByte()
 		if count <= 0:
 			return
-		count = min(count, 16)
-		s.A1, s.A0, s.OV = (accu1 >> (16 - count)) & 1, 0, 0
+		if count > 16:
+			count = 16
+		s.A1, s.A0, s.OV = (accu1 >> (16 - count)) & 1, 0, 0	#+suffix-u
 		accu1 <<= count
 		self.cpu.accu1.setWord(accu1)
