@@ -2,7 +2,7 @@
 #
 # AWL simulator
 #
-# Copyright 2012-2017 Michael Buesch <m@bues.ch>
+# Copyright 2012-2019 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -26,14 +26,17 @@ from awlsim.common.compat import *
 from awlsim.common.project import *
 from awlsim.common.env import *
 from awlsim.common.exceptions import *
+from awlsim.common.profiler import *
+from awlsim.common.util import *
 
-from awlsim.core.util import *
 from awlsim.core.cpu import * #+cimport
 from awlsim.core.hardware import * #+cimport
 from awlsim.core.hardware_loader import *
 
 import sys
 
+#cimport cython #@cy
+
 
 __all__ = [
 	"AwlSim",
@@ -47,12 +50,12 @@ def AwlSim_decorator_profiled(profileLevel):
 		@functools.wraps(func) #@nocy
 		def profiled_wrapper(self, *args, **kwargs):
 			if self._profileLevel >= profileLevel:
-				self._profileStart()
+				self._profileStart() #@nocov
 			try:
 				func(self, *args, **kwargs)
 			finally:
 				if self._profileLevel >= profileLevel:
-					self._profileStop()
+					self._profileStop() #@nocov
 		return profiled_wrapper
 	return profiled_decorator
 
@@ -77,6 +80,7 @@ class AwlSim(object): #+cdef
 	def __init__(self):
 		self.__registeredHardware = []
 		self.__registeredHardwareCount = 0
+		self.__hwStartupRequired = True
 		self._fatalHwErrors = True
 		self.cpu = S7CPU()
 		self.cpu.setPeripheralReadCallback(self.__peripheralReadCallback)
@@ -87,46 +91,23 @@ class AwlSim(object): #+cdef
 	def getCPU(self):
 		return self.cpu
 
-	def __setProfiler(self, profileLevel):
+	def __setProfiler(self, profileLevel): #@nocov
 		self._profileLevel = profileLevel
 		if self._profileLevel <= 0:
 			return
 
-		try:
-			import cProfile as profileModule
-		except ImportError:
-			profileModule = None
-		self.__profileModule = profileModule
-		try:
-			import pstats as pstatsModule
-		except ImportError:
-			pstatsModule = None
-		self.__pstatsModule = pstatsModule
+		self.__profiler = Profiler()
 
-		if not self.__profileModule or\
-		   not self.__pstatsModule:
-			raise AwlSimError("Failed to load cProfile/pstats modules. "
-				"Cannot enable profiling.")
+	def _profileStart(self): #@nocov
+		self.__profiler.start()
 
-		self.__profiler = self.__profileModule.Profile()
+	def _profileStop(self): #@nocov
+		self.__profiler.stop()
 
-	def _profileStart(self):
-		self.__profiler.enable()
-
-	def _profileStop(self):
-		self.__profiler.disable()
-
-	def getProfileStats(self):
+	def getProfileStats(self): #@nocov
 		if self._profileLevel <= 0:
 			return None
-
-		sio = StringIO()
-		ps = self.__pstatsModule.Stats(self.__profiler,
-					       stream = sio)
-		ps.sort_stats("time")
-		ps.print_stats()
-
-		return sio.getvalue()
+		return self.__profiler.getResult()
 
 	def _handleSimException(self, e, fatal = True):
 		"""Handle an exception and add some information
@@ -142,7 +123,7 @@ class AwlSim(object): #+cdef
 			raise e
 		else:
 			# Non-fatal. Just log an error.
-			printError(str(e))
+			printError(str(e)) #@nocov
 
 	@throwsAwlSimError
 	def __handleMaintenanceRequest(self, e):
@@ -158,14 +139,14 @@ class AwlSim(object): #+cdef
 				self.cpu.startup()
 			else:
 				assert(0)
-		except MaintenanceRequest as e:
+		except MaintenanceRequest as e: #@nocov
 			raise AwlSimError("Recursive maintenance request")
 
 	@profiled(2)
 	@throwsAwlSimError
 	def reset(self):
 		self.cpu.reset()
-		self.unregisterAllHardware(inCpuReset = True)
+		self.unregisterAllHardware()
 
 	@profiled(2)
 	@throwsAwlSimError
@@ -200,7 +181,13 @@ class AwlSim(object): #+cdef
 	@profiled(2)
 	@throwsAwlSimError
 	def startup(self):
-		self.__hwStartup()
+		# Startup the hardware modules, if required.
+		if self.__hwStartupRequired:
+			self.hardwareStartup()
+		# Next time we need a startup again.
+		self.__hwStartupRequired = True
+
+		# Startup the CPU core.
 		try:
 			self.__readHwInputs()
 			self.cpu.startup()
@@ -210,7 +197,7 @@ class AwlSim(object): #+cdef
 
 	def runCycle(self): #+cpdef
 		if self._profileLevel >= 1:
-			self._profileStart()
+			self._profileStart() #@nocov
 
 		try:
 			if self.__registeredHardwareCount:
@@ -218,13 +205,14 @@ class AwlSim(object): #+cdef
 			self.cpu.runCycle()
 			if self.__registeredHardwareCount:
 				self.__writeHwOutputs()
+			self.cpu.sleepCyclePadding()
 		except AwlSimError as e:
 			self._handleSimException(e)
 		except MaintenanceRequest as e:
 			self.__handleMaintenanceRequest(e)
 
 		if self._profileLevel >= 1:
-			self._profileStop()
+			self._profileStop() #@nocov
 
 	@throwsAwlSimError
 	def shutdown(self):
@@ -233,7 +221,7 @@ class AwlSim(object): #+cdef
 		"""
 		self.unregisterAllHardware()
 		ps = self.getProfileStats()
-		if ps:
+		if ps: #@nocov
 			sys.stdout.write("\n\nAwlsim core profile stats "
 					 "(level %d) follow:\n" %\
 					 self._profileLevel)
@@ -241,18 +229,11 @@ class AwlSim(object): #+cdef
 			sys.stdout.write("\n")
 			sys.stdout.flush()
 
-	def unregisterAllHardware(self, inCpuReset = False):
-		newHwList = []
+	def unregisterAllHardware(self):
 		for hw in self.__registeredHardware:
-			if not hw.getParam("removeOnReset") and inCpuReset:
-				# This module has "removeOnReset" set to False
-				# and we are in a CPU reset.
-				# Do shutdown the module, but keep it in the
-				# list of registered modules.
-				newHwList.append(hw)
 			hw.shutdown()
-		self.__registeredHardware = newHwList
-		self.__registeredHardwareCount = len(self.__registeredHardware)
+		self.__registeredHardware = []
+		self.__registeredHardwareCount = 0
 
 	def registerHardware(self, hwClassInst):
 		"""Register a new hardware interface."""
@@ -260,6 +241,7 @@ class AwlSim(object): #+cdef
 		if hwClassInst.getParamValueByName("enabled"):
 			self.__registeredHardware.append(hwClassInst)
 			self.__registeredHardwareCount = len(self.__registeredHardware)
+			self.__hwStartupRequired = True
 
 	def registerHardwareClass(self, hwClass, parameters={}):
 		"""Register a new hardware interface class.
@@ -279,7 +261,9 @@ class AwlSim(object): #+cdef
 
 		return HwModLoader.loadModule(name).getInterface()
 
-	def __hwStartup(self):
+	@profiled(2)
+	@throwsAwlSimError
+	def hardwareStartup(self):
 		"""Startup all attached hardware modules.
 		"""
 
@@ -292,27 +276,40 @@ class AwlSim(object): #+cdef
 				hw.startup()
 			except AwlSimError as e:
 				# Always fatal in startup.
-				self._handleSimException(e, fatal = True)
+				self._handleSimException(e, fatal=True)
+		self.__hwStartupRequired = False
 
+#@cy	@cython.boundscheck(False)
 	def __readHwInputs(self): #+cdef
 		"""Read all hardware module inputs.
 		"""
 #@cy		cdef AbstractHardwareInterface hw
+#@cy		cdef uint32_t i
 
-		for hw in self.__registeredHardware:
+		# Note: Bounds checking of the indexing operator [] is disabled
+		#       by @cython.boundscheck(False) in this method.
+
+		for i in range(self.__registeredHardwareCount):
 			try:
+				hw = self.__registeredHardware[i]
 				hw.readInputs()
 			except AwlSimError as e:
 				self._handleSimException(e,
 					fatal = self._fatalHwErrors)
 
+#@cy	@cython.boundscheck(False)
 	def __writeHwOutputs(self): #+cdef
 		"""Write all hardware module outputs.
 		"""
 #@cy		cdef AbstractHardwareInterface hw
+#@cy		cdef uint32_t i
 
-		for hw in self.__registeredHardware:
+		# Note: Bounds checking of the indexing operator [] is disabled
+		#       by @cython.boundscheck(False) in this method.
+
+		for i in range(self.__registeredHardwareCount):
 			try:
+				hw = self.__registeredHardware[i]
 				hw.writeOutputs()
 			except AwlSimError as e:
 				self._handleSimException(e,
@@ -358,5 +355,5 @@ class AwlSim(object): #+cdef
 			return False
 		return retOk
 
-	def __repr__(self):
+	def __repr__(self): #@nocov
 		return str(self.cpu)
