@@ -2,7 +2,7 @@
 #
 # AWL simulator - operators
 #
-# Copyright 2012-2017 Michael Buesch <m@bues.ch>
+# Copyright 2012-2018 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -25,13 +25,13 @@ from awlsim.common.compat import *
 
 from awlsim.common.datatypehelpers import * #+cimport
 from awlsim.common.exceptions import *
+from awlsim.common.util import *
 
 from awlsim.core.operatortypes import * #+cimport
 from awlsim.core.memory import * #+cimport
 from awlsim.core.offset import * #+cimport
 from awlsim.core.statusword import * #+cimport
 from awlsim.core.timers import * #+cimport
-from awlsim.core.util import *
 
 
 class AwlOperator(object): #+cdef
@@ -61,7 +61,7 @@ class AwlOperator(object): #+cdef
 	isExtended = False #@nocy
 
 	# Possible label index.
-	labelIndex = None #@nocy
+	labelIndex = 0xFFFFFFFF #@nocy
 
 	# Interface index number.
 	# May be set by the symbol resolver.
@@ -76,18 +76,18 @@ class AwlOperator(object): #+cdef
 	# Only set for resolved symbolic accesses.
 	dataType = None #@nocy
 
-#@cy	cdef _cy_init(self):
+#@cy	cdef void _cy_init(self):
 #@cy		self.immediate = 0
 #@cy		self.immediateBytes = None
 #@cy		self.pointer = None
 #@cy		self.isExtended = False
-#@cy		self.labelIndex = None
+#@cy		self.labelIndex = 0xFFFFFFFF
 #@cy		self.interfaceIndex = 0xFFFF
 #@cy		self.compound = False
 #@cy		self.dataType = None
 
 	def __eq__(self, other): #@nocy
-#@cy	cdef __eq(self, AwlOperator other):
+#@cy	cdef _Bool __eq(self, AwlOperator other):
 		return (self is other) or (\
 			isinstance(other, AwlOperator) and\
 			self.operType == other.operType and\
@@ -138,7 +138,7 @@ class AwlOperator(object): #+cdef
 		from awlsim.common.sources import AwlSource
 		try:
 			return self.immediateBytes.decode(AwlSource.COMPAT_ENCODING)
-		except UnicodeError as e:
+		except UnicodeError as e: #@nocov
 			raise AwlSimError("Invalid characters in operator (decode).")
 
 	@immediateStr.setter
@@ -146,7 +146,7 @@ class AwlOperator(object): #+cdef
 		from awlsim.common.sources import AwlSource
 		try:
 			self.immediateBytes = bytearray(newStr.encode(AwlSource.COMPAT_ENCODING))
-		except UnicodeError as e:
+		except UnicodeError as e: #@nocov
 			raise AwlSimError("Invalid characters in operator (encode).")
 
 	def setInsn(self, newInsn):
@@ -159,7 +159,7 @@ class AwlOperator(object): #+cdef
 		self.labelIndex = newLabelIndex
 
 	def isImmediate(self): #@nocy
-#@cy	cpdef _Bool isImmediate(self):
+#@cy	cdef _Bool isImmediate(self):
 		return self.operType > AwlOperatorTypes._IMM_START and\
 		       self.operType < AwlOperatorTypes._IMM_END
 
@@ -190,15 +190,17 @@ class AwlOperator(object): #+cdef
 				raise AwlSimError("Operator value too big",
 						  insn=self.insn)
 		if widths is not None:
-			if not (AwlOperatorWidths.makeMask(self.width) & widths):
+			if not (makeAwlOperatorWidthMask(self.width) & widths):
 				raise AwlSimError("Invalid operator width. "
 					"Got %d, but expected %s." % (
 					self.width,
 					listToHumanStr(AwlOperatorWidths.maskToList(widths))))
 
-	def checkDataTypeCompat(self, cpu, dataType):
+	def checkDataTypeCompat(self, _cpu, dataType):
+#@cy		cdef S7CPU cpu
 		from awlsim.core.datatypes import AwlDataType
 
+		cpu = _cpu
 		assert(isinstance(dataType, AwlDataType))
 
 		if self.operType in (AwlOperatorTypes.NAMED_LOCAL,
@@ -216,11 +218,8 @@ class AwlOperator(object): #+cdef
 				(str(dataType), dataType.width, str(oper), operWidth))
 
 		if dataType.type == AwlDataType.TYPE_UDT_X:
-			try:
-				udt = cpu.udts[dataType.index]
-				if udt._struct.getSize() * 8 != self.width:
-					raise ValueError
-			except (KeyError, ValueError) as e:
+			udt = cpu.getUDT(dataType.index)
+			if not udt or udt._struct.getSize() * 8 != self.width:
 				mismatch(dataType, self, self.width)
 		elif dataType.type in {AwlDataType.TYPE_POINTER,
 				       AwlDataType.TYPE_ANY}:
@@ -244,7 +243,8 @@ class AwlOperator(object): #+cdef
 						"must be only one single character "
 						"long.")
 			else:
-				if self.isImmediate():
+				if (self.isImmediate() and
+				    self.operType != AwlOperatorTypes.IMM):
 					raise AwlSimError("Invalid immediate '%s'"
 						"for CHAR data type." %\
 						str(self))
@@ -269,8 +269,8 @@ class AwlOperator(object): #+cdef
 				mismatch(dataType, self, self.width)
 
 	# Resolve this indirect operator to a direct operator.
-	def resolve(self, store=True): #@nocy
-#@cy	cpdef AwlOperator resolve(self, _Bool store=True):
+	def resolve(self, store): #@nocy
+#@cy	cdef AwlOperator resolve(self, _Bool store):
 		# This already is a direct operator.
 		if self.operType == AwlOperatorTypes.NAMED_LOCAL:
 			# This is a named-local access (#abc).
@@ -284,12 +284,11 @@ class AwlOperator(object): #+cdef
 
 	# Make an area-spanning pointer value (32 bit) to this memory area.
 	def makePointerValue(self): #@nocy
-#@cy	cpdef uint32_t makePointerValue(self):
-#@cy		cdef uint32_t area
+#@cy	cdef uint32_t makePointerValue(self) except? 0xFFFFFFFF:
+#@cy		cdef int64_t area
 
-		try:
-			area = AwlIndirectOpConst.optype2area[self.operType]
-		except KeyError as e:
+		area = AwlIndirectOpConst.optype2area(self.operType)
+		if area < 0 or area > 0xFFFFFFFF: #+suffix-u #+unlikely
 			raise AwlSimError("Could not transform operator '%s' "
 				"into a pointer." % str(self))
 		return area | self.offset.toPointerValue()
@@ -330,7 +329,7 @@ class AwlOperator(object): #+cdef
 		if self.operType == AwlOperatorTypes.IMM_REAL:
 			return str(dwordToPyFloat(self.immediate))
 		elif self.operType == AwlOperatorTypes.IMM_S5T:
-			seconds = Timer.s5t_to_seconds(self.immediate)
+			seconds = Timer_s5t_to_seconds(self.immediate)
 			return "S5T#" + AwlDataType.formatTime(seconds)
 		elif self.operType == AwlOperatorTypes.IMM_TIME:
 			return "T#" + AwlDataType.formatTime(self.immediate / 1000.0)
@@ -459,7 +458,7 @@ class AwlOperator(object): #+cdef
 			return "__UNSPEC"
 		try:
 			return self.type2str[self.operType]
-		except KeyError:
+		except KeyError: #@nocov
 			assert(0)
 
 #
@@ -503,28 +502,21 @@ class AwlIndirectOp(AwlOperator): #+cdef
 
 	def assertType(self, types, lowerLimit=None, upperLimit=None):
 		types = toSet(types)
-		if not AwlIndirectOpConst.area2optype_fetch[self.area] in types and\
-		   not AwlIndirectOpConst.area2optype_store[self.area] in types:
-			self._raiseTypeError(AwlIndirectOpConst.area2optype_fetch[self.area], types)
+		if not AwlIndirectOpConst.area2optype(self.area, False) in types and\
+		   not AwlIndirectOpConst.area2optype(self.area, True) in types:
+			self._raiseTypeError(AwlIndirectOpConst.area2optype(self.area, False), types)
 		assert(lowerLimit is None)
 		assert(upperLimit is None)
 
-	# Possible offset oper types for indirect access
-	__possibleOffsetOperTypes = (AwlOperatorTypes.MEM_M,
-				     AwlOperatorTypes.MEM_L,
-				     AwlOperatorTypes.MEM_DB,
-				     AwlOperatorTypes.MEM_DI)
-
 	# Resolve this indirect operator to a direct operator.
-	def resolve(self, store=True): #@nocy
-#@cy	cpdef AwlOperator resolve(self, _Bool store=True):
+	def resolve(self, store): #@nocy
+#@cy	cdef AwlOperator resolve(self, _Bool store):
 #@cy		cdef _Bool bitwiseDirectOffset
 #@cy		cdef AwlOffset directOffset
 #@cy		cdef AwlOperator offsetOper
-#@cy		cdef set possibleWidths
 #@cy		cdef uint32_t offsetValue
 #@cy		cdef uint64_t pointer
-#@cy		cdef uint32_t optype
+#@cy		cdef int32_t optype
 
 		bitwiseDirectOffset = True
 		offsetOper = self.offsetOper
@@ -537,49 +529,54 @@ class AwlIndirectOp(AwlOperator): #+cdef
 					"address register.")
 			if self.area > PointerConst.AREA_MASK_S:
 				# Is extended area
-				possibleWidths = {8, 16, 32}
 				bitwiseDirectOffset = False
+				if offsetOper.width not in (8, 16, 32):
+					raise AwlSimError("Offset operator in indirect "
+						"access is not of 8, 16 or 32 bit width.")
 			else:
 				# Is standard area
-				possibleWidths = {32,}
-			if offsetOper.operType not in self.__possibleOffsetOperTypes:
+				if offsetOper.width != 32:
+					raise AwlSimError("Offset operator in indirect "
+						"access is not of 32 bit width.")
+			if offsetOper.operType not in (AwlOperatorTypes.MEM_M,
+						       AwlOperatorTypes.MEM_L,
+						       AwlOperatorTypes.MEM_DB,
+						       AwlOperatorTypes.MEM_DI):
 				raise AwlSimError("Offset operator in indirect "
 					"access is not a valid memory offset.")
-			if offsetOper.width not in possibleWidths:
-				print(offsetOper.width)
-				raise AwlSimError("Offset operator in indirect "
-					"access is not of %s bit width." %\
-					listToHumanStr(possibleWidths))
-			offsetValue = self.insn.cpu.fetch(offsetOper,
-							  AwlOperatorWidths.WIDTH_MASK_8_16_32)
-			pointer = (self.area | (offsetValue & 0x0007FFFF))
+			offsetValue = AwlMemoryObject_asScalar(
+				self.insn.cpu.fetch(offsetOper,
+						    AwlOperatorWidths.WIDTH_MASK_8_16_32))
+			pointer = (self.area | (offsetValue & 0x0007FFFF)) #+suffix-u
 		else:
 			# Register-indirect access
 			if offsetOper.operType != AwlOperatorTypes.IMM_PTR:
 				raise AwlSimError("Offset operator in "
 					"register-indirect access is not a "
 					"pointer immediate.")
-			offsetValue = self.insn.cpu.fetch(offsetOper,
-							  AwlOperatorWidths.WIDTH_MASK_8_16_32) &\
-							  0x0007FFFF
+			offsetValue = AwlMemoryObject_asScalar(
+				self.insn.cpu.fetch(offsetOper,
+						    (AwlOperatorWidths.WIDTH_MASK_8_16_32 &
+						     0x0007FFFF))) #+suffix-u
 			if self.area == PointerConst.AREA_NONE_S:
 				# Area-spanning access
-				pointer = (self.insn.cpu.getAR(self.addressRegister).get() +\
-					   offsetValue) & 0xFFFFFFFF
+				pointer = ((self.insn.cpu.getAR(self.addressRegister).get() +
+					    offsetValue) & 0xFFFFFFFF) #+suffix-u
 			else:
 				# Area-internal access
-				pointer = ((self.insn.cpu.getAR(self.addressRegister).get() +
-					    offsetValue) & 0x0007FFFF) |\
-					  self.area
+				pointer = (((self.insn.cpu.getAR(self.addressRegister).get() +
+					     offsetValue) & 0x0007FFFF) | #+suffix-u
+					   self.area)
 		# Create a direct operator
-		try:
-			if store:
-				optype = AwlIndirectOpConst.area2optype_store[
-						pointer & AwlIndirectOpConst.EXT_AREA_MASK_S]
-			else:
-				optype = AwlIndirectOpConst.area2optype_fetch[
-						pointer & AwlIndirectOpConst.EXT_AREA_MASK_S]
-		except KeyError:
+		if store:
+			optype = AwlIndirectOpConst.area2optype(
+					(pointer & AwlIndirectOpConst.EXT_AREA_MASK_S),
+					True)
+		else:
+			optype = AwlIndirectOpConst.area2optype(
+					(pointer & AwlIndirectOpConst.EXT_AREA_MASK_S),
+					False)
+		if optype < 0: #+unlikely
 			raise AwlSimError("Invalid area code (%X hex) in indirect addressing" %\
 				((pointer & AwlIndirectOpConst.EXT_AREA_MASK_S) >>\
 				 PointerConst.AREA_SHIFT))
@@ -588,7 +585,7 @@ class AwlIndirectOp(AwlOperator): #+cdef
 			directOffset = make_AwlOffset_fromPointerValue(pointer)
 		else:
 			# 'pointer' is a byte offset
-			directOffset = make_AwlOffset(pointer & 0x0000FFFF, 0)
+			directOffset = make_AwlOffset(pointer & 0x0000FFFF, 0) #+suffix-u
 		if self.width != 1 and directOffset.bitOffset:
 			raise AwlSimError("Bit offset (lowest three bits) in %d-bit "
 				"indirect addressing is not zero. "
@@ -596,24 +593,24 @@ class AwlIndirectOp(AwlOperator): #+cdef
 				(self.width, str(directOffset)))
 		return make_AwlOperator(optype, self.width, directOffset, self.insn)
 
-	def __pointerError(self):
+	def __pointerError(self): #@nocov
 		# This is a programming error.
 		# The caller should resolve() the operator first.
 		raise AwlSimBug("Can not transform indirect operator "
 			"into a pointer. Resolve it first.")
 
 	def makePointer(self):
-		self.__pointerError()
+		self.__pointerError() #@nocov
 
 	def makePointerValue(self): #@nocy
-#@cy	cpdef uint32_t makePointerValue(self):
-		self.__pointerError()
+#@cy	cdef uint32_t makePointerValue(self) except? 0xFFFFFFFF:
+		self.__pointerError() #@nocov
 
 	def makeDBPointer(self):
-		self.__pointerError()
+		self.__pointerError() #@nocov
 
 	def makeANYPointer(self, areaShifted=None):
-		self.__pointerError()
+		self.__pointerError() #@nocov
 
 	def __repr__(self):
 		return "__INDIRECT" #TODO
@@ -631,7 +628,7 @@ class AwlIndirectOp(AwlOperator): #+cdef
 #                    AR_2 => This is a register-indirect access with AR2.
 # offsetOper -> This is the AwlOperator for the offset.
 #               For memory-indirect access, this must be an AwlOperator
-#               with "type in __possibleOffsetOperTypes".
+#               with type MEM_M, MEM_L, MEM_DB or MEM_DI.
 #               For register-indirect access, this must be an AwlOperator
 #               with "type==IMM_PTR".
 # insn -> The instruction this operator is used in. May be None.
