@@ -2,7 +2,7 @@
 #
 # AWL simulator - SFBs
 #
-# Copyright 2012-2014 Michael Buesch <m@bues.ch>
+# Copyright 2012-2017 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -20,18 +20,22 @@
 #
 
 from __future__ import division, absolute_import, print_function, unicode_literals
+#from awlsim.common.cython_support cimport * #@cy
 from awlsim.common.compat import *
 
-from awlsim.core.systemblocks.system_sfb_m1 import *
-from awlsim.core.systemblocks.system_sfb_0 import *
-from awlsim.core.systemblocks.system_sfb_1 import *
-from awlsim.core.systemblocks.system_sfb_2 import *
-from awlsim.core.systemblocks.system_sfb_3 import *
-from awlsim.core.systemblocks.system_sfb_4 import *
-from awlsim.core.systemblocks.system_sfb_5 import *
+from awlsim.core.systemblocks.system_sfb_m1 import * #+cimport
+from awlsim.core.systemblocks.system_sfb_0 import * #+cimport
+from awlsim.core.systemblocks.system_sfb_1 import * #+cimport
+from awlsim.core.systemblocks.system_sfb_2 import * #+cimport
+from awlsim.core.systemblocks.system_sfb_3 import * #+cimport
+from awlsim.core.systemblocks.system_sfb_4 import * #+cimport
+from awlsim.core.systemblocks.system_sfb_5 import * #+cimport
 
+__all__ = [
+	"_SFB_table",
+]
 
-SFB_table = {
+_SFB_table = { #+cdef-dict
 	-1	: SFBm1,	# __SFB_NOP
 	0	: SFB0,		# CTU
 	1	: SFB1,		# CTD
