@@ -2,7 +2,7 @@
 #
 # AWL simulator - SFCs
 #
-# Copyright 2012-2017 Michael Buesch <m@bues.ch>
+# Copyright 2012-2018 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -25,11 +25,11 @@ from awlsim.common.compat import *
 
 from awlsim.common.datatypehelpers import * #+cimport
 from awlsim.common.exceptions import *
+from awlsim.common.util import *
 
 from awlsim.core.systemblocks.systemblocks import * #+cimport
 from awlsim.core.blockinterface import *
 from awlsim.core.datatypes import *
-from awlsim.core.util import *
 
 
 class SFCm3(SFC): #+cdef
@@ -41,8 +41,9 @@ class SFCm3(SFC): #+cdef
 		),
 	}
 
-	def run(self): #+cpdef
-		shutdownType = wordToSignedPyInt(self.fetchInterfaceFieldByName("SHUTDOWN_TYPE"))
+	def run(self): #+cpdef #@nocov
+		shutdownType = wordToSignedPyInt(AwlMemoryObject_asScalar(
+			self.fetchInterfaceFieldByName("SHUTDOWN_TYPE")))
 		if shutdownType == 1:
 			raise MaintenanceRequest(MaintenanceRequest.TYPE_SHUTDOWN,
 						 "SFC -3 shutdown request")
