@@ -2,7 +2,7 @@
 #
 # AWL simulator - User defined data types (UDT)
 #
-# Copyright 2014-2017 Michael Buesch <m@bues.ch>
+# Copyright 2014-2018 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -23,13 +23,16 @@ from __future__ import division, absolute_import, print_function, unicode_litera
 #from awlsim.common.cython_support cimport * #@cy
 from awlsim.common.compat import *
 
+from awlsim.common.blockinfo import *
 from awlsim.common.enumeration import *
 from awlsim.common.exceptions import *
+from awlsim.common.util import *
 
-from awlsim.core.util import *
 from awlsim.core.blocks import * #+cimport
 from awlsim.core.datastructure import * #+cimport
 
+#from awlsim.core.cpu cimport * #@cy
+
 from awlsim.awlcompiler.translator import *
 
 
@@ -50,7 +53,7 @@ class UDTField(object): #+cdef
 		self.dataType = dataType
 		self.initBytes = initBytes
 
-	def __repr__(self):
+	def __repr__(self): #@nocov
 		if self.initBytes:
 			return "%s : %s := %s" %\
 				(self.name, str(self.dataType),
@@ -99,18 +102,20 @@ class UDT(Block): #+cdef
 		self.fields.append(field)
 		self.fieldNameMap[field.name] = field
 
-	def __buildField(self, cpu, field):
+	def __buildField(self, _cpu, field):
+#@cy		cdef S7CPU cpu
+
 		from awlsim.core.datatypes import AwlDataType
 
+		cpu = _cpu
 		if field.dataType.width < 0:
 			# The size of the field is unknown, yet.
 			# Try to resolve it.
 			if field.dataType.type == AwlDataType.TYPE_UDT_X:
 				# This UDT embeds another UDT.
 				# Get the embedded UDT and build it.
-				try:
-					udt = cpu.udts[field.dataType.index]
-				except KeyError:
+				udt = cpu.getUDT(field.dataType.index)
+				if not udt:
 					raise AwlSimError("The '%s' embeds "
 						"a 'UDT %d', which does not "
 						"exist." %\
@@ -155,5 +160,12 @@ class UDT(Block): #+cdef
 				"one variable in '%s'" % str(self))
 		self.__structState = self.STRUCT_BUILT
 
-	def __repr__(self):
+	def getBlockInfo(self):
+		"""Get a BlockInfo instance for this block.
+		"""
+		return BlockInfo(blockType=BlockInfo.TYPE_UDT,
+				 blockIndex=self.index,
+				 identHash=self.identHash)
+
+	def __repr__(self): #@nocov
 		return "UDT %d" % self.index
