@@ -2,7 +2,7 @@
 #
 # AWL simulator - GUI CPU widget
 #
-# Copyright 2012-2018 Michael Buesch <m@bues.ch>
+# Copyright 2012-2020 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -28,210 +28,11 @@ from awlsim.gui.cpustate import *
 from awlsim.gui.awlsimclient import *
 from awlsim.gui.icons import *
 from awlsim.gui.linkconfig import *
+from awlsim.gui.runstate import *
+from awlsim.gui.validatorsched import *
 
 
-class RunState(QObject):
-	# Signal: Emitted, if the state changed.
-	# The parameter is 'self'.
-	stateChanged = Signal(QObject)
-
-	EnumGen.start
-	STATE_OFFLINE	= EnumGen.item
-	STATE_ONLINE	= EnumGen.item
-	STATE_LOAD	= EnumGen.item
-	STATE_RUN	= EnumGen.item
-	STATE_EXCEPTION	= EnumGen.item
-	EnumGen.end
-
-	def __init__(self):
-		QObject.__init__(self)
-		self.state = self.STATE_OFFLINE
-		self.setCoreDetails()
-
-	def __emitStateChanged(self):
-		self.stateChanged.emit(self)
-		QApplication.processEvents(QEventLoop.ExcludeUserInputEvents,
-					   50)
-
-	def setState(self, newState):
-		if self.state != newState:
-			self.state = newState
-			self.__emitStateChanged()
-
-	def setCoreDetails(self, spawned=True,
-			   host=None, port=None,
-			   haveTunnel=False):
-		self.spawned = spawned
-		self.host = host
-		self.port = port
-		self.haveTunnel = haveTunnel
-		self.__emitStateChanged()
-
-	def __eq__(self, other):
-		if isinstance(self, RunState) and\
-		   isinstance(other, RunState):
-			return self.state == other.state
-		if isinstance(self, RunState) and\
-		   isInteger(other):
-			return self.state == other
-		raise RuntimeError
-
-	def __ne__(self, other):
-		return not self.__eq__(other)
-
-class OnlineSelectAction(QAction):
-	def __init__(self, parent):
-		QAction.__init__(self, getIcon("network"), "", parent)
-
-		self.setCheckable(True)
-		self.__handleToggle(self.isChecked())
-
-		self.toggled.connect(self.__handleToggle)
-
-	def __handleToggle(self, checked):
-		if checked:
-			self.setText("Go offline")
-		else:
-			self.setText("Go online (Connect to a CPU)")
-
-class RunSelectAction(QAction):
-	def __init__(self, parent):
-		QAction.__init__(self, getIcon("run"), "", parent)
-
-		self.setCheckable(True)
-		self.__handleToggle(self.isChecked())
-
-		self.toggled.connect(self.__handleToggle)
-
-	def __handleToggle(self, checked):
-		if checked:
-			self.setText("Stop CPU (STOP mode)")
-			self.setIcon(getIcon("stop"))
-		else:
-			self.setText("Start CPU (RUN mode)")
-			self.setIcon(getIcon("run"))
-
-class DiagSelectAction(QAction):
-	def __init__(self, parent):
-		QAction.__init__(self, getIcon("glasses"), "", parent)
-
-		self.setCheckable(True)
-		self.__handleToggle(self.isChecked())
-
-		self.toggled.connect(self.__handleToggle)
-
-	def __handleToggle(self, checked):
-		if checked:
-			self.setText("Disable online diagnosis")
-		else:
-			self.setText("Enable online diagnosis")
-
-class CpuInspectToolBar(QToolBar):
-	def __init__(self, parent=None):
-		QToolBar.__init__(self, parent)
-		self.setObjectName("CpuInspectToolBar")
-		self.setWindowTitle("CPU inspection tool bar")
-
-		self.blocksAction = QAction(getIcon("plugin"),
-					    "Add inspection: Online blocks",
-					    self)
-		self.addAction(self.blocksAction)
-		self.inputsAction = QAction(getIcon("inputs"),
-					    "Add inspection: Input memory (I / E)",
-					    self)
-		self.addAction(self.inputsAction)
-		self.outputsAction = QAction(getIcon("outputs"),
-					     "Add inspection: Output memory (Q / A)",
-					     self)
-		self.addAction(self.outputsAction)
-		self.flagsAction = QAction(getIcon("flags"),
-					   "Add inspection: Flag memory (M)",
-					   self)
-		self.addAction(self.flagsAction)
-		self.dbAction = QAction(getIcon("datablock"),
-					"Add inspection: Data block (DB)",
-					self)
-		self.addAction(self.dbAction)
-		self.timerAction = QAction(getIcon("timer"),
-					   "Add inspection: Timer (T)",
-					   self)
-		self.addAction(self.timerAction)
-		self.counterAction = QAction(getIcon("counter"),
-					     "Add inspection: Counter (C / Z)",
-					     self)
-		self.addAction(self.counterAction)
-		self.cpuAction = QAction(getIcon("cpu"),
-					 "Add inspection: CPU overview",
-					 self)
-		self.addAction(self.cpuAction)
-		self.lcdAction = QAction(getIcon("lcd"),
-					 "Add inspection: LCD",
-					 self)
-		self.addAction(self.lcdAction)
-
-	def connectToCpuWidget(self, cpuWidget):
-		self.blocksAction.triggered.connect(cpuWidget.newWin_Blocks)
-		self.inputsAction.triggered.connect(cpuWidget.newWin_E)
-		self.outputsAction.triggered.connect(cpuWidget.newWin_A)
-		self.flagsAction.triggered.connect(cpuWidget.newWin_M)
-		self.dbAction.triggered.connect(cpuWidget.newWin_DB)
-		self.timerAction.triggered.connect(cpuWidget.newWin_T)
-		self.counterAction.triggered.connect(cpuWidget.newWin_Z)
-		self.cpuAction.triggered.connect(cpuWidget.newWin_CPU)
-		self.lcdAction.triggered.connect(cpuWidget.newWin_LCD)
-
-class CpuControlToolBar(QToolBar):
-	def __init__(self, parent=None):
-		QToolBar.__init__(self, parent)
-		self.setObjectName("CpuControlToolBar")
-		self.setWindowTitle("CPU control tool bar")
-
-		self.onlineAction = OnlineSelectAction(self)
-		self.addAction(self.onlineAction)
-		self.resetAction = QAction(getIcon("doc_delete"),
-					   "Reset the CPU",
-					   self)
-		self.addAction(self.resetAction)
-		self.downloadAction = QAction(getIcon("download"),
-					      "Download all sources to CPU",
-					      self)
-		self.addAction(self.downloadAction)
-		self.downloadSingleAction = QAction(getIcon("download_one"),
-						    "Download single source to CPU",
-						    self)
-		self.addAction(self.downloadSingleAction)
-		self.runAction = RunSelectAction(self)
-		self.addAction(self.runAction)
-		self.diagAction = DiagSelectAction(self)
-		self.addAction(self.diagAction)
-
-	def connectToCpuWidget(self, cpuWidget):
-		self.onlineAction.toggled.connect(cpuWidget._onlineToggled)
-		self.resetAction.triggered.connect(cpuWidget.resetCpu)
-		self.downloadAction.triggered.connect(cpuWidget.download)
-		self.downloadSingleAction.triggered.connect(cpuWidget.downloadSingle)
-		self.runAction.toggled.connect(cpuWidget._runStateToggled)
-		self.diagAction.toggled.connect(cpuWidget._onlineDiagToggled)
-
-		cpuWidget.reqRunButtonState.connect(self.__setRun)
-		cpuWidget.reqOnlineButtonState.connect(self.__setOnline)
-		cpuWidget.reqOnlineDiagButtonState.connect(self.__setOnlineDiag)
-
-	def __setRun(self, en):
-		if en != self.runAction.isChecked():
-			self.runAction.trigger()
-
-	def __setOnline(self, en):
-		if en != self.onlineAction.isChecked():
-			self.onlineAction.trigger()
-
-	def __setOnlineDiag(self, en):
-		if en != self.diagAction.isChecked():
-			self.diagAction.trigger()
-
 class CpuWidget(QWidget):
-	# Signal: The CPU run-state changed
-	runStateChanged = Signal(RunState)
 	# Signal: The online-diag state changed
 	onlineDiagChanged = Signal(bool)
 	# Signal: Some configuration value changed
@@ -240,6 +41,9 @@ class CpuWidget(QWidget):
 	haveInsnDump = Signal(AwlSimMessage_INSNSTATE)
 	# Signal: Have a new ident hashes message
 	haveIdentsMsg = Signal(AwlSimMessage_IDENTS)
+	# CPU-stats signal.
+	# Parameter: AwlSimMessage_CPUSTATE instance.
+	haveCpuStats = Signal(AwlSimMessage_CPUSTATS)
 
 	# Signal: Request a new run button state.
 	reqRunButtonState = Signal(bool)
@@ -254,12 +58,14 @@ class CpuWidget(QWidget):
 		self.layout().setContentsMargins(QMargins(7, 0, 0, 0))
 
 		self.mainWidget = mainWidget
-		self.state = RunState()
 		self.__runStateChangeBlocked = Blocker()
 
 		self.__onlineBtnPressed = False
 		self.__runBtnPressed = False
 		self.__onlineDiagBtnPressed = False
+		self.__onlineDiagIdentHash = None
+		self.__onlineDiagFromLine = None
+		self.__onlineDiagToLine = None
 
 		self.__coreMsgTimer = QTimer(self)
 		self.__coreMsgTimer.setSingleShot(False)
@@ -272,20 +78,24 @@ class CpuWidget(QWidget):
 		client = self.getSimClient()
 		client.haveException.connect(self.__handleCpuException)
 		client.haveCpuDump.connect(self.__handleCpuDump)
+		client.haveCpuStats.connect(self.haveCpuStats)
 		client.haveInsnDump.connect(self.haveInsnDump)
 		client.haveMemoryUpdate.connect(self.__handleMemoryUpdate)
 		client.haveIdentsMsg.connect(self.__handleIdentsMsg)
 
-		self.stateMdi = StateMdiArea(self)
+		self.stateMdi = StateMdiArea(client=client, parent=self)
 		self.stateMdi.setViewMode(QMdiArea.SubWindowView)
 		self.stateMdi.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
 		self.stateMdi.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
 		self.layout().addWidget(self.stateMdi, 0, 0)
 
-		self.state.stateChanged.connect(self.runStateChanged)
-
-		self.newWin_Blocks()
-		self.newWin_CPU()
+		self.stateMdi.subWinAdded.connect(lambda w: self.__uploadMemReadAreas())
+		self.stateMdi.subWinClosed.connect(lambda w: self.__stateMdiWindowClosed(w))
+		self.stateMdi.settingsChanged.connect(lambda: self.__uploadMemReadAreas())
+		self.stateMdi.contentChanged.connect(
+			lambda: self.mainWidget.setDirty(self.mainWidget.DIRTY_SLIGHT))
+		self.stateMdi.openByIdentHash.connect(
+			lambda mdiWin, identHash: self.mainWidget.openByIdentHash(identHash))
 
 	def getProject(self):
 		return self.mainWidget.getProject()
@@ -293,59 +103,51 @@ class CpuWidget(QWidget):
 	def getSimClient(self):
 		return self.mainWidget.getSimClient()
 
+	def getEditMdiArea(self):
+		return self.mainWidget.editMdiArea
+
 	def __stateMdiWindowClosed(self, mdiWin):
 		QTimer.singleShot(0, self.__uploadMemReadAreas)
 
-	def __addWindow(self, win):
-		mdiWin = StateMdiSubWindow(win)
-		mdiWin.closed.connect(self.__stateMdiWindowClosed)
-		self.stateMdi.addSubWindow(mdiWin, Qt.Window)
-		win.configChanged.connect(self.__stateWinConfigChanged)
-		win.show()
-		self.__uploadMemReadAreas()
-
 	def newWin_CPU(self):
-		self.__addWindow(State_CPU(self.getSimClient(), self))
+		win = State_CPU(self.getSimClient())
+		self.stateMdi.addCpuStateWindow(win)
 
 	def newWin_DB(self):
-		self.__addWindow(State_Mem(self.getSimClient(),
-					   AbstractDisplayWidget.ADDRSPACE_DB,
-					   self))
+		win = State_Mem(self.getSimClient(),
+				AbstractDisplayWidget.ADDRSPACE_DB)
+		self.stateMdi.addCpuStateWindow(win)
 
 	def newWin_E(self):
-		self.__addWindow(State_Mem(self.getSimClient(),
-					   AbstractDisplayWidget.ADDRSPACE_E,
-					   self))
+		win = State_Mem(self.getSimClient(),
+				AbstractDisplayWidget.ADDRSPACE_E)
+		self.stateMdi.addCpuStateWindow(win)
 
 	def newWin_A(self):
-		self.__addWindow(State_Mem(self.getSimClient(),
-					   AbstractDisplayWidget.ADDRSPACE_A,
-					   self))
+		win = State_Mem(self.getSimClient(),
+				AbstractDisplayWidget.ADDRSPACE_A)
+		self.stateMdi.addCpuStateWindow(win)
 
 	def newWin_M(self):
-		self.__addWindow(State_Mem(self.getSimClient(),
-					   AbstractDisplayWidget.ADDRSPACE_M,
-					   self))
+		win = State_Mem(self.getSimClient(),
+				AbstractDisplayWidget.ADDRSPACE_M)
+		self.stateMdi.addCpuStateWindow(win)
 
 	def newWin_T(self):
-		self.__addWindow(State_Timer(self.getSimClient(),
-					     self))
+		win = State_Timer(self.getSimClient())
+		self.stateMdi.addCpuStateWindow(win)
 
 	def newWin_Z(self):
-		self.__addWindow(State_Counter(self.getSimClient(),
-					       self))
+		win = State_Counter(self.getSimClient())
+		self.stateMdi.addCpuStateWindow(win)
 
 	def newWin_LCD(self):
-		self.__addWindow(State_LCD(self.getSimClient(), self))
+		win = State_LCD(self.getSimClient())
+		self.stateMdi.addCpuStateWindow(win)
 
 	def newWin_Blocks(self):
-		mdiWin = State_Blocks(self.getSimClient(), self)
-		self.__addWindow(mdiWin)
-		mdiWin.openItem.connect(
-			lambda identHash: self.mainWidget.openByIdentHash(identHash))
-
-	def __stateWinConfigChanged(self, stateWin):
-		self.__uploadMemReadAreas()
+		win = State_Blocks(self.getSimClient())
+		self.stateMdi.addCpuStateWindow(win)
 
 	# Upload the used memory area descriptors to the core.
 	def __uploadMemReadAreas(self):
@@ -363,7 +165,8 @@ class CpuWidget(QWidget):
 						     sync = True)
 			client.setPeriodicDumpInterval(300 if wantDump else 0)
 		except AwlSimError as e:
-			self.state.setState(RunState.STATE_EXCEPTION)
+			with MessageBox.awlSimErrorBlocked:
+				client.runState.setState(RunState.STATE_EXCEPTION)
 			MessageBox.handleAwlSimError(self,
 				"Error in awlsim core", e)
 			return False
@@ -377,8 +180,10 @@ class CpuWidget(QWidget):
 		# Set our state to exception/stopped.
 		# This will stop the CPU, if it wasn't already stopped.
 		# Subsequent exception handlers might do additional steps.
-		self.state.setState(RunState.STATE_EXCEPTION)
-		self.stop()
+		with MessageBox.awlSimErrorBlocked:
+			client = self.getSimClient()
+			client.runState.setState(RunState.STATE_EXCEPTION)
+			self.stop()
 
 	def __handleCpuDump(self, dumpText):
 		for mdiWin in self.stateMdi.subWindowList():
@@ -447,12 +252,14 @@ class CpuWidget(QWidget):
 			self.__startCoreMessageHandler()
 
 			# Put the GUI into RUN mode.
-			self.state.setState(RunState.STATE_RUN)
+			client.runState.setState(RunState.STATE_RUN)
 		except AwlSimError as e:
-			self.state.setState(RunState.STATE_EXCEPTION)
+			with MessageBox.awlSimErrorBlocked:
+				client.runState.setState(RunState.STATE_EXCEPTION)
 			MessageBox.handleAwlSimError(self,
 				"Could not start CPU", e)
-			self.stop()
+			with MessageBox.awlSimErrorBlocked:
+				self.stop()
 		except MaintenanceRequest as e:
 			self.__handleMaintenance(e)
 
@@ -464,11 +271,13 @@ class CpuWidget(QWidget):
 			while client.processMessages(0.02):
 				pass
 		except AwlSimError as e:
-			self.state.setState(RunState.STATE_EXCEPTION)
+			with MessageBox.awlSimErrorBlocked:
+				client.runState.setState(RunState.STATE_EXCEPTION)
 			MessageBox.handleAwlSimError(self,
 				"Core server error", e)
-			self.stop()
-			self.__stopCoreMessageHandler()
+			with MessageBox.awlSimErrorBlocked:
+				self.stop()
+				self.__stopCoreMessageHandler()
 		except MaintenanceRequest as e:
 			self.__handleMaintenance(e)
 		except Exception:
@@ -515,13 +324,17 @@ class CpuWidget(QWidget):
 				client.requestBlockInfo(reqOBInfo=True,
 							reqFCInfo=True,
 							reqFBInfo=True,
-							reqDBInfo=True)
+							reqDBInfo=True,
+							reqUDTInfo=True)
+			client.getCpuStats()
 		except AwlSimError as e:
-			self.state.setState(RunState.STATE_EXCEPTION)
+			with MessageBox.awlSimErrorBlocked:
+				client.runState.setState(RunState.STATE_EXCEPTION)
 			MessageBox.handleAwlSimError(self,
 				"Core server error", e)
-			self.stop()
-			self.__stopCoreMessageHandler()
+			with MessageBox.awlSimErrorBlocked:
+				self.stop()
+				self.__stopCoreMessageHandler()
 		except MaintenanceRequest as e:
 			self.__handleMaintenance(e)
 		except Exception:
@@ -534,12 +347,13 @@ class CpuWidget(QWidget):
 		self.haveIdentsMsg.emit(identsMsg)
 
 	def __stop(self):
+		client = self.getSimClient()
+
 		# Make sure the button is released.
 		with self.__runStateChangeBlocked:
 			self.stop()
 
 		if self.isOnline():
-			client = self.getSimClient()
 			try:
 				client.setRunState(False)
 			except AwlSimError as e:
@@ -549,7 +363,7 @@ class CpuWidget(QWidget):
 		# Re-Start the message handler.
 		self.__startCoreMessageHandler()
 
-		self.state.setState(RunState.STATE_ONLINE)
+		client.runState.setState(RunState.STATE_ONLINE)
 
 	def stop(self):
 		self.reqRunButtonState.emit(False)
@@ -581,12 +395,12 @@ class CpuWidget(QWidget):
 			else:
 				client.setMode_ONLINE(self, linkConfig)
 
-			self.state.setCoreDetails(
+			client.runState.setCoreDetails(
 				spawned = linkConfig.getSpawnLocalEn(),
 				host = linkConfig.getConnectHost(),
 				port = linkConfig.getConnectPort(),
 				haveTunnel = (linkConfig.getTunnel() == linkConfig.TUNNEL_SSH))
-			self.state.setState(RunState.STATE_ONLINE)
+			client.runState.setState(RunState.STATE_ONLINE)
 
 			if client.getRunState():
 				# The core is already running.
@@ -603,7 +417,8 @@ class CpuWidget(QWidget):
 				client.setMode_OFFLINE()
 			MessageBox.handleAwlSimError(self,
 				"Error while trying to connect to CPU", e)
-			self.goOffline()
+			with MessageBox.awlSimErrorBlocked:
+				self.goOffline()
 			return
 		except MaintenanceRequest as e:
 			self.__handleMaintenance(e)
@@ -618,7 +433,7 @@ class CpuWidget(QWidget):
 		# Release the stop-button.
 		# This will _not_ stop the CPU, as we're offline already.
 		self.stop()
-		self.state.setState(RunState.STATE_OFFLINE)
+		client.runState.setState(RunState.STATE_OFFLINE)
 		self.__stopCoreMessageHandler()
 
 	def _onlineToggled(self, onlineBtnPressed):
@@ -637,6 +452,30 @@ class CpuWidget(QWidget):
 	def goOffline(self):
 		self.reqOnlineButtonState.emit(False)
 
+	def __validatePreDownload(self, project):
+		guiSettings = project.getGuiSettings()
+		if not guiSettings.getPreDownloadValidationEn():
+			return True
+
+		printInfo("Validating project before downloading...")
+		valSched = GuiValidatorSched.get()
+		exception = valSched.syncValidation(project)
+		if exception is valSched.TIMEOUT:
+			printError("Project validation failed. Loading anyway...")
+		elif exception is not None:
+			res = MessageBox.handleAwlSimError(self,
+				"\nPre-download validation of the project failed.\n"
+				"Continuing the download may bring the CPU to STOP.\n\n"
+				"Do you want to continue or cancel "
+				"downloading the project to the CPU?",
+				exception,
+				okButton=False,
+				continueButton=True,
+				cancelButton=True)
+			if res != MessageBox.Accepted:
+				return False
+		return True
+
 	# Reset/clear the CPU and upload all sources.
 	def download(self):
 		# Make sure we are online.
@@ -647,7 +486,14 @@ class CpuWidget(QWidget):
 		client = self.getSimClient()
 		project = self.getProject()
 		try:
-			self.state.setState(RunState.STATE_LOAD)
+			client.runState.setState(RunState.STATE_LOAD)
+
+			if not self.__validatePreDownload(project):
+				if self.__runBtnPressed:
+					client.runState.setState(RunState.STATE_RUN)
+				else:
+					client.runState.setState(RunState.STATE_ONLINE)
+				return False
 
 			client.setRunState(False)
 			client.reset()
@@ -655,15 +501,17 @@ class CpuWidget(QWidget):
 			client.loadProject(project)
 			client.build()
 
-			self.state.setState(RunState.STATE_ONLINE)
+			client.runState.setState(RunState.STATE_ONLINE)
 		except AwlParserError as e:
-			self.state.setState(RunState.STATE_ONLINE)
-			self.stop()
+			with MessageBox.awlSimErrorBlocked:
+				client.runState.setState(RunState.STATE_ONLINE)
+				self.stop()
 			MessageBox.handleAwlParserError(self, e)
 			return False
 		except AwlSimError as e:
-			self.state.setState(RunState.STATE_ONLINE)
-			self.stop()
+			with MessageBox.awlSimErrorBlocked:
+				client.runState.setState(RunState.STATE_ONLINE)
+				self.stop()
 			MessageBox.handleAwlSimError(self,
 				"Error while loading code", e)
 			return False
@@ -683,8 +531,7 @@ class CpuWidget(QWidget):
 
 	# Download the current source.
 	def downloadSingle(self):
-		editMdiArea = self.mainWidget.editMdiArea
-		mdiSubWin = editMdiArea.activeOpenSubWindow
+		mdiSubWin = self.getEditMdiArea().activeOpenSubWindow
 		source = libSelections = None
 		if mdiSubWin:
 			if mdiSubWin.TYPE in {mdiSubWin.TYPE_AWL,
@@ -712,7 +559,14 @@ class CpuWidget(QWidget):
 		client = self.getSimClient()
 		project = self.getProject()
 		try:
-			self.state.setState(RunState.STATE_LOAD)
+			client.runState.setState(RunState.STATE_LOAD)
+
+			if not self.__validatePreDownload(project):
+				if self.__runBtnPressed:
+					client.runState.setState(RunState.STATE_RUN)
+				else:
+					client.runState.setState(RunState.STATE_ONLINE)
+				return False
 
 			if mdiSubWin.TYPE == mdiSubWin.TYPE_AWL:
 				printVerbose("Single AWL download: %s/%s" %\
@@ -741,17 +595,19 @@ class CpuWidget(QWidget):
 				assert(0)
 
 			if self.__runBtnPressed:
-				self.state.setState(RunState.STATE_RUN)
+				client.runState.setState(RunState.STATE_RUN)
 			else:
-				self.state.setState(RunState.STATE_ONLINE)
+				client.runState.setState(RunState.STATE_ONLINE)
 		except AwlParserError as e:
-			self.state.setState(RunState.STATE_ONLINE)
-			self.stop()
+			with MessageBox.awlSimErrorBlocked:
+				client.runState.setState(RunState.STATE_ONLINE)
+				self.stop()
 			MessageBox.handleAwlParserError(self, e)
 			return False
 		except AwlSimError as e:
-			self.state.setState(RunState.STATE_ONLINE)
-			self.stop()
+			with MessageBox.awlSimErrorBlocked:
+				client.runState.setState(RunState.STATE_ONLINE)
+				self.stop()
 			MessageBox.handleAwlSimError(self,
 				"Error while loading code (single source)", e)
 			return False
@@ -800,8 +656,8 @@ class CpuWidget(QWidget):
 			else:
 				self.__stop()
 
-	def handleDirtyChange(self, dirty):
-		if dirty:
+	def handleDirtyChange(self, dirtyLevel):
+		if dirtyLevel == self.mainWidget.DIRTY_FULL:
 			self.reqOnlineDiagButtonState.emit(False)
 
 	def _onlineDiagToggled(self, onlineDiagBtnPressed):
@@ -815,12 +671,25 @@ class CpuWidget(QWidget):
 		try:
 			client = self.getSimClient()
 			if self.__onlineDiagBtnPressed and source:
-				client.setInsnStateDump(enable=True,
-							sourceId=source.identHash,
-							fromLine=fromLine, toLine=toLine,
-							sync=False)
+				identHash = source.identHash
+				if (self.__onlineDiagIdentHash != identHash or
+				    self.__onlineDiagFromLine != fromLine or
+				    self.__onlineDiagToLine != toLine):
+					client.setInsnStateDump(enable=True,
+								sourceId=identHash,
+								fromLine=fromLine, toLine=toLine,
+								userData=42,
+								ob1Div=1,
+								sync=False)
+					self.__onlineDiagIdentHash = identHash
+					self.__onlineDiagFromLine = fromLine
+					self.__onlineDiagToLine = toLine
 			else:
-				client.setInsnStateDump(enable=False, sync=False)
+				if self.__onlineDiagIdentHash:
+					client.setInsnStateDump(enable=False, sync=False)
+					self.__onlineDiagIdentHash = None
+					self.__onlineDiagFromLine = None
+					self.__onlineDiagToLine = None
 		except AwlSimError as e:
 			MessageBox.handleAwlSimError(self,
 				"Failed to setup instruction dumping", e)
