@@ -23,8 +23,6 @@ from __future__ import division, absolute_import, print_function, unicode_litera
 #from awlsim.common.cython_support cimport * #@cy
 from awlsim.common.compat import *
 
-from awlsim.common.codevalidator import *
-
 from awlsim.gui.editwidget import EditWidget
 from awlsim.gui.symtabwidget import SymTabView
 from awlsim.gui.libtablewidget import LibTableView
@@ -32,6 +30,8 @@ from awlsim.gui.fup.fupwidget import FupWidget
 from awlsim.gui.sourcecodeedit import SourceCodeEdit
 from awlsim.gui.icons import *
 from awlsim.gui.util import *
+from awlsim.gui.runstate import *
+from awlsim.gui.validatorsched import *
 
 
 __all__ = [
@@ -72,6 +72,11 @@ class EditMdiArea(QMdiArea):
 		QMdiArea.__init__(self, parent=mainWidget)
 		self.mainWidget = mainWidget
 
+		self.__onlineDiagMdiSubWin = None
+
+		GuiValidatorSched.get().haveValidationResult.connect(
+			self.__handleDocumentValidationResult)
+
 		# Init the editor find dialog.
 		SourceCodeEdit.initFindDialog(self)
 
@@ -83,6 +88,7 @@ class EditMdiArea(QMdiArea):
 		self.resetArea()
 
 		self.subWindowActivated.connect(self.__handleSubWindowActivated)
+		self.getSimClient().haveException.connect(self.__handleAwlSimError)
 
 	def getMainWidget(self):
 		return self.mainWidget
@@ -93,6 +99,9 @@ class EditMdiArea(QMdiArea):
 	def getProject(self):
 		return self.getProjectTreeModel().getProject()
 
+	def getSimClient(self):
+		return self.getMainWidget().getSimClient()
+
 	@property
 	def activeOpenSubWindow(self):
 		"""Get the currently active and open sub window.
@@ -114,6 +123,10 @@ class EditMdiArea(QMdiArea):
 			del mdiSubWin
 		self._guiSettings = GuiSettings()
 
+	def __handleAwlSimError(self, e):
+		for mdiSubWin in self.subWindowList():
+			mdiSubWin.handleAwlSimError(e)
+
 	def __handleSubWindowActivated(self, mdiSubWin):
 		"""An MDI sub window has just been activated.
 		"""
@@ -122,7 +135,8 @@ class EditMdiArea(QMdiArea):
 		self.__handleSubWinCopyAvailChanged(mdiSubWin)
 		self.__handleSubWinCutAvailChanged(mdiSubWin)
 		self.__handleSubWinPasteAvailChanged(mdiSubWin)
-		self.__handleDocumentValidation()
+		GuiValidatorSched.get().startAsyncValidation(self.getProject,
+							     delaySec=0.5)
 		self.__refreshOnlineDiagState()
 		self.__setFindDialogReference(mdiSubWin)
 
@@ -181,11 +195,14 @@ class EditMdiArea(QMdiArea):
 		mdiSubWin.pasteAvailableChanged.connect(
 			lambda pasteAvail: self.__handleSubWinPasteAvailChanged(mdiSubWin))
 		mdiSubWin.resizeFont.connect(self.__handleSourceCodeFontResize)
-		mdiSubWin.validateDocument.connect(self.__handleDocumentValidation)
+		mdiSubWin.validateDocument.connect(
+			lambda delaySec: GuiValidatorSched.get().startAsyncValidation(project=self.getProject,
+										      delaySec=delaySec))
 		mdiSubWin.visibleLinesChanged.connect(self.__handleVisibleLinesChanged)
 
 		self.__handleSubWinFocusChanged(mdiSubWin, True)
-		QTimer.singleShot(0, self.__handleDocumentValidation)
+		GuiValidatorSched.get().startAsyncValidation(project=self.getProject,
+							     delaySec=0.5)
 		mdiSubWin.setCpuRunState(self.__cpuRunState)
 
 		return mdiSubWin
@@ -228,25 +245,6 @@ class EditMdiArea(QMdiArea):
 
 		self.sourceChanged.emit()
 
-	def __handleDocumentValidation(self):
-		"""Handle a background document validation request.
-		"""
-		validator = AwlValidator.get()
-		if validator:
-			project = self.getProject()
-			validator.validate(project=project)
-			QTimer.singleShot(100, self.__checkDocumentValidationResult)
-
-	def __checkDocumentValidationResult(self):
-		"""Poll the background source code validation result.
-		"""
-		validator = AwlValidator.get()
-		if validator:
-			running, exception = validator.getState()
-			self.__handleDocumentValidationResult(exception)
-			if running:
-				QTimer.singleShot(100, self.__checkDocumentValidationResult)
-
 	def __handleDocumentValidationResult(self, exception):
 		"""Handle a background source code validator exception.
 		"""
@@ -267,12 +265,14 @@ class EditMdiArea(QMdiArea):
 
 		# Enable the active sub window, if possible.
 		if activeMdiSubWin and activeMdiSubWin.ONLINE_DIAG:
+			self.__onlineDiagMdiSubWin = activeMdiSubWin
 			if self.__onlineDiagEnabled:
 				activeMdiSubWin.enableOnlineDiag(True)
 			source = activeMdiSubWin.getSource()
 			fromLine, toLine = activeMdiSubWin.getVisibleLineRange()
 			self.visibleLinesChanged.emit(source, fromLine, toLine)
 		else:
+			self.__onlineDiagMdiSubWin = None
 			self.visibleLinesChanged.emit(None, -1, -1)
 
 	def __handleVisibleLinesChanged(self, mdiSubWin, source, visibleFromLine, visibleToLine):
@@ -285,7 +285,7 @@ class EditMdiArea(QMdiArea):
 		self.__refreshOnlineDiagState()
 
 	def handleInsnDump(self, insnDumpMsg):
-		mdiSubWin = self.activeOpenSubWindow
+		mdiSubWin = self.__onlineDiagMdiSubWin
 		if mdiSubWin:
 			mdiSubWin.handleInsnDump(insnDumpMsg)
 
@@ -300,6 +300,10 @@ class EditMdiArea(QMdiArea):
 		self.__cpuRunState = runState
 		for mdiSubWin in self.subWindowList():
 			mdiSubWin.setCpuRunState(runState)
+			if runState in (RunState.STATE_LOAD,
+					RunState.STATE_RUN):
+				# Clear all errors.
+				mdiSubWin.handleAwlSimError(None)
 
 	def undoIsAvailable(self):
 		mdiSubWin = self.activeOpenSubWindow
@@ -418,7 +422,8 @@ class EditMdiSubWindow(QMdiSubWindow):
 
 	# Signal: Validation request.
 	#	  A code validation should take place.
-	validateDocument = Signal()
+	#	  Parameter: Number of seconds to delay before validation start.
+	validateDocument = Signal(float)
 
 	# Signal: The visible AWL line range changed
 	#         Parameters are: EditMdiSubWindow, AwlSource, visibleFromLine, visibleToLine
@@ -460,6 +465,13 @@ class EditMdiSubWindow(QMdiSubWindow):
 		self.closed.emit(self)
 		QMdiSubWindow.closeEvent(self, ev)
 
+	def wheelEvent(self, ev):
+		QMdiSubWindow.wheelEvent(self, ev)
+		# Always accept the wheel event.
+		# This avoids forwarding it to the parent MDI area,
+		# if the scroll happened in this MDI sub window.
+		ev.accept()
+
 	def updateTitle(self):
 		pass
 
@@ -527,9 +539,6 @@ class EditMdiSubWindow(QMdiSubWindow):
 	def setGuiSettings(self, guiSettings):
 		pass
 
-	def handleDocumentValidationResult(self, exception):
-		pass
-
 	def enableOnlineDiag(self, enabled):
 		return False
 
@@ -542,6 +551,12 @@ class EditMdiSubWindow(QMdiSubWindow):
 	def handleIdentsMsg(self, identsMsg):
 		pass
 
+	def handleAwlSimError(self, exception):
+		pass
+
+	def handleDocumentValidationResult(self, exception):
+		self.handleAwlSimError(exception)
+
 	def setCpuRunState(self, runState):
 		pass
 
@@ -566,13 +581,22 @@ class AwlEditMdiSubWindow(EditMdiSubWindow):
 		self.editWidget.copyAvailable.connect(self.cutAvailableChanged)
 		self.editWidget.resizeFont.connect(self.resizeFont)
 		self.editWidget.validateDocument.connect(
-			lambda editWidget: self.validateDocument.emit())
+			lambda editWidget: self.validateDocument.emit(0.0))
+
+		self.windowStateChanged.connect(self.__handleWindowStateChange)
 
 		self.updateTitle()
 
 	def sizeHint(self):
 		return QSize(600, 500)
 
+	def resizeEvent(self, ev):
+		EditMdiSubWindow.resizeEvent(self, ev)
+		self.__emitVisibleLinesSignal()
+
+	def __handleWindowStateChange(self, oldState, newState):
+		self.__emitVisibleLinesSignal()
+
 	def updateTitle(self):
 		title = ""
 		source = self.editWidget.getSource()
@@ -676,7 +700,7 @@ class AwlEditMdiSubWindow(EditMdiSubWindow):
 	def setGuiSettings(self, guiSettings):
 		self.editWidget.setSettings(guiSettings)
 
-	def handleDocumentValidationResult(self, exception):
+	def handleAwlSimError(self, exception):
 		self.editWidget.handleValidationResult(exception)
 
 	def enableOnlineDiag(self, enabled):
@@ -721,8 +745,12 @@ class FupEditMdiSubWindow(EditMdiSubWindow):
 		self.setWidget(self.fupWidget)
 
 		self.fupWidget.diagramChanged.connect(self.sourceChanged)
+		self.fupWidget.diagramChanged.connect(self.__handleDiagramChanged)
 		self.fupWidget.undoAvailableChanged.connect(self.undoAvailableChanged)
 		self.fupWidget.redoAvailableChanged.connect(self.redoAvailableChanged)
+		self.fupWidget.clipboardCopyAvailableChanged.connect(self.copyAvailableChanged)
+		self.fupWidget.clipboardCutAvailableChanged.connect(self.cutAvailableChanged)
+		self.fupWidget.clipboardPasteAvailableChanged.connect(self.pasteAvailableChanged)
 
 		self.updateTitle()
 
@@ -792,6 +820,30 @@ class FupEditMdiSubWindow(EditMdiSubWindow):
 	def redo(self):
 		return self.fupWidget.redo()
 
+	def cutIsAvailable(self):
+		return self.fupWidget.clipboardCutIsAvailable()
+
+	def cut(self):
+		return self.fupWidget.clipboardCut()
+
+	def copyIsAvailable(self):
+		return self.fupWidget.clipboardCopyIsAvailable()
+
+	def copy(self):
+		return self.fupWidget.clipboardCopy()
+
+	def pasteIsAvailable(self):
+		return self.fupWidget.clipboardPasteIsAvailable()
+
+	def paste(self, text=None):
+		return self.fupWidget.clipboardPaste(text)
+
+	def handleAwlSimError(self, exception):
+		self.fupWidget.handleAwlSimError(exception)
+
+	def __handleDiagramChanged(self):
+		self.validateDocument.emit(0.5)
+
 class KopEditMdiSubWindow(EditMdiSubWindow):
 	TYPE = EditMdiSubWindow.TYPE_KOP
 
