@@ -26,6 +26,9 @@ from awlsim.common.compat import *
 from awlsim.gui.util import *
 from awlsim.gui.cpuwidget import *
 from awlsim.gui.sourcecodeedit import *
+from awlsim.gui.runstate import *
+
+import random
 
 
 class EditSubWidget(QWidget):
@@ -354,6 +357,14 @@ class EditWidget(SourceCodeEdit):
 	__runAniNoDown = tuple(c.replace("x", " source NOT DOWNLOADED to CPU ")
 			       for c in __runAniTemplate)
 
+	__didYouKnow = (
+		"blocks of code can be commented or uncommented "
+		"by selecting it and then pressing Ctrl + /",
+
+		"pressing Ctrl and moving the mouse scroll wheel "
+		"resizes the text font",
+	)
+
 	def __init__(self, parent=None,
 		     readOnly=False,
 		     withHeader=True, withCpuStats=True):
@@ -418,8 +429,10 @@ class EditWidget(SourceCodeEdit):
 			self.cpuStatsWidget.contextMenuReq.connect(self.__cpuStatsContextMenuPopup)
 			self.__cpuStatsMenu.closed.connect(self.__cpuStatsContextMenuClosed)
 
-	def getSourceId(self):
-		return self.__source.identHash
+		self.setPlaceholderText(
+			"Please enter your AWL program here...\n\n\n"
+			"Did you know\n" + random.choice(self.__didYouKnow)
+		)
 
 	def shutdown(self):
 		pass
@@ -466,6 +479,9 @@ class EditWidget(SourceCodeEdit):
 			self.__updateSource()
 		return self.__source
 
+	def getSourceId(self):
+		return self.getSource().identHash
+
 	def setSettings(self, guiSettings):
 		self.enableAutoIndent(guiSettings.getEditorAutoIndentEn())
 		self.enablePasteIndent(guiSettings.getEditorPasteIndentEn())
@@ -631,7 +647,7 @@ class EditWidget(SourceCodeEdit):
 		if self.__runState == RunState.STATE_RUN:
 			cpuHashes = [ s.identHash for s in identsMsg.awlSources ]
 			self.__setSourceMatchesCpuSource(
-				self.getSource().identHash in cpuHashes)
+				self.getSourceId() in cpuHashes)
 		else:
 			# The CPU is not in RUN state.
 			# We don't make a big deal out of code mismatch.
@@ -868,6 +884,19 @@ class EditWidget(SourceCodeEdit):
 				return
 		self.setErraticLine(None)
 
+	def _lineIsCommented(self, lineText):
+		return lineText.lstrip().startswith("//")
+
+	def _commentLine(self, lineText):
+		return "//" + lineText
+
+	def _uncommentLine(self, lineText):
+		if self._lineIsCommented(lineText):
+			idx = lineText.find("//")
+			if idx >= 0:
+				lineText = lineText[:idx] + lineText[idx+len("//"):]
+		return lineText
+
 class EditDialog(QDialog):
 	"""AWL/STL edit dialog."""
 
