@@ -2,7 +2,7 @@
 #
 # AWL simulator - GUI main window
 #
-# Copyright 2012-2018 Michael Buesch <m@bues.ch>
+# Copyright 2012-2020 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -37,8 +37,51 @@ from awlsim.gui.hwmodconfig import *
 from awlsim.gui.icons import *
 from awlsim.gui.templatedialog import *
 from awlsim.gui.library import *
+from awlsim.gui.runstate import *
+from awlsim.gui.toolbars import *
+from awlsim.gui.validatorsched import *
 
 
+class LoadProgressDialog(QDialog):
+	def __init__(self, parent=None):
+		QDialog.__init__(self, parent)
+		self.setLayout(QGridLayout())
+		self.setContentsMargins(30, 30, 30, 30)
+		self.setWindowModality(Qt.ApplicationModal)
+		self.__isShown = False
+		self.__icon = QLabel(self)
+		self.layout().addWidget(self.__icon, 0, 0)
+		self.layout().setColumnMinimumWidth(1, 60)
+		self.__text = QLabel(self)
+		self.layout().addWidget(self.__text, 0, 2)
+
+	def closeEvent(self, ev):
+		if self.__isShown:
+			ev.ignore()
+			return
+		QDialog.closeEvent(self, ev)
+
+	def showProgress(self):
+		self.__isShown = True
+		self.show()
+		QApplication.processEvents(QEventLoop.ExcludeUserInputEvents, 50)
+
+	def hideProgress(self):
+		self.__isShown = False
+		self.hide()
+		QApplication.processEvents(QEventLoop.ExcludeUserInputEvents, 50)
+
+	def setCpuRunState(self, cpuRunState):
+		if cpuRunState.state == cpuRunState.STATE_LOAD:
+			self.setWindowTitle("Awlsim - Downloading...")
+			self.__icon.setPixmap(getIcon("download").pixmap(64, 64))
+			self.__text.setText("Downloading project to CPU.\n\n"
+					    "Please be patient.\n"
+					    "This might take a few seconds.")
+			self.showProgress()
+		else:
+			self.hideProgress()
+
 class CpuDockWidget(QDockWidget):
 	def __init__(self, mainWidget, parent=None):
 		QDockWidget.__init__(self, "", parent)
@@ -95,7 +138,7 @@ class MainWidget(QWidget):
 	# Signal: Project loaded
 	projectLoaded = Signal(Project)
 	# Signal: Dirty-status changed
-	dirtyChanged = Signal(bool)
+	dirtyChanged = Signal(int)
 	# Signal: Source text focus changed
 	textFocusChanged = Signal(bool)
 	# Signal: UndoAvailable state changed
@@ -109,18 +152,33 @@ class MainWidget(QWidget):
 	# Signal: PasteAvailable state changed
 	pasteAvailableChanged = Signal(bool)
 
+	# Document dirty levels
+	EnumGen.start
+	DIRTY_NO	= EnumGen.item
+	DIRTY_SLIGHT	= EnumGen.item
+	DIRTY_FULL	= EnumGen.item
+	EnumGen.end
+
 	def __init__(self, mainWindow, parent=None):
 		QWidget.__init__(self, parent)
 		self.setLayout(QGridLayout(self))
 		self.mainWindow = mainWindow
 
-		self.simClient = GuiAwlSimClient()
+		self.simClient = GuiAwlSimClient(self)
 
 		self.editMdiArea = EditMdiArea(self)
 		self.layout().addWidget(self.editMdiArea, 0, 0)
 
 		self.filename = None
-		self.dirty = False
+		self.__dirtyLevel = self.DIRTY_NO
+		self.__cpuRunState = RunState()
+		self.__insnPerSecond = 0.0
+		self.__avgCycleTime = 0.0
+		self.__minCycleTime = 0.0
+		self.__maxCycleTime = 0.0
+		self.__padCycleTime = 0.0
+
+		self.__updateStatusBar()
 
 		self.editMdiArea.sourceChanged.connect(self.somethingChanged)
 		self.editMdiArea.focusChanged.connect(self.textFocusChanged)
@@ -138,15 +196,20 @@ class MainWidget(QWidget):
 		return self.projectTreeModel.getProject()
 
 	def isDirty(self):
-		return self.dirty
+		return self.__dirtyLevel == self.DIRTY_FULL
 
-	def setDirty(self, dirty, force=False):
-		if dirty != self.dirty or force:
-			self.dirty = dirty
-			self.dirtyChanged.emit(self.dirty)
+	def setDirty(self, dirtyLevel=DIRTY_FULL, force=False):
+		if dirtyLevel != self.__dirtyLevel or force:
+			if (not force and
+			    self.__dirtyLevel == self.DIRTY_FULL and
+			    dirtyLevel == self.DIRTY_SLIGHT):
+				# Cannot go from full to slight.
+				return
+			self.__dirtyLevel = dirtyLevel
+			self.dirtyChanged.emit(self.__dirtyLevel)
 
 	def somethingChanged(self):
-		self.setDirty(True)
+		self.setDirty(self.DIRTY_FULL)
 
 	def getFilename(self):
 		return self.filename
@@ -157,6 +220,16 @@ class MainWidget(QWidget):
 	def getCpuWidget(self):
 		return self.mainWindow.cpuWidget
 
+	def __updateCpuViewConfig(self):
+		"""Update the GuiCpuStateViewSettings objects in the project.
+		"""
+		settingsList = []
+
+		stateMdiArea = self.getCpuWidget().stateMdi
+		settingsList.append(stateMdiArea.getSettings())
+
+		self.getProject().getGuiSettings().setCpuStateViewSettingsList(settingsList)
+
 	def newFile(self, filename=None):
 		if isWinStandalone:
 			executableName = "awlsim-gui.exe"
@@ -186,7 +259,7 @@ class MainWidget(QWidget):
 			return
 
 	def loadFile(self, filename, newIfNotExist=False):
-		if self.dirty:
+		if self.isDirty():
 			res = QMessageBox.question(self,
 				"Unsaved project",
 				"The current project is modified and contains unsaved changes.\n "
@@ -211,7 +284,8 @@ class MainWidget(QWidget):
 			else:
 				assert(0)
 		self.getCpuWidget().goOffline()
-		if not fileExists(filename) and newIfNotExist:
+		self.getCpuWidget().stateMdi.reset()
+		if not os.path.exists(filename) and newIfNotExist:
 			# The file does not exist. We implicitly create it.
 			# The actual file will be created when the project is saved.
 			isNewProject = True
@@ -221,24 +295,26 @@ class MainWidget(QWidget):
 			isNewProject = False
 			try:
 				self.projectTreeModel.loadProjectFile(filename, self)
+				guiSettings = self.getProject().getGuiSettings()
+				for viewSettings in guiSettings.getCpuStateViewSettingsList():
+					# We currently only have one CPU state view.
+					# If the project has multiple viewSettings, all but the
+					# last one will be lost.
+					self.getCpuWidget().stateMdi.loadFromCpuStateViewSettings(viewSettings)
 			except AwlSimError as e:
 				QMessageBox.critical(self,
 					"Failed to load project file", str(e))
 				return False
 		self.filename = filename
 		if isNewProject or not self.getProject().getProjectFile():
-			self.setDirty(True, force=True)
+			self.setDirty(self.DIRTY_FULL, force=True)
 		else:
-			self.setDirty(False, force=True)
+			self.setDirty(self.DIRTY_NO, force=True)
 		self.projectLoaded.emit(self.getProject())
 		return True
 
 	def load(self):
-		if isPyQt and isQt4:
-			getOpenFileName = QFileDialog.getOpenFileNameAndFilter
-		else:
-			getOpenFileName = QFileDialog.getOpenFileName
-		fn, fil = getOpenFileName(self,
+		fn, fil = QFileDialog.getOpenFileName(self,
 			"Open project", "",
 			"Awlsim project or AWL/STL source (*.awlpro *.awl);;"
 			"Awlsim project (*.awlpro);;"
@@ -250,6 +326,7 @@ class MainWidget(QWidget):
 
 	def saveFile(self, filename):
 		try:
+			self.__updateCpuViewConfig()
 			res = self.projectTreeModel.saveProjectFile(filename, self)
 			if res == 0: # Failure
 				return False
@@ -260,16 +337,12 @@ class MainWidget(QWidget):
 				"Failed to write project file", str(e))
 			return False
 		self.filename = filename
-		self.setDirty(dirty = False, force = True)
+		self.setDirty(self.DIRTY_NO, force=True)
 		return True
 
 	def save(self, newFile=False):
 		if newFile or not self.filename:
-			if isPyQt and isQt4:
-				getSaveFileName = QFileDialog.getSaveFileNameAndFilter
-			else:
-				getSaveFileName = QFileDialog.getSaveFileName
-			fn, fil = getSaveFileName(self,
+			fn, fil = QFileDialog.getSaveFileName(self,
 				"Awlsim project save as", "",
 				"Awlsim project (*.awlpro)",
 				"*.awlpro")
@@ -340,7 +413,6 @@ class MainWidget(QWidget):
 		libSelection: The AwlLibEntrySelection() instance.
 		Returns True, if the selection has successfully been added.
 		"""
-		assert(isinstance(libSelection, AwlLibEntrySelection))
 		return self.projectTreeModel.libSelectionAdd(libSelection,
 							     parentWidget=self)
 
@@ -476,6 +548,62 @@ class MainWidget(QWidget):
 			return projectTreeModel.entryActivate(index, parentWidget=self)
 		return False
 
+	def handleCpuRunStateChange(self, cpuRunState):
+		"""CPU RunState changed.
+		"""
+		self.__cpuRunState = cpuRunState
+		self.__updateStatusBar()
+		self.mainWindow.loadProgressDialog.setCpuRunState(cpuRunState)
+
+	def handleCpuStats(self, statsMsg):
+		"""Received new AwlSimMessage_CPUSTATS.
+		"""
+		self.__insnPerSecond = statsMsg.insnPerSecond
+		self.__avgCycleTime = statsMsg.avgCycleTime
+		self.__minCycleTime = statsMsg.minCycleTime
+		self.__maxCycleTime = statsMsg.maxCycleTime
+		self.__padCycleTime = statsMsg.padCycleTime
+		self.__updateStatusBar()
+
+	def __updateStatusBar(self):
+		"""Update the main window status bar.
+		"""
+		status = []
+
+		if self.__cpuRunState == RunState.STATE_OFFLINE:
+			status.append("CPU: offline")
+		elif self.__cpuRunState == RunState.STATE_ONLINE:
+			status.append("CPU: online / STOP")
+		elif self.__cpuRunState == RunState.STATE_LOAD:
+			status.append("CPU: loading")
+		elif self.__cpuRunState == RunState.STATE_RUN:
+			status.append("CPU: RUN")
+		elif self.__cpuRunState == RunState.STATE_EXCEPTION:
+			status.append("CPU: EXCEPTION")
+
+		if self.__cpuRunState == RunState.STATE_RUN:
+			if self.__insnPerSecond > 0.0:
+				usPerInsnStr = "%.02f" % ((1.0 / self.__insnPerSecond) * 1000000.0)
+				status.append("%s stmt/s (%s µs/stmt)" % (
+					      floatToHumanReadable(self.__insnPerSecond),
+					      usPerInsnStr))
+
+			if (self.__avgCycleTime > 0.0 and
+			    self.__minCycleTime > 0.0 and
+			    self.__maxCycleTime > 0.0):
+				avgCycleTimeStr = "%.01f" % (self.__avgCycleTime * 1000.0)
+				minCycleTimeStr = "%.01f" % (self.__minCycleTime * 1000.0)
+				maxCycleTimeStr = "%.01f" % (self.__maxCycleTime * 1000.0)
+				padCycleTimeStr = "%.01f" % (self.__padCycleTime * 1000.0)
+				status.append("OB1: avg: %s ms  min: %s ms  max: %s ms  padding: %s ms" % (
+					      avgCycleTimeStr,
+					      minCycleTimeStr,
+					      maxCycleTimeStr,
+					      padCycleTimeStr))
+
+		statusBar = self.mainWindow.statusBar()
+		statusBar.showMessage("  --  ".join(status))
+
 class MainWindow(QMainWindow):
 	TITLE = "Awlsim PLC v%s" % VERSION_STRING
 
@@ -491,12 +619,22 @@ class MainWindow(QMainWindow):
 
 		mainwnd = cls(initialAwlSource)
 		mainwnd.show()
+
+		if initialAwlSource and not mainwnd.mainWidget.isDirty():
+			# Revert back from DIRTY_SLIGHT to DIRTY_NO.
+			mainwnd.mainWidget.setDirty(mainwnd.mainWidget.DIRTY_NO,
+						    force=True)
+
 		return mainwnd
 
 	def __init__(self, awlSource=None, parent=None):
 		QMainWindow.__init__(self, parent)
 		self.setWindowIcon(getIcon("cpu"))
 
+		self.__profiler = None
+
+		self.setStatusBar(QStatusBar(self))
+
 		self.mainWidget = MainWidget(self, self)
 		self.cpuDockWidget = CpuDockWidget(self.mainWidget, self)
 		self.treeDockWidget = ProjectTreeDockWidget(self.mainWidget, self)
@@ -636,10 +774,22 @@ class MainWindow(QMainWindow):
 		menu.addAction(getIcon("browser"), "Awlsim &homepage...", self.awlsimHomepage)
 		menu.addSeparator()
 		menu.addAction(getIcon("cpu"), "&About...", self.about)
+		menu.addSeparator()
+		self.__actProfileStart = menu.addAction(getIcon("enable"),
+							"Start profiling",
+							self.profileStart)
+		self.__actProfileStop = menu.addAction(getIcon("disable"),
+						       "Stop profiling",
+						       self.profileStop)
+		profEnabled = (AwlSimEnv.getProfileLevel() > 0)
+		self.__actProfileStart.setVisible(profEnabled)
+		self.__actProfileStop.setVisible(False)
 		self.menuBar().addMenu(menu)
 
+		self.__loadProgressDialog = LoadProgressDialog(self)
+
 		self.__sourceTextHasFocus = False
-		self.__dirtyChanged(False)
+		self.__dirtyChanged(MainWidget.DIRTY_NO)
 		self.__textFocusChanged(False)
 		self.__undoAvailableChanged(False)
 		self.__redoAvailableChanged(False)
@@ -647,6 +797,8 @@ class MainWindow(QMainWindow):
 		self.__cutAvailableChanged(False)
 		self.__pasteAvailableChanged(False)
 
+		client = self.getSimClient()
+
 		self.mainWidget.projectLoaded.connect(self.__handleProjectLoaded)
 		self.mainWidget.dirtyChanged.connect(self.__dirtyChanged)
 		self.mainWidget.textFocusChanged.connect(self.__textFocusChanged)
@@ -660,20 +812,28 @@ class MainWindow(QMainWindow):
 		self.inspectTb.connectToCpuWidget(self.cpuWidget)
 		self.mainWidget.dirtyChanged.connect(self.cpuWidget.handleDirtyChange)
 		self.editMdiArea.visibleLinesChanged.connect(self.cpuWidget.updateVisibleLineRange)
+		GuiValidatorSched.get().haveValidationResult.connect(self.projectTreeModel.handleAwlSimError)
 		self.cpuWidget.onlineDiagChanged.connect(self.editMdiArea.enableOnlineDiag)
 		self.cpuWidget.haveInsnDump.connect(self.editMdiArea.handleInsnDump)
 		self.cpuWidget.haveIdentsMsg.connect(self.editMdiArea.handleIdentsMsg)
 		self.cpuWidget.haveIdentsMsg.connect(self.projectTreeModel.handleIdentsMsg)
-		self.cpuWidget.runStateChanged.connect(self.editMdiArea.setCpuRunState)
-		self.cpuWidget.runStateChanged.connect(self.projectTreeModel.setCpuRunState)
+		self.cpuWidget.haveCpuStats.connect(self.mainWidget.handleCpuStats)
 		self.cpuWidget.configChanged.connect(self.mainWidget.somethingChanged)
 		self.projectTreeModel.projectContentChanged.connect(self.mainWidget.somethingChanged)
+		client.haveException.connect(self.projectTreeModel.handleAwlSimError)
+		client.runState.stateChanged.connect(self.editMdiArea.setCpuRunState)
+		client.runState.stateChanged.connect(self.projectTreeModel.setCpuRunState)
+		client.runState.stateChanged.connect(self.mainWidget.handleCpuRunStateChange)
 
 		if awlSource:
 			self.mainWidget.loadFile(awlSource, newIfNotExist=True)
 
 		self.__restoreState()
 
+	@property
+	def loadProgressDialog(self):
+		return self.__loadProgressDialog
+
 	@property
 	def editMdiArea(self):
 		return self.mainWidget.editMdiArea
@@ -774,14 +934,14 @@ class MainWindow(QMainWindow):
 		cpuDockEn = self.cpuDockWidget.toggleViewAction().isChecked()
 		self.inspectTb.toggleViewAction().setEnabled(cpuDockEn)
 
-	def __dirtyChanged(self, isDirty):
-		self.saveAct.setEnabled(isDirty)
-		self.tbSaveAct.setEnabled(isDirty)
+	def __dirtyChanged(self, dirtyLevel):
+		self.saveAct.setEnabled(dirtyLevel != MainWidget.DIRTY_NO)
+		self.tbSaveAct.setEnabled(dirtyLevel != MainWidget.DIRTY_NO)
 
 		filename = self.mainWidget.getFilename()
 		if filename:
 			postfix = " -- " + os.path.basename(filename)
-			if isDirty:
+			if dirtyLevel == MainWidget.DIRTY_FULL:
 				postfix += "*"
 		else:
 			postfix = ""
@@ -850,6 +1010,7 @@ class MainWindow(QMainWindow):
 		self.getSimClient().shutdown()
 		ev.accept()
 		QMainWindow.closeEvent(self, ev)
+		self.profileStop()
 
 	def keyPressEvent(self, ev):
 		if ev.matches(QKeySequence.Save):
@@ -870,7 +1031,7 @@ class MainWindow(QMainWindow):
 		QMessageBox.about(self, "About Awlsim PLC",
 			"Awlsim PLC version %s\n"
 			"\n"
-			"Copyright 2012-2018 Michael Büsch <m@bues.ch>\n"
+			"Copyright 2012-2019 Michael Büsch <m@bues.ch>\n"
 			"\n"
 			"Project home:  %s\n"
 			"\n"
@@ -889,3 +1050,26 @@ class MainWindow(QMainWindow):
 			"with this program; if not, write to the Free Software Foundation, Inc., "
 			"51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA." %\
 			(VERSION_STRING, AWLSIM_HOME_URL))
+
+	def profileStart(self):
+		if not self.__profiler:
+			try:
+				self.__profiler = Profiler()
+				self.__profiler.start()
+			except AwlSimError as e:
+				MessageBox.handleAwlSimError(self,
+					"Failed to start profiler.", e)
+				self.__profiler = None
+				return
+
+			self.__actProfileStart.setVisible(False)
+			self.__actProfileStop.setVisible(True)
+
+	def profileStop(self):
+		if self.__profiler:
+			self.__profiler.stop()
+			printInfo("GUI profiler dump:\n" + self.__profiler.getResult())
+			self.__profiler = None
+
+			self.__actProfileStart.setVisible(True)
+			self.__actProfileStop.setVisible(False)
