@@ -2,7 +2,7 @@
 #
 # AWL simulator - Project tree widget
 #
-# Copyright 2017-2018 Michael Buesch <m@bues.ch>
+# Copyright 2017-2020 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -29,6 +29,9 @@ from awlsim.library.libselection import AwlLibEntrySelection
 from awlsim.gui.cpuwidget import RunState
 from awlsim.gui.util import *
 from awlsim.gui.icons import *
+from awlsim.gui.runstate import *
+from awlsim.gui.fup.fupwidget import FupFactory, FupWidget
+from awlsim.gui.validatorsched import *
 
 
 __all__ = [
@@ -119,6 +122,30 @@ class ProjectTreeModel(QAbstractItemModel):
 		"""
 		return self.mainWidget.editMdiArea
 
+	def handleAwlSimError(self, exception):
+		"""An AwlSimError occurred (due to normal operation or validation).
+		"""
+		anyChanged = False
+		project = self.getProject()
+		for source in project.getAllSources():
+			# Set error flag on corresponding source.
+			if exception and exception.getSourceId() == source.identHash:
+				# Set error flag.
+				if not source.userData.get("gui-erroneous", False):
+					anyChanged = True
+				source.userData["gui-erroneous"] = True
+			else:
+				# Clear error flag.
+				if source.userData.get("gui-erroneous", False):
+					anyChanged = True
+				source.userData["gui-erroneous"] = False
+
+		# Mark the program container as changed.
+		if anyChanged:
+			index = self.idToIndex(self.INDEXID_SRCS)
+			roles = (Qt.DecorationRole, Qt.ToolTipRole)
+			self.dataChanged.emit(index, index, roles)
+
 	def handleIdentsMsg(self, identsMsg):
 		"""Handle a new identifier message from AwlSimClient().
 		identsMsg: AwlSimMessage_IDENTS() instance.
@@ -528,7 +555,6 @@ class ProjectTreeModel(QAbstractItemModel):
 		symbol: The Symbol() instance to add.
 		Returns True, if adding was successfull.
 		"""
-		assert(isinstance(symbol, Symbol))
 		project = self.getProject()
 		mnemonics = project.getCpuConf().getConfiguredMnemonics()
 		symTabSources = project.getSymTabSources()
@@ -600,7 +626,6 @@ class ProjectTreeModel(QAbstractItemModel):
 		libSelection: The AwlLibEntrySelection() to add.
 		Returns True, if adding was successfull.
 		"""
-		assert(isinstance(libSelection, AwlLibEntrySelection))
 		project = self.getProject()
 		libSelections = project.getLibSelections()
 
@@ -758,6 +783,7 @@ class ProjectTreeModel(QAbstractItemModel):
 			# Try to load it.
 			self.__loadPlainAwlSourceFile(filename, parentWidget)
 		self.projectLoaded.emit()
+		GuiValidatorSched.get().startAsyncValidation(self.getProject)
 
 	def saveProjectFile(self, filename, parentWidget):
 		if self.__isAdHocProject:
@@ -785,6 +811,7 @@ class ProjectTreeModel(QAbstractItemModel):
 			return 0
 		project = self.__project
 
+		# Check if there is any deprecated file-backing.
 		if (any(src.isFileBacked() for src in project.getAwlSources()) or\
 		    any(src.isFileBacked() for src in project.getSymTabSources())) and\
 		    not self.__warnedFileBacked:
@@ -797,12 +824,35 @@ class ProjectTreeModel(QAbstractItemModel):
 				"in the source menu.")
 			self.__warnedFileBacked = True
 
+		# Re-generate UUIDs, if we are saving to a new file.
+		if project.getProjectFile() != filename:
+			printInfo("Save-file name changed. "
+				  "Re-generating project file UUIDs.")
+			self.__regenAllUUIDs()
+
+		# Set the new file name and write to that file.
 		project.setProjectFile(filename)
 		project.toFile()
 		self.__isAdHocProject = False
 
 		return 1
 
+	def __regenAllUUIDs(self):
+		"""Re-generate all UUIDs in all sources.
+		"""
+		for source in self.__project.getFupSources():
+			# Parse the FUP source and regenerate all UUIDs.
+			fupWidget = FupWidget(None, None)
+			factory = FupFactory(fupWidget=fupWidget)
+			factory.parse(source.sourceBytes)
+			fupWidget.regenAllUUIDs()
+			source.sourceBytes = factory.compose()
+
+			# Update the source in the MDI window, if any.
+			mdiSubWin = source.userData.get("gui-edit-window")
+			if mdiSubWin:
+				mdiSubWin.setSource(source)
+
 	def idToIndex(self, idxId, column = 0):
 		for table in (self.id2row_toplevel,
 			      self.id2row_srcs):
@@ -976,19 +1026,26 @@ class ProjectTreeModel(QAbstractItemModel):
 		"""
 
 		def getSourceContainerIcon(sourceList, okIconName):
+			# If there is an error in the source, show error icon.
 			# If the identHash of any source does not match what's on the CPU,
 			# use a warning icon. But only do that, if the CPU is
 			# in RUN state.
-			if self.__cpuRunState == RunState.STATE_RUN and\
-			   any(not source.userData.get("gui-cpu-idents-match", True)
+			if any(source.userData.get("gui-erroneous", False)
 			       for source in sourceList):
+				return getIcon("exit")
+			elif (self.__cpuRunState == RunState.STATE_RUN and
+			      any(not source.userData.get("gui-cpu-idents-match", True)
+				  for source in sourceList)):
 				return getIcon("warning")
 			return getIcon(okIconName)
 
 		def getSourceIcon(sourceList, okIconName):
 			if itemNr >= len(sourceList):
-				return False
-			return getSourceContainerIcon((sourceList[itemNr],), okIconName)
+				return None
+			source = sourceList[itemNr]
+			if not source.enabled:
+				return getIcon("disable")
+			return getSourceContainerIcon((source,), okIconName)
 
 		def getProgramContainerIcon(okIconName):
 			project = self.getProject()
@@ -1043,9 +1100,14 @@ class ProjectTreeModel(QAbstractItemModel):
 		"""
 
 		def getSourceContainerTip(sourceList, okToolTip):
+			# If there is an error in the source, show error tool tip.
 			# If the identHash of any source does not match what's on the CPU,
 			# use a warning tool tip. But only do that, if the CPU is
 			# in RUN state.
+			if any(source.userData.get("gui-erroneous", False)
+			       for source in sourceList):
+				return "ERROR: There is an error in the source file.\n"\
+				       "Please open it to fix the problem."
 			if self.__cpuRunState == RunState.STATE_RUN and\
 			   any(not source.userData.get("gui-cpu-idents-match", True)
 			       for source in sourceList):
@@ -1150,20 +1212,10 @@ class ProjectTreeModel(QAbstractItemModel):
 				title = "Project"
 			if self.mainWidget.isDirty():
 				title += "*"
-			if self.__cpuRunState == RunState.STATE_OFFLINE:
-				title += " (CPU: offline)"
-			elif self.__cpuRunState == RunState.STATE_ONLINE:
-				title += " (CPU: online / STOP)"
-			elif self.__cpuRunState == RunState.STATE_LOAD:
-				title += " (CPU: loading)"
-			elif self.__cpuRunState == RunState.STATE_RUN:
-				title += " (CPU: RUN)"
-			elif self.__cpuRunState == RunState.STATE_EXCEPTION:
-				title += " (CPU: EXCEPTION)"
 			return title
 		return None
 
-	def __handleProjectDirtyChanged(self, dirty):
+	def __handleProjectDirtyChanged(self, dirtyLevel):
 		self.headerDataChanged.emit(Qt.Horizontal, 0, 0)
 
 	def mimeData(self, indexes):
@@ -1338,44 +1390,44 @@ class SourceContextMenu(QMenu):
 
 		if withEditButton:
 			self.addAction(getIcon("doc_edit"),
-				       "&Edit '%s'..." % itemName, self.__edit)
+				       "&Edit...", self.__edit)
 		self.addSeparator()
 		if withAddButton:
 			self.addAction(getIcon("doc_new"),
-				       "&Add %s" % itemCategoryName, self.__add)
+				       "&Add new %s" % itemCategoryName, self.__add)
 		if withDeleteButton:
 			self.addAction(getIcon("doc_close"),
-				       "&Delete '%s'..." % itemName, self.__delete)
+				       "&Delete...", self.__delete)
 		if withRenameButton:
 			self.addAction(getIcon("doc_edit"),
-				       "&Rename '%s'..." % itemName, self.__rename)
+				       "&Rename...", self.__rename)
 		self.addSeparator()
 		if withCopyButton:
 			self.addAction(getIcon("copy"),
-				       "&Copy '%s'..." % itemName, self.__copy)
+				       "&Copy", self.__copy)
 		if withCutButton:
 			self.addAction(getIcon("cut"),
-				       "&Cut '%s'..." % itemName, self.__cut)
+				       "&Cut", self.__cut)
 		if withPasteButton:
 			self.addAction(getIcon("paste"),
-				       "&Paste '%s'..." % itemCategoryName, self.__paste)
+				       "&Paste", self.__paste)
 		self.addSeparator()
 		if withImportButton:
 			self.addAction(getIcon("doc_import"),
-				       "&Import %s..." % itemCategoryName, self.__import)
+				       "&Import...", self.__import)
 		if withExportButton:
 			self.addAction(getIcon("doc_export"),
-				       "&Export '%s'..." % itemName, self.__export)
+				       "&Export...", self.__export)
 		if withIntegrateButton:
 			self.addAction("&Integrate '%s' into project..." % itemName,
 				       self.__integrate)
 		self.addSeparator()
 		if withEnableButton:
-			self.addAction("E&nable '%s'" % itemName,
-				       self.__enable)
+			self.addAction(getIcon("enable"),
+				       "E&nable", self.__enable)
 		if withDisableButton:
-			self.addAction("D&isable '%s'" % itemName,
-				       self.__disable)
+			self.addAction(getIcon("disable"),
+				       "D&isable", self.__disable)
 
 	def __edit(self):
 		self.edit.emit()
@@ -1582,7 +1634,7 @@ class ProjectTreeView(QTreeView):
 		menu.import_.connect(handleImport)
 		menu.export.connect(handleExport)
 		menu.enable.connect(handleEnable)
-		menu.exec_(QCursor.pos())
+		menu.exec_(QCursor.pos() + QPoint(3, 3))
 
 	def __mouseBtnPressed(self, index):
 		model, buttons = self.model(), QApplication.mouseButtons()
