@@ -2,7 +2,7 @@
 #
 # AWL simulator - GUI utility functions
 #
-# Copyright 2012-2018 Michael Buesch <m@bues.ch>
+# Copyright 2012-2020 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -25,11 +25,18 @@ from awlsim_loader.common import *
 from awlsim_loader.coreclient import *
 import awlsim_loader.cython_helper as cython_helper
 from awlsim.common.monotonic import monotonic_time #+cimport
+from awlsim.common.datatypehelpers import * #+cimport
+from awlsim.common.profiler import *
+from awlsim.core.datatypes import AwlDataType
+from awlsim.core.symbolparser import SymbolTable, Symbol, SymTabParser
 
 import sys
 import traceback
 import xml.sax.saxutils as saxutils
 
+if isPy2Compat:
+	printWarning("WARNING: Python 2.x is not supported by awlsim-gui.")
+
 if isPyPy or isJython:
 	# PySide does not work on PyPy or Jython, yet.
 	printError("Running awlsim-gui on the PyPy or Jython interpreter is not supported.")
@@ -85,8 +92,15 @@ def handleFatalException(parentWidget=None):
 
 
 class MessageBox(QDialog):
-	def __init__(self, parent, title, text,
-		     verboseText=None, icon=QMessageBox.Critical):
+	def __init__(self,
+		     parent,
+		     title,
+		     text,
+		     verboseText=None,
+		     icon=QMessageBox.Critical,
+		     okButton=True,
+		     continueButton=False,
+		     cancelButton=False):
 		QDialog.__init__(self, parent)
 		self.setLayout(QGridLayout())
 		self.setWindowTitle(title)
@@ -109,12 +123,26 @@ class MessageBox(QDialog):
 		else:
 			self.verboseCheckBox = None
 
-		self.okButton = QPushButton("&Ok", self)
-		self.layout().addWidget(self.okButton, 2, 1)
+		buttonsLayout = QHBoxLayout()
+		if okButton:
+			self.okButton = QPushButton("&Ok", self)
+			buttonsLayout.addWidget(self.okButton)
+		if continueButton:
+			self.continueButton = QPushButton("C&ontinue", self)
+			buttonsLayout.addWidget(self.continueButton)
+		if cancelButton:
+			self.cancelButton = QPushButton("&Cancel", self)
+			buttonsLayout.addWidget(self.cancelButton)
+		self.layout().addLayout(buttonsLayout, 2, 1)
 
 		self.__updateText()
 
-		self.okButton.released.connect(self.accept)
+		if okButton:
+			self.okButton.released.connect(self.accept)
+		if continueButton:
+			self.continueButton.released.connect(self.accept)
+		if cancelButton:
+			self.cancelButton.released.connect(self.reject)
 		if self.verboseCheckBox:
 			self.verboseCheckBox.stateChanged.connect(self.__updateText)
 
@@ -126,30 +154,34 @@ class MessageBox(QDialog):
 			self.textBox.setText(self.text)
 
 	@classmethod
-	def error(cls, parent, text, verboseText=None):
-		dlg = cls(parent = parent,
-			  title = "Awlsim - Error",
-			  text = text,
-			  verboseText = verboseText,
-			  icon = QMessageBox.Critical)
+	def error(cls, parent, text, verboseText=None, **kwargs):
+		dlg = cls(parent=parent,
+			  title="Awlsim - Error",
+			  text=text,
+			  verboseText=verboseText,
+			  icon=QMessageBox.Critical,
+			  **kwargs)
 		res = dlg.exec_()
 		dlg.deleteLater()
 		return res
 
 	@classmethod
-	def warning(cls, parent, text, verboseText=None):
-		dlg = cls(parent = parent,
-			  title = "Awlsim - Warning",
-			  text = text,
-			  verboseText = verboseText,
-			  icon = QMessageBox.Warning)
+	def warning(cls, parent, text, verboseText=None, **kwargs):
+		dlg = cls(parent=parent,
+			  title="Awlsim - Warning",
+			  text=text,
+			  verboseText=verboseText,
+			  icon=QMessageBox.Warning,
+			  **kwargs)
 		res = dlg.exec_()
 		dlg.deleteLater()
 		return res
 
+	awlSimErrorBlocked = Blocker()
+
 	@classmethod
-	def handleAwlSimError(cls, parent, description, exception):
-		if exception.getSeenByUser():
+	def handleAwlSimError(cls, parent, description, exception, **kwargs):
+		if exception.getSeenByUser() or cls.awlSimErrorBlocked:
 			return cls.Accepted
 		exception.setSeenByUser()
 		def maketext(verbose):
@@ -160,10 +192,14 @@ class MessageBox(QDialog):
 			text += "\n\n"
 			text += exception.getReport(verbose)
 			return text
-		return cls.error(parent, maketext(False), maketext(True))
+		return cls.error(parent=parent,
+				 text=maketext(False),
+				 verboseText=maketext(True),
+				 **kwargs)
 
 	@classmethod
-	def handleAwlParserError(cls, parent, exception):
-		return cls.handleAwlSimError(parent = parent,
-					     description = None,
-					     exception = exception)
+	def handleAwlParserError(cls, parent, exception, **kwargs):
+		return cls.handleAwlSimError(parent=parent,
+					     description=None,
+					     exception=exception,
+					     **kwargs)
