@@ -2,7 +2,7 @@
 #
 # AWL simulator - GUI text value line edit widget
 #
-# Copyright 2014 Michael Buesch <m@bues.ch>
+# Copyright 2014-2015 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -20,6 +20,7 @@
 #
 
 from __future__ import division, absolute_import, print_function, unicode_literals
+#from awlsim.common.cython_support cimport * #@cy
 from awlsim.common.compat import *
 
 from awlsim.gui.util import *
@@ -31,7 +32,12 @@ class ValidatorCallback(QValidator):
 		self.callback = callback
 
 	def validate(self, inputString, pos):
-		return self.callback(inputString, pos)
+		state = self.callback(inputString, pos)
+		if isPySide:
+			return state
+		elif isPyQt:
+			return (state, inputString, pos)
+		assert(0)
 
 class ValueLineEdit(QLineEdit):
 	valueChanged = Signal(str)
@@ -54,16 +60,24 @@ class ValueLineEdit(QLineEdit):
 
 	def __setColors(self, validInput):
 		pal = self.palette()
-		if self.__editing:
-			pal.setColor(QPalette.Base, QColor("#FFFFC0"))
+		if self.isEnabledTo(None):
+			if self.__editing:
+				pal.setColor(QPalette.Base, QColor("#FFFFC0"))
+			else:
+				pal.setColor(QPalette.Base, Qt.white)
 		else:
-			pal.setColor(QPalette.Base, Qt.white)
+			pal.setColor(QPalette.Base, pal.color(QPalette.Window))
 		if validInput:
 			pal.setColor(QPalette.Text, Qt.black)
 		else:
 			pal.setColor(QPalette.Text, Qt.red)
 		self.setPalette(pal)
 
+	def changeEvent(self, ev):
+		QLineEdit.changeEvent(self, ev)
+		if not self.isEnabledTo(None):
+			self.__setColors(True)
+
 	def __validator(self, inputString, pos):
 		res = self.__validatorCallback(inputString, pos)
 		self.__setColors(res == QValidator.Acceptable)
