@@ -2,7 +2,7 @@
 #
 # AWL simulator - Dummy hardware interface
 #
-# Copyright 2013-2014 Michael Buesch <m@bues.ch>
+# Copyright 2013-2017 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -20,15 +20,24 @@
 #
 
 from __future__ import division, absolute_import, print_function, unicode_literals
+#from awlsim.common.cython_support cimport * #@cy
 from awlsim.common.compat import *
 
-from awlsim.core.hardware import *
-from awlsim.core.operators import AwlOperator
-from awlsim.core.datatypes import AwlOffset
+from awlsim.common.util import *
+from awlsim.common.exceptions import *
+
+#from awlsimhw_dummy.main cimport * #@cy
+
+from awlsim.core.hardware_params import *
+from awlsim.core.hardware import * #+cimport
+from awlsim.core.operators import * #+cimport
+from awlsim.core.offset import * #+cimport
+from awlsim.core.cpu import * #+cimport
 
 
-class HardwareInterface(AbstractHardwareInterface):
-	name = "dummy"
+class HardwareInterface_Dummy(AbstractHardwareInterface): #+cdef
+	name		= "dummy"
+	description	= "Dummy hardware module that does nothing."
 
 	def __init__(self, sim, parameters={}):
 		AbstractHardwareInterface.__init__(self,
@@ -41,22 +50,29 @@ class HardwareInterface(AbstractHardwareInterface):
 	def doShutdown(self):
 		pass # Do nothing
 
-	def readInputs(self):
+	def readInputs(self): #+cdef
 		pass # Do nothing
 
-	def writeOutputs(self):
+	def writeOutputs(self): #+cdef
 		pass # Do nothing
 
-	def directReadInput(self, accessWidth, accessOffset):
+	def directReadInput(self, accessWidth, accessOffset): #@nocy
+#@cy	cdef bytearray directReadInput(self, uint32_t accessWidth, uint32_t accessOffset):
 		if accessOffset < self.inputAddressBase:
-			return None
+			return bytearray()
 		# Just read the current value from the CPU and return it.
-		return self.sim.cpu.fetch(AwlOperator(AwlOperator.MEM_E,
-						      accessWidth,
-						      AwlOffset(accessOffset)))
+		try:
+			return self.sim.cpu.fetchInputRange(accessOffset, accessWidth // 8)
+		except AwlSimError as e:
+			# We may be out of process image range. Just return 0.
+			return bytearray( (0,) * (accessWidth // 8) )
 
-	def directWriteOutput(self, accessWidth, accessOffset, data):
+	def directWriteOutput(self, accessWidth, accessOffset, data): #@nocy
+#@cy	cdef ExBool_t directWriteOutput(self, uint32_t accessWidth, uint32_t accessOffset, bytearray data) except ExBool_val:
 		if accessOffset < self.outputAddressBase:
 			return False
 		# Just pretend we wrote it somewhere.
 		return True
+
+# Module entry point
+HardwareInterface = HardwareInterface_Dummy
