@@ -2,7 +2,7 @@
 #
 # PiLC bootstrap
 #
-# Copyright 2016-2018 Michael Buesch <m@bues.ch>
+# Copyright 2016-2020 Michael Buesch <m@bues.ch>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -27,27 +27,24 @@ basedir="$basedir/.."
 
 
 MAIN_MIRROR="http://mirrordirector.raspbian.org/raspbian/"
-DEFAULT_SUITE=stretch
+#MAIN_MIRROR="http://mirror1.hs-esslingen.de/pub/Mirrors/archive.raspbian.org/raspbian/"
+#MAIN_MIRROR="http://ftp.gwdg.de/pub/linux/debian/raspbian/raspbian/"
+DEFAULT_SUITE=buster
 
-RPIGPIO_VERSION="0.6.3"
-RPIGPIO_FILE="RPi.GPIO-$RPIGPIO_VERSION.tar.gz"
-RPIGPIO_MIRROR="https://netcologne.dl.sourceforge.net/project/raspberry-gpio-python/$RPIGPIO_FILE"
-RPIGPIO_SHA256="9366ff36104a39368759929e71f0d8ad6a88553497b3064cbc40f4248806cc19"
-
-SPIDEV_VERSION="3.2"
-SPIDEV_FILE="spidev-$SPIDEV_VERSION.tar.gz"
-SPIDEV_MIRROR="https://pypi.python.org/packages/36/83/73748b6e1819b57d8e1df8090200195cdae33aaa22a49a91ded16785eedd/$SPIDEV_FILE"
-SPIDEV_SHA256="09d2b5122f0dd79910713a11f9a0020f71537224bf829916def4fffc0ea59456"
+KEYRING_VERSION="20120528.2"
+KEYRING_BASEURL="$MAIN_MIRROR/pool/main/r/raspbian-archive-keyring"
+KEYRING_TGZ_FILE="raspbian-archive-keyring_${KEYRING_VERSION}.tar.gz"
+KEYRING_TGZ_SHA256="fdf50f775b60901a2783f21a6362e2bf5ee6203983e884940b163faa1293c002"
 
 PPL_VERSION="0.1.1"
 PPL_FILE="ppl_v$PPL_VERSION.zip"
 PPL_MIRROR="./libs/pixtend/v1/ppl/$PPL_FILE"
 PPL_SHA256="103edcdbc377f8b478fcbc89117cbad143500c611cb714568f55513cece220d4"
 
-PPL2_VERSION="0.1.1-awlsim1"
+PPL2_VERSION="0.1.3"
 PPL2_FILE="pplv2_v$PPL2_VERSION.zip"
 PPL2_MIRROR="./libs/pixtend/v2/pplv2/$PPL2_FILE"
-PPL2_SHA256="ba5259e612beccb55f664bfa22d27390ec766ce655b193c27695fe8a2ceca606"
+PPL2_SHA256="cab6e7cd9062ffbf81a8f570ea0cad663addd8fe22e31cb75e887ae89e425651"
 
 
 info()
@@ -117,11 +114,25 @@ cleanup()
 	done
 }
 
+do_install()
+{
+	install "$@" || die "Failed install $*"
+}
+
+do_systemctl()
+{
+	info "systemctl $*"
+	systemctl "$@" || die "Failed to systemctl $*"
+}
+
 write_image()
 {
 	local image="$1"
 	local dev="$2"
 
+	[ -b "$dev" ] || die "$dev is not a block device"
+	mount | grep -q "$dev" && die "$dev is mounted. Refusing to write to it!"
+
 	if have_program blkdiscard; then
 		info "Discarding $dev ..."
 		blkdiscard "$dev" ||\
@@ -132,93 +143,10 @@ write_image()
 
 	info "Writing $image to $dev ..."
 
-	[ -b "$dev" ] || die "$dev is not a block device"
-	mount | grep -q "$dev" && die "$dev is mounted. Refusing to write to it!"
-
 	dd if="$image" of="$dev" bs=32M status=progress ||\
 		die "Failed to write image."
 }
 
-boot_config_file()
-{
-	cat <<EOF
-# For more options and information see
-# http://www.raspberrypi.org/documentation/configuration/config-txt.md
-# Some settings may impact device functionality. See link above for details
-
-# uncomment if you get no picture on HDMI for a default "safe" mode
-#hdmi_safe=1
-
-# uncomment this if your display has a black border of unused pixels visible
-# and your display can output without overscan
-#disable_overscan=1
-
-# uncomment the following to adjust overscan. Use positive numbers if console
-# goes off screen, and negative if there is too much border
-#overscan_left=16
-#overscan_right=16
-#overscan_top=16
-#overscan_bottom=16
-
-# uncomment to force a console size. By default it will be display's size minus
-# overscan.
-#framebuffer_width=1280
-#framebuffer_height=720
-
-# uncomment if hdmi display is not detected and composite is being output
-#hdmi_force_hotplug=1
-
-# uncomment to force a specific HDMI mode (this will force VGA)
-#hdmi_group=1
-#hdmi_mode=1
-
-# uncomment to force a HDMI mode rather than DVI. This can make audio work in
-# DMT (computer monitor) modes
-#hdmi_drive=2
-
-# uncomment to increase signal to HDMI, if you have interference, blanking, or
-# no display
-#config_hdmi_boost=4
-
-# uncomment for composite PAL
-#sdtv_mode=2
-
-# GPU memory
-gpu_mem=16
-
-#uncomment to overclock the arm. 700 MHz is the default.
-#arm_freq=800
-
-# Uncomment some or all of these to enable the optional hardware interfaces
-dtparam=i2c_arm=on
-dtparam=i2c_vc=on
-#dtparam=i2s=on
-dtparam=spi=on
-
-# I2C-1 baud rate 100 kHz
-dtparam=i2c1_baudrate=100000
-
-# Enable DS1307 real time clock
-dtoverlay=i2c-rtc,ds1307
-
-# Uncomment this to enable the lirc-rpi module
-#dtoverlay=lirc-rpi
-
-# Additional overlays and parameters are documented /boot/overlays/README
-
-# Enable audio (loads snd_bcm2835)
-dtparam=audio=on
-EOF
-}
-
-policy_rcd_file()
-{
-	cat <<EOF
-#!/bin/sh
-exit 101
-EOF
-}
-
 download()
 {
 	local target="$1"
@@ -290,32 +218,10 @@ build_pythonpack()
 		die "Failed to remove $name build files."
 }
 
-build_rpigpio()
-{
-	local archive="$RPIGPIO_FILE"
-	for python in pypy; do
-		build_pythonpack "$python" "RPi.GPIO-$RPIGPIO_VERSION" \
-				 "$archive" "RPi.GPIO-$RPIGPIO_VERSION" 0
-	done
-	rm "/tmp/$archive" ||\
-		die "Failed to remove /tmp/$archive."
-}
-
-build_spidev()
-{
-	local archive="$SPIDEV_FILE"
-	for python in pypy; do
-		build_pythonpack "$python" "spidev-$SPIDEV_VERSION" \
-				 "$archive" "spidev-$SPIDEV_VERSION" 0
-	done
-	rm "/tmp/$archive" ||\
-		die "Failed to remove /tmp/$archive."
-}
-
 build_ppl()
 {
 	local archive="$PPL_FILE"
-	for python in python python3 pypy; do
+	for python in python3; do
 		build_pythonpack "$python" "ppl-$PPL_VERSION" \
 				 "$archive" "ppl-$PPL_VERSION" 1
 	done
@@ -326,7 +232,7 @@ build_ppl()
 build_ppl2()
 {
 	local archive="$PPL2_FILE"
-	for python in python python3 pypy; do
+	for python in python3; do
 		build_pythonpack "$python" "ppl2-$PPL2_VERSION" \
 				 "$archive" "ppl2-$PPL2_VERSION" 1
 	done
@@ -341,67 +247,84 @@ pilc_bootstrap_first_stage()
 	[ "$(id -u)" = "0" ] || die "Permission denied. Must be root."
 
 	# Check host tools (first/third stage).
+	assert_program 7z
+	assert_program chroot
+	assert_program dd
 	assert_program debootstrap
 	assert_program git
-	assert_program chroot
-	assert_program rsync
-	assert_program parted
-	assert_program mkfs.vfat
+	assert_program gpg
+	assert_program install
 	assert_program mkfs.ext4
-	assert_program 7z
+	assert_program mkfs.vfat
+	assert_program parted
+	assert_program rsync
+	assert_program setarch
 	assert_program tar
 	assert_program unzip
-	assert_program install
+	assert_program wget
 	[ -x "$opt_qemu" ] ||\
 		die "The qemu binary '$opt_qemu' is not executable."
 
+	info "Cleaning tmp..."
+	rm -rf "$opt_target_dir"/tmp/*
+	do_install -d -o root -g root -m 1777 "$opt_target_dir/tmp"
+
+	info "Downloading and extracting keys..."
+	do_install -o root -g root -m 644 \
+		"$basedir_pilc/CF8A1AF502A2AA2D763BAE7E82B129927FA3303E.gpg" \
+		"$opt_target_dir/tmp/"
+	download "$opt_target_dir/tmp/$KEYRING_TGZ_FILE" \
+		 "$KEYRING_BASEURL/$KEYRING_TGZ_FILE" \
+		 "$KEYRING_TGZ_SHA256"
+	tar -C "$opt_target_dir/tmp" -x -f "$opt_target_dir/tmp/$KEYRING_TGZ_FILE" ||\
+		die "Failed to extract keys."
+	local raspbian_asc="$opt_target_dir/tmp/raspbian-archive-keyring-$KEYRING_VERSION/raspbian.public.key"
+	local raspbian_gpg="$raspbian_asc.gpg"
+	gpg --dearmor < "$raspbian_asc" > "$raspbian_gpg" ||\
+		die "Failed to convert key."
+
 	# debootstrap first stage.
 	if [ $opt_skip_debootstrap1 -eq 0 ]; then
 		info "Running debootstrap first stage..."
-		debootstrap --arch="$opt_arch" --foreign --verbose \
-			--keyring="$basedir/pilc/raspbian.public.key.gpg" \
+		setarch linux32 \
+			debootstrap --arch="$opt_arch" --foreign \
+			--components="main,contrib,non-free" \
+			--keyring="$raspbian_gpg" \
 			"$opt_suite" "$opt_target_dir" "$MAIN_MIRROR" \
 			|| die "debootstrap failed"
-		mkdir -p "$opt_target_dir/usr/share/keyrings" ||\
-			die "Failed to create keyrings dir."
-		cp "$basedir/pilc/raspbian.archive.public.key.gpg" \
-		   "$opt_target_dir/usr/share/keyrings/" ||\
-			die "Failed to copy raspbian.archive.public.key.gpg."
-		cp /usr/share/keyrings/debian-archive-keyring.gpg \
-		   "$opt_target_dir/usr/share/keyrings/debian-archive-keyring.gpg" ||\
-			die "Failed to copy debian-archive-keyring.gpg."
 	fi
 	[ -d "$opt_target_dir" ] ||\
 		die "Target directory '$opt_target_dir' does not exist."
 
 	# Avoid the start of daemons during second stage.
-	policy_rcd_file > "$opt_target_dir/usr/sbin/policy-rc.d" ||\
-		die "Failed to create policy-rc.d"
-	chmod 755 "$opt_target_dir/usr/sbin/policy-rc.d" ||\
-		die "Failed to chmod policy-rc.d"
-
-	info "Cleaning tmp..."
-	rm -rf "$opt_target_dir"/tmp/*
+	do_install -o root -g root -m 755 \
+		"$basedir_pilc/templates/policy-rc.d" \
+		"$opt_target_dir/usr/sbin/"
 
 	# Copy qemu.
 	local qemu_bin="$opt_target_dir/$opt_qemu"
 	if ! [ -x "$qemu_bin" ]; then
 		info "Copying qemu binary from '$opt_qemu' to '$qemu_bin'..."
-		mkdir -p "$(dirname "$qemu_bin")" ||\
-			die "Failed to make qemu base directory."
-		cp "$opt_qemu" "$qemu_bin" ||\
-			die "Failed to copy qemu binary."
+		do_install -d -o root -g root -m 755 \
+			"$(dirname "$qemu_bin")"
+		do_install -T -o root -g root -m 755 \
+			"$opt_qemu" "$qemu_bin"
 	fi
 
-	info "Copying PiLC bootstrap script..."
-	cp "$basedir/pilc/pilc-bootstrap.sh" "$opt_target_dir/" ||\
-		die "Failed to copy bootstrap script."
+	info "Copying PiLC bootstrap script and templates..."
+	do_install -o root -g root -m 755 \
+		"$basedir_pilc/pilc-bootstrap.sh" \
+		"$opt_target_dir/"
+	cp -r "$basedir_pilc/templates" "$opt_target_dir/tmp/" ||\
+		die "Failed to copy PiLC templates"
+	cp -r "$basedir_pilc/deb" "$opt_target_dir/tmp/" ||\
+		die "Failed to copy PiLC deb packages"
 
 	info "Checking out awlsim..."
 	local awlsim_dir="$opt_target_dir/tmp/awlsim"
 	local checkout_dir="$awlsim_dir/src"
 	rm -rf "$awlsim_dir"
-	mkdir -p "$awlsim_dir" || die "mkdir failed"
+	do_install -d -o root -g root -m 755 "$awlsim_dir"
 	git clone --no-checkout "$basedir/.git" "$checkout_dir" ||\
 		die "Failed to clone"
 	(
@@ -418,10 +341,12 @@ pilc_bootstrap_first_stage()
 	) || die
 
 	# Fetch packages
-	download "$opt_target_dir/tmp/$RPIGPIO_FILE" "$RPIGPIO_MIRROR" "$RPIGPIO_SHA256"
-	download "$opt_target_dir/tmp/$SPIDEV_FILE" "$SPIDEV_MIRROR" "$SPIDEV_SHA256"
-	download "$opt_target_dir/tmp/$PPL_FILE" "$PPL_MIRROR" "$PPL_SHA256"
-	download "$opt_target_dir/tmp/$PPL2_FILE" "$PPL2_MIRROR" "$PPL2_SHA256"
+	download "$opt_target_dir/tmp/$PPL_FILE" \
+		 "$PPL_MIRROR" \
+		 "$PPL_SHA256"
+	download "$opt_target_dir/tmp/$PPL2_FILE" \
+		 "$PPL2_MIRROR" \
+		 "$PPL2_SHA256"
 
 	# Second stage will mount a few filesystems.
 	# Keep track to umount them in cleanup.
@@ -443,18 +368,28 @@ pilc_bootstrap_second_stage()
 	export LANGUAGE=C
 	export LANG=C
 	if [ "$opt_rpiver" = "1" -o "$opt_rpiver" = "0" ]; then
-		info "Optimizing for RPi 1.x, zero(w) or later"
-		local march="armv6kz"
+		info "Optimizing for RPi 1, zero(w)"
+		local march="-march=armv6kz"
+		local mtune="-mtune=arm1176jzf-s"
 	elif [ "$opt_rpiver" = "2" ]; then
-		info "Optimizing for RPi 2.x or later"
-		local march="armv7-a"
+		info "Optimizing for RPi 2"
+		local march="-march=armv7-a"
+		local mtune="-mtune=cortex-a7"
+	elif [ "$opt_rpiver" = "3" ]; then
+		info "Optimizing for RPi 3"
+		local march="-march=armv8-a"
+		local mtune="-mtune=cortex-a53"
+	elif [ "$opt_rpiver" = "4" ]; then
+		info "Optimizing for RPi 4"
+		local march="-march=armv8-a"
+		local mtune="-mtune=cortex-a72"
 	else
-		info "Optimizing for RPi 3.x or later"
-		local march="armv8-a"
+		info "Optimizing for generic RPi"
+		local march="-march=armv6kz"
+		local mtune=
 	fi
-	export CFLAGS="-O3 -march=$march -mfpu=vfp -mfloat-abi=hard -pipe"
+	export CFLAGS="-O3 $march $mtune -mfpu=vfp -mfloat-abi=hard -pipe"
 	export CXXFLAGS="$CFLAGS"
-#	export CC=clang LINKCC=clang LDSHARED="clang -shared" CXX=clang++
 
 	# debootstrap second stage.
 	if [ $opt_skip_debootstrap2 -eq 0 ]; then
@@ -463,27 +398,24 @@ pilc_bootstrap_second_stage()
 			die "Debootstrap second stage failed."
 	fi
 
+	info "Disabling raspi-copies-and-fills..."
+	rm -f /etc/ld.so.preload || die "Failed to disable raspi-copies-and-fills"
+
 	info "Mounting /proc..."
-	mkdir -p /proc ||\
-		die "Failed to create /proc mountpoint."
-	mount -t proc proc /proc ||\
-		die "Mounting /proc failed."
+	do_install -d -o root -g root -m 755 /proc
+	mount -t proc proc /proc || die "Mounting /proc failed."
 
 	info "Mounting /sys..."
-	mkdir -p /sys ||\
-		die "Failed to create /sys mountpoint."
-	mount -t sysfs sysfs /sys ||\
-		die "Mounting /sys failed."
+	do_install -d -o root -g root -m 755 /sys
+	mount -t sysfs sysfs /sys || die "Mounting /sys failed."
 
 	info "Mounting /dev/shm..."
-	mkdir -p /dev/shm ||\
-		die "Failed to create /dev/shm mountpoint."
-	mount -t tmpfs tmpfs /dev/shm ||\
-		die "Mounting /dev/shm failed."
+	do_install -d -o root -g root -m 755 /dev/shm
+	mount -t tmpfs tmpfs /dev/shm || die "Mounting /dev/shm failed."
 
 	info "Writing apt configuration..."
-cat > /etc/apt/sources.list <<EOF
-deb $MAIN_MIRROR $opt_suite main firmware rpi
+	cat > /etc/apt/sources.list <<EOF
+deb $MAIN_MIRROR $opt_suite main contrib non-free rpi firmware
 deb http://archive.raspberrypi.org/debian/ $opt_suite main ui
 EOF
 	[ $? -eq 0 ] || die "Failed to set sources.list"
@@ -491,20 +423,13 @@ EOF
 		die "Failed to set apt.conf.d"
 
 	info "Creating /etc/fstab"
-	mkdir -p /config ||\
-		die "Failed to create /config"
-	cat > /etc/fstab <<EOF
-proc		/proc			proc		auto,defaults		0 0
-debugfs		/sys/kernel/debug	debugfs		auto,defaults		0 0
-configfs	/config			configfs	auto,defaults		0 0
-tmpfs		/tmp			tmpfs		auto,mode=1777		0 0
-/dev/mmcblk0p2	/			ext4		auto,noatime,errors=remount-ro	0 1
-/dev/mmcblk0p1	/boot			vfat		auto,noatime		0 0
-EOF
+	do_install -d -o root -g root -m 755 /config
+	do_install -T -o root -g root -m 644 \
+		/tmp/templates/fstab \
+		/etc/fstab
 
 	info "Writing misc /etc stuff..."
-	echo "PiLC" > /etc/hostname ||\
-		die "Failed to set hostname"
+	echo "pilc" > /etc/hostname || die "Failed to set hostname"
 	printf 'PiLC GNU/Linux (based on Raspbian) \\n \\l\n\n' > /etc/issue ||\
 		die "Failed to create /etc/issue"
 	printf 'PiLC GNU/Linux (based on Raspbian)\n' > /etc/issue.net ||\
@@ -531,171 +456,88 @@ EOF
 		/etc/os-release ||\
 		die "Failed to set os-release BUG_REPORT_URL."
 
-	info "Updating packages..."
-cat <<EOF | debconf-set-selections
-debconf	debconf/priority	select	high
-debconf	debconf/frontend	select	Noninteractive
-locales	locales/locales_to_be_generated	multiselect	en_US.UTF-8 UTF-8
-locales	locales/default_environment_locale	select	None
-EOF
-	[ $? -eq 0 ] || die "Failed to configure debconf settings"
-	apt-key add /usr/share/keyrings/raspbian.archive.public.key.gpg ||\
+	info "Updating package list..."
+	cat /tmp/templates/debconf-set-selections-preinstall.conf | debconf-set-selections ||\
+		die "Failed to configure debconf settings"
+	apt-key add /tmp/CF8A1AF502A2AA2D763BAE7E82B129927FA3303E.gpg ||\
 		die "apt-key add failed"
-	apt-get -y update ||\
+	local apt_opts="-y -o Acquire::Retries=3"
+	apt-get $apt_opts update ||\
 		die "apt-get update failed"
-	apt-get -y dist-upgrade ||\
+	apt-get $apt_opts install apt-transport-https ||\
+		die "apt-get install apt-transport-https failed"
+	apt-get $apt_opts install \
+		debian-keyring \
+		raspberrypi-archive-keyring \
+		raspbian-archive-keyring \
+		|| die "apt-get install keyrings failed"
+
+	info "Upgrading system..."
+	apt-get $apt_opts dist-upgrade ||\
 		die "apt-get dist-upgrade failed"
 
 	info "Installing packages..."
-	apt-get -y install \
+	apt-get $apt_opts install \
 		aptitude \
-		autoconf \
-		automake \
-		bc \
+		bash \
 		build-essential \
-		bwidget \
-		clang \
 		console-setup \
-		cython \
 		cython3 \
 		dbus \
 		debconf-utils \
+		debsums \
 		devscripts \
+		ethtool \
+		fdisk \
+		firmware-atheros \
 		firmware-brcm80211 \
+		firmware-libertas \
+		firmware-linux \
+		firmware-linux-free \
+		firmware-linux-nonfree \
+		firmware-misc-nonfree \
+		firmware-realtek \
 		git \
-		gnu-fdisk \
 		htop \
 		i2c-tools \
 		irqbalance \
 		iw \
-		libboost-python-dev \
-		libgl1-mesa-dev \
-		libglu1-mesa-dev \
-		libglib2.0-dev \
-		libgtk2.0-dev \
-		libncurses5-dev \
-		libmodbus-dev \
-		libreadline-gplv2-dev \
-		libtk-img \
-		libudev-dev \
-		libusb-1.0-0-dev \
-		libxmu-dev \
 		locales \
 		nano \
 		ntp \
 		openssh-server \
-		openssh-blacklist \
-		openssh-blacklist-extra \
 		parted \
-		pkg-config \
-		pypy \
-		pypy-dev \
-		pypy-setuptools \
-		python \
-		python-all-dev \
-		python-cairo \
-		python-dev \
-		python-gtk2 \
-		python-rpi.gpio \
-		python-serial \
-		python-setuptools \
-		python-smbus \
-		python-spidev \
-		python-tk \
 		python3 \
-		python3-all-dev \
-		python3-cairo \
+		python3-cffi \
 		python3-dev \
-		python3-rpi.gpio \
 		python3-serial \
 		python3-setuptools \
-		python3-smbus \
 		python3-spidev \
-		python3-tk \
-		raspberrypi-bootloader \
-		raspi-config \
+		rng-tools \
 		schedtool \
-		screen \
 		sudo \
 		systemd \
-		tcl-dev \
-		tclx \
-		tk-dev \
 		tmux \
 		vim \
 		wireless-tools \
-		wpasupplicant ||\
-		die "apt-get install failed"
-cat <<EOF | debconf-set-selections
-debconf	debconf/frontend	select	Dialog
-locales	locales/default_environment_locale	select	en_US.UTF-8
-EOF
-	[ $? -eq 0 ] || die "Failed to configure debconf settings"
+		wpasupplicant \
+		|| die "apt-get install failed"
+
+	info "Configuring locales..."
 	dpkg-reconfigure -u locales ||\
 		die "Failed to reconfigure locales"
-	apt-get -y clean ||\
-		die "apt-get clean failed"
 
-	# Build python modules
-	build_rpigpio
-	build_spidev
-	build_ppl
-	build_ppl2
-
-	info "Removing ssh keys..."
-	if [ -e "$(first /etc/ssh/ssh_host_*_key*)" ]; then
-		rm /etc/ssh/ssh_host_*_key* ||\
-			die "Failed to remove ssh keys."
-	fi
-	echo 1 > /etc/ssh/sshd_not_to_be_run ||\
-		die "Failed to create /etc/ssh/sshd_not_to_be_run"
-	echo 1 > /etc/ssh/ssh_create_keys ||\
-		die "Failed to create /etc/ssh/ssh_create_keys"
+	info "Configuring console..."
+	sed -i -e 's|CHARMAP=.*|CHARMAP="UTF-8"|' \
+		-e 's|FONTFACE=.*|FONTFACE=""|' \
+		-e 's|FONTSIZE=.*|FONTSIZE=""|' \
+		/etc/default/console-setup ||\
+		die "Failed to edit /etc/default/console-setup"
 
 	info "Creating /etc/rc.local..."
-	cat > /etc/rc.local <<EOF
-#!/bin/sh -e
-#
-# rc.local
-#
-# This script is executed at the end of each multiuser runlevel.
-# Make sure that the script will "exit 0" on success or any other
-# value on error.
-#
-# In order to enable or disable this script just change the execution
-# bits.
-#
-
-# Re-generate ssh keys, if requested.
-if [ -e /etc/ssh/ssh_create_keys ]; then
-	/bin/rm -f /etc/ssh/ssh_host_*_key*
-	LC_ALL=C LANGUAGE=C LANG=C /usr/sbin/dpkg-reconfigure openssh-server
-	/bin/rm /etc/ssh/sshd_not_to_be_run
-	/bin/rm /etc/ssh/ssh_create_keys
-	/etc/init.d/ssh start
-fi
-
-# Workaround firmware issue leaving i2c0 in an non-ALT0 state.
-for i in 28 29; do
-	/bin/echo \$i > /sys/class/gpio/export
-	/bin/echo in > /sys/class/gpio/gpio\${i}/direction
-done
-
-# Add /dev/ttyS0 link for convenience.
-if ! [ -e /dev/ttyS0 ]; then
-	/bin/ln -s /dev/ttyAMA0 /dev/ttyS0
-fi
-
-exit 0
-EOF
-	[ $? -eq 0 ] || die "Failed to create /etc/rc.local"
-	chmod 755 /etc/rc.local || die "Failed to chmod /etc/rc.local"
-
-	info "Creating /etc/modules-load.d/i2c.conf..."
-	cat > /etc/modules-load.d/i2c.conf <<EOF
-i2c_dev
-EOF
-	[ $? -eq 0 ] || die "Failed to create /etc/modules-load.d/i2c.conf"
+	do_install -o root -g root -m 755 \
+		/tmp/templates/rc.local \
+		/etc/
 
 	info "Creating users/groups..."
 	userdel -f pi
@@ -710,50 +552,106 @@ EOF
 		die "Failed to create user pi."
 	printf 'raspberry\nraspberry\n' | passwd pi ||\
 		die "Failed to set 'pi' password."
-	echo 'pi ALL=(ALL:ALL) ALL' > "/etc/sudoers.d/00-pi" ||\
-		die "Failed to create /etc/sudoers.d/00-pi"
 
 	info "Initializing home directory..."
-	mkdir -p /home/pi/.vim || die "Failed to mkdir /home/pi/.vim"
-	cat > /home/pi/.vim/vimrc <<EOF
-set nocompatible
-set autoindent
-syntax enable
-set backspace=indent,start
-set number
-EOF
-	[ $? -eq 0 ] || die "Failed to create /home/pi/.vim/vimrc"
-	cat > /home/pi/.tmux.conf <<EOF
-# Default new panes and windows to be opened in the current panes path
-bind-key c new-window -c "#{pane_current_path}"
-bind-key % split-window -h -c "#{pane_current_path}"
-bind-key "\"" split-window -c "#{pane_current_path}"
-EOF
-	[ $? -eq 0 ] || die "Failed to create /home/pi/.tmux.conf"
+	do_install -d -o pi -g pi -m 755 /home/pi/.vim
+	do_install -o pi -g pi -m 644 \
+		/tmp/templates/vimrc \
+		/home/pi/.vim/
+	do_install -T -o pi -g pi -m 644 \
+		/tmp/templates/tmux.conf \
+		/home/pi/.tmux.conf
+
+	info "Installing raspbian packages..."
+	apt-get $apt_opts install \
+		libraspberrypi-bin \
+		libraspberrypi-dev \
+		libraspberrypi-doc \
+		python3-rpi.gpio \
+		raspberrypi-bootloader \
+		raspberrypi-kernel \
+		raspberrypi-net-mods \
+		raspberrypi-sys-mods \
+		raspi-config \
+		raspi-gpio \
+		raspinfo \
+		rpi-eeprom \
+		rpi-eeprom-images \
+		|| die "apt-get install failed"
+
+	info "Running debconf-set-selections..."
+	cat /tmp/templates/debconf-set-selections-postinstall.conf | debconf-set-selections ||\
+		die "Failed to configure debconf settings"
+
+	info "Cleaning apt..."
+	apt-get $apt_opts autoremove --purge ||\
+		die "apt-get autoremove failed"
+	apt-get $apt_opts clean ||\
+		die "apt-get clean failed"
+
+	info "Disabling some services..."
+	do_systemctl mask apt-daily.service
+	do_systemctl mask apt-daily.timer
+	do_systemctl mask apt-daily-upgrade.timer
+	do_systemctl mask rsync.service
+	do_systemctl mask exim4.service
+	do_systemctl mask triggerhappy.service
+	do_systemctl mask triggerhappy.socket
+	do_systemctl mask alsa-state.service
+	do_systemctl mask alsa-restore.service
+	do_systemctl mask alsa-utils.service
+
+	info "Building and installing PiLC system package..."
+	(
+		cd /tmp/deb/pilc-system || die "Failed to cd to pilc-system"
+		debuild -uc -us -b -d || die "debuild failed"
+		dpkg -i ../pilc-system_*.deb || die "Failed to install pilc-system"
+		# Copy debs
+		rm -rf /home/pi/deb/pilc-system
+		do_install -d -o pi -g pi -m 755 /home/pi/deb/pilc-system
+		do_install -o pi -g pi -m 644 \
+			../*pilc-system*.deb \
+			/home/pi/deb/pilc-system/
+	) || die
+
+	info "Updating /etc/hosts..."
+	if ! grep -qe pilc /etc/hosts; then
+		printf '\n127.0.0.1\tpilc\n' >> /etc/hosts ||\
+			die "Failed to update /etc/hosts"
+	fi
+
+	info "Building Python modules..."
+	build_ppl
+	build_ppl2
 
 	info "Building awlsim..."
 	(
-		cd /tmp/awlsim/src ||\
-			die "Failed to cd"
+		cd /tmp/awlsim/src || die "Failed to cd"
 		if [ $opt_cython -eq 0 ]; then
 			# Disable cython
 			sed -i -e '/Package: cython/,/^$/ d' \
 				debian/control ||\
-				die "Failed to patch control file"
+				die "Failed to patch control file (cython)"
 			sed -i -e 's/export AWLSIM_CYTHON_BUILD=1/export AWLSIM_CYTHON_BUILD=0/' \
 				debian/rules ||\
-				die "Failed to patch rules file"
+				die "Failed to patch rules file (cython)"
 		fi
+
+		# Update the systemd service file.
+		sed -i -e 's|AWLSIM_SCHED=|AWLSIM_SCHED=realtime-if-multicore|g' \
+		       -e 's|AWLSIM_PRIO=|AWLSIM_PRIO=50|g' \
+		       -e 's|AWLSIM_AFFINITY=|AWLSIM_AFFINITY=-1,-2,-3|g' \
+		       -e 's|AWLSIM_MLOCK=|AWLSIM_MLOCK=1|g' \
+		       awlsim-server.service ||\
+		       die "Failed to patch awlsim-server.service"
+
+		# Build the packages.
 		debuild -uc -us -b -d || die "debuild failed"
-		info "Built awlsim files:"
-		ls .. || die "Failed to list results"
 
 		info "Installing awlsim..."
 		# Core
 		dpkg -i ../python3-awlsim_*.deb ||\
 			die "Failed to install python3-awlsim"
-		dpkg -i ../pypy-awlsim_*.deb ||\
-			die "Failed to install pypy-awlsim"
 		if [ $opt_cython -ne 0 ]; then
 			dpkg -i ../cython3-awlsim_*.deb ||\
 				die "Failed to install cython3-awlsim"
@@ -761,8 +659,6 @@ EOF
 		# hardware: dummy
 		dpkg -i ../python3-awlsimhw-dummy_*.deb ||\
 			die "Failed to install python3-awlsimhw-dummy"
-		dpkg -i ../pypy-awlsimhw-dummy_*.deb ||\
-			die "Failed to install pypy-awlsimhw-dummy"
 		if [ $opt_cython -ne 0 ]; then
 			dpkg -i ../cython3-awlsimhw-dummy_*.deb ||\
 				die "Failed to install cython3-awlsimhw-dummy"
@@ -777,8 +673,6 @@ EOF
 		# hardware: profibus
 		dpkg -i ../python3-awlsimhw-profibus_*.deb ||\
 			die "Failed to install python3-awlsimhw-profibus"
-		dpkg -i ../pypy-awlsimhw-profibus_*.deb ||\
-			die "Failed to install pypy-awlsimhw-profibus"
 		if [ $opt_cython -ne 0 ]; then
 			dpkg -i ../cython3-awlsimhw-profibus_*.deb ||\
 				die "Failed to install cython3-awlsimhw-profibus"
@@ -786,8 +680,6 @@ EOF
 		# hardware: RPi GPIO
 		dpkg -i ../python3-awlsimhw-rpigpio_*.deb ||\
 			die "Failed to install python3-awlsimhw-rpigpio"
-		dpkg -i ../pypy-awlsimhw-rpigpio_*.deb ||\
-			die "Failed to install pypy-awlsimhw-rpigpio"
 		if [ $opt_cython -ne 0 ]; then
 			dpkg -i ../cython3-awlsimhw-rpigpio_*.deb ||\
 				die "Failed to install cython3-awlsimhw-rpigpio"
@@ -795,8 +687,6 @@ EOF
 		# hardware: PiXtend
 		dpkg -i ../python3-awlsimhw-pixtend_*.deb ||\
 			die "Failed to install python3-awlsimhw-pixtend"
-		dpkg -i ../pypy-awlsimhw-pixtend_*.deb ||\
-			die "Failed to install pypy-awlsimhw-pixtend"
 		if [ $opt_cython -ne 0 ]; then
 			dpkg -i ../cython3-awlsimhw-pixtend_*.deb ||\
 				die "Failed to install cython3-awlsimhw-pixtend"
@@ -814,53 +704,44 @@ EOF
 			die "Failed to install awlsim-proupgrade"
 		# Copy debs
 		rm -rf /home/pi/deb/awlsim
-		mkdir -p /home/pi/deb/awlsim ||\
-			die "mkdir /home/pi/deb/awlsim failed"
-		cp ../*awlsim*.deb ../*awlsim*.buildinfo ../*awlsim*.changes \
-			/home/pi/deb/awlsim/ ||\
-			die "Failed to copy awlsim debs"
+		do_install -d -o pi -g pi -m 755 /home/pi/deb/awlsim
+		do_install -o pi -g pi -m 644 \
+			../*awlsim*.deb \
+			/home/pi/deb/awlsim/
 		# Copy examples
-		cp examples/EXAMPLE.awlpro /home/pi/generic-example.awlpro ||\
-			die "Failed to copy EXAMPLE.awlpro."
-		cp examples/raspberrypi-gpio.awlpro /home/pi/raspberrypi-gpio-example.awlpro ||\
-			die "Failed to copy raspberrypi-gpio.awlpro."
-		cp examples/raspberrypi-profibus.awlpro /home/pi/raspberrypi-profibus-example.awlpro ||\
-			die "Failed to copy raspberrypi-profibus.awlpro."
-		cp examples/raspberrypi-pixtend.awlpro /home/pi/raspberrypi-pixtend-example.awlpro ||\
-			die "Failed to copy raspberrypi-pixtend.awlpro."
-
-		#TODO run the testsuite
-
-		#TODO install unit via package
-		info "Installing awlsim service unit..."
-		local awlsim_prefix=/usr
-		local pyver=3
-		local site="$awlsim_prefix/lib/python$pyver/dist-packages"
-		cat awlsim-server.service.in |\
-		sed -e 's|@USER@|root|g' \
-		    -e 's|@GROUP@|root|g' \
-		    -e "s|@PREFIX@|$awlsim_prefix|g" \
-		    -e 's|@PROJECT@|/etc/awlsim-server.awlpro|g' \
-		    -e "s|@PYTHON@|/usr/bin/python$pyver|g" \
-		    -e "s|@PYTHON_SITE@|$site|g" >\
-		    /etc/systemd/system/awlsim-server.service ||\
-		    die "Failed to create awlsim-server.service"
-		systemctl enable awlsim-server.service ||\
-			die "Failed to enable awlsim-server-service"
+		do_install -T -o pi -g pi -m 644 \
+			examples/EXAMPLE.awlpro \
+			/home/pi/generic-example.awlpro
+		do_install -T -o pi -g pi -m 644 \
+			examples/raspberrypi-gpio.awlpro \
+			/home/pi/raspberrypi-gpio-example.awlpro
+		do_install -T -o pi -g pi -m 644 \
+			examples/raspberrypi-profibus.awlpro \
+			/home/pi/raspberrypi-profibus-example.awlpro
+		do_install -T -o pi -g pi -m 644 \
+			examples/raspberrypi-pixtend.awlpro \
+			/home/pi/raspberrypi-pixtend-example.awlpro
 	) || die
 	info "Building pyprofibus..."
 	(
 		cd /tmp/awlsim/pyprofibus ||\
 			die "Failed to cd"
+		if [ $opt_cython -eq 0 ]; then
+			# Disable cython
+			sed -i -e '/Package: cython/,/^$/ d' \
+				debian/control ||\
+				die "Failed to patch control file (cython)"
+			sed -i -e 's/export PYPROFIBUS_CYTHON_BUILD=1/export PYPROFIBUS_CYTHON_BUILD=0/' \
+				debian/rules ||\
+				die "Failed to patch rules file (cython)"
+		fi
+
+		# Build the packages.
 		debuild -uc -us -b -d || die "debuild failed"
-		info "Built pyprofibus files:"
-		ls .. || die "Failed to list results"
 
 		info "Installing pyprofibus..."
 		dpkg -i ../python3-pyprofibus_*.deb ||\
 			die "Failed to install python3-pyprofibus"
-		dpkg -i ../pypy-pyprofibus_*.deb ||\
-			die "Failed to install pypy-pyprofibus"
 		dpkg -i ../profisniff_*.deb ||\
 			die "Failed to install profisniff"
 		dpkg -i ../gsdparser_*.deb ||\
@@ -868,51 +749,16 @@ EOF
 
 		# Copy debs
 		rm -rf /home/pi/deb/pyprofibus
-		mkdir -p /home/pi/deb/pyprofibus ||\
-			die "mkdir /home/pi/deb/pyprofibus failed"
-		cp ../*pyprofibus*.deb ../*pyprofibus*.buildinfo ../*pyprofibus*.changes \
-			../profisniff_*.deb ../gsdparser_*.deb \
-			/home/pi/deb/pyprofibus/ ||\
-			die "Failed to copy pyprofibus debs"
+		do_install -d -o pi -g pi -m 755 /home/pi/deb/pyprofibus
+		do_install -o pi -g pi -m 644 \
+			../*pyprofibus*.deb \
+			../profisniff_*.deb \
+			../gsdparser_*.deb \
+			/home/pi/deb/pyprofibus/
 	) || die
 	rm -r /tmp/awlsim ||\
 		die "Failed to remove awlsim checkout."
 
-	info "Configuring network..."
-	cat > /etc/network/interfaces.d/lo <<EOF
-auto lo
-iface lo inet loopback
-EOF
-	[ $? -eq 0 ] || die "Failed to create /etc/network/interfaces.d/lo"
-	cat > /etc/network/interfaces.d/enx <<EOF
-allow-hotplug /enx*=enx
-iface enx inet dhcp
-iface enx inet6 auto
-EOF
-	[ $? -eq 0 ] || die "Failed to create /etc/network/interfaces.d/enx"
-	cat > /etc/network/interfaces.d/eth <<EOF
-allow-hotplug /eth*=eth
-iface eth inet dhcp
-iface eth inet6 auto
-EOF
-	[ $? -eq 0 ] || die "Failed to create /etc/network/interfaces.d/eth"
-	for i in $(seq 0 9); do
-		cat > /etc/network/interfaces.d/eth$i <<EOF
-allow-hotplug eth$i
-iface eth$i inet dhcp
-iface eth$i inet6 auto
-EOF
-		[ $? -eq 0 ] || die "Failed to create /etc/network/interfaces.d/eth$i"
-	done
-	cat > /etc/network/interfaces.d/wlan0 <<EOF
-#allow-hotplug wlan0
-#iface wlan0 inet dhcp
-#	wpa-ssid 'enter your SSID here'
-#	wpa-psk 'enter your pre-shared-key (PSK) here'
-#iface wlan0 inet6 auto
-EOF
-	[ $? -eq 0 ] || die "Failed to create /etc/network/interfaces.d/wlan0"
-
 	info "Updating home directory permissions..."
 	chown -R pi:pi /home/pi || die "Failed to change /home/pi permissions."
 
@@ -922,8 +768,13 @@ EOF
 			die "Failed to remove policy-rc.d"
 	fi
 
+	# Install this last. It won't work correctly in the qemu environment.
+	info "Installing raspi-copies-and-fills..."
+	apt-get $apt_opts install --reinstall raspi-copies-and-fills ||\
+		die "apt-get install failed"
+
 	info "Stopping processes..."
-	for i in dbus ssh atd irqbalance; do
+	for i in dbus ssh irqbalance; do
 		/etc/init.d/$i stop
 	done
 }
@@ -949,32 +800,12 @@ pilc_bootstrap_third_stage()
 	rm -rf "$opt_target_dir"/tmp/*
 
 	info "Configuring boot..."
-	cat > "$opt_target_dir/boot/cmdline.txt" <<EOF
-dwc_otg.lpm_enable=0 console=tty1 root=/dev/mmcblk0p2 rootfstype=ext4 elevator=deadline isolcpus=2,3 rcu_nocbs=2,3 nohz_full=2,3 fsck.repair=yes net.ifnames=0 rootwait quiet
-EOF
-	[ $? -eq 0 ] || die "Failed to create /boot/cmdline.txt"
-	boot_config_file > "$opt_target_dir/boot/config.txt" ||\
-		die "Failed to create /boot/config.txt"
-	local img="$(first "$opt_target_dir/boot/"vmlinuz-*-rpi)"
-	if [ -e "$img" ]; then
-		mv "$img" "$opt_target_dir/boot/kernel.img" ||\
-			die "Failed to create kernel.img"
-	fi
-	local img="$(first "$opt_target_dir/boot/"initrd.img-*-rpi)"
-	if [ -e "$img" ]; then
-		mv "$img" "$opt_target_dir/boot/initrd.img" ||\
-			die "Failed to create initrd.img"
-	fi
-	local img="$(first "$opt_target_dir/boot/"vmlinuz-*-rpi2)"
-	if [ -e "$img" ]; then
-		mv "$img" "$opt_target_dir/boot/kernel7.img" ||\
-			die "Failed to create kernel7.img"
-	fi
-	local img="$(first "$opt_target_dir/boot/"initrd.img-*-rpi2)"
-	if [ -e "$img" ]; then
-		mv "$img" "$opt_target_dir/boot/initrd7.img" ||\
-			die "Failed to create initrd7.img"
-	fi
+	do_install -T -o root -g root -m 644 \
+		"$basedir_pilc/templates/boot_cmdline.txt" \
+		"$opt_target_dir/boot/cmdline.txt"
+	do_install -T -o root -g root -m 644 \
+		"$basedir_pilc/templates/boot_config.txt" \
+		"$opt_target_dir/boot/config.txt"
 
 	# Prepare image paths.
 	local imgfile="${opt_target_dir}${opt_imgsuffix}.img"
@@ -988,9 +819,17 @@ EOF
 
 	# Create images.
 	if [ "$opt_img" -ne 0 ]; then
+		# Calculate image size.
+		local imgsize_b="$(expr "$opt_imgsize" \* 1000 \* 1000 \* 1000)"
+		local imgsize_mib="$(expr "$imgsize_b" \/ 1024 \/ 1024)"
+		# Reduce the size to make sure it fits every SD card.
+		local imgsize_mib_red="$(expr \( "$imgsize_mib" \* 98 \) \/ 100)"
+		[ -n "$imgsize_mib_red" ] || die "Failed to calculate image size"
+		info "SD image size = $imgsize_mib_red MiB"
+
 		info "Creating boot image..."
 		mkfs.vfat -F 32 -i 7771B0BB -n boot -C "$bootimgfile" \
-			$(expr \( 64 \* 1024 \) - \( 4 \* 1024 \) ) ||\
+			$(expr \( 256 \* 1024 \) ) ||\
 			die "Failed to create boot partition file system."
 		mkdir "$mp_bootimgfile" ||\
 			die "Failed to make boot partition mount point."
@@ -1005,7 +844,7 @@ EOF
 			die "Failed to remove boot partition mount point."
 
 		info "Creating root image..."
-		mkfs.ext4 "$rootimgfile" $(expr \( 4000 - 64 \) \* 1024 ) ||\
+		mkfs.ext4 "$rootimgfile" $(expr \( "$imgsize_mib_red" - \( 256 + 4 + 4 \) \) \* 1024 ) ||\
 			die "Failed to create root filesystem."
 		mkdir "$mp_rootimgfile" ||\
 			die "Failed to make root partition mount point."
@@ -1013,9 +852,10 @@ EOF
 			die "Failed to mount root partition."
 		rsync -aHAX --inplace \
 			--exclude='boot/*' \
-			--exclude='proc/*' \
-			--exclude='sys/*' \
 			--exclude='dev/shm/*' \
+			--exclude='proc/*' \
+			--exclude='run/*' \
+			--exclude='sys/*' \
 			--exclude='tmp/*' \
 			--exclude="$(basename "$opt_qemu")" \
 			"$opt_target_dir/" "$mp_rootimgfile/" ||\
@@ -1026,13 +866,13 @@ EOF
 			die "Failed to remove root partition mount point."
 
 		info "Creating image '$imgfile'..."
-		dd if=/dev/zero of="$imgfile" bs=1M count=4000 conv=sparse ||\
+		dd if=/dev/zero of="$imgfile" bs=1M count="$imgsize_mib_red" conv=sparse ||\
 			die "Failed to create image file."
 		parted "$imgfile" <<EOF
-		    unit b
-		    mklabel msdos
-		    mkpart primary fat32 $(expr 4 \* 1024 \* 1024) $(expr 64 \* 1024 \* 1024 - 1)
-		    mkpart primary ext4 $(expr 64 \* 1024 \* 1024) 100%
+unit b
+mklabel msdos
+mkpart primary fat32 $(expr 4 \* 1024 \* 1024) $(expr \( 256 + 4 \) \* 1024 \* 1024)
+mkpart primary ext4 $(expr \( 256 + 4 + 4 \) \* 1024 \* 1024) 100%
 EOF
 		[ $? -eq 0 ] || die "Failed to create partitions."
 
@@ -1045,7 +885,7 @@ EOF
 
 		info "Integrating root image..."
 		dd if="$rootimgfile" of="$imgfile"\
-			seek=64 bs=1M conv=notrunc,sparse ||\
+			seek="$(expr 256 + 4 + 4)" bs=1M conv=notrunc,sparse ||\
 			die "Failed to integrate root partition."
 		rm "$rootimgfile" ||\
 			die "Failed to delete root partition image."
@@ -1085,9 +925,12 @@ usage()
 	echo " --qemu-bin|-Q PATH      Select qemu-user-static binary."
 	echo "                         Default: $default_qemu"
 	echo
-	echo " --img-suffix|-s SUFFIX  Image file suffix."
+	echo " --img-suffix|-X SUFFIX  Image file suffix."
 	echo "                         Default: $default_imgsuffix"
 	echo
+	echo " --img-size|-S SIZEGB    Image file size, in Gigabytes (base 1000)."
+	echo "                         Default: $default_imgsize"
+	echo
 	echo " --no-img|-I             Do not create an image."
 	echo "                         Default: Create image."
 	echo
@@ -1106,16 +949,20 @@ usage()
 	echo " --quick|-q              Quick build. This is a shortcut for:"
 	echo "                         --no-cython --no-zimg"
 	echo
-	echo " --rpiver|-R VERSION     Minimum Raspberry Pi version to build for."
-	echo "                         Can be either 0, 1, 2 or 3."
+	echo " --rpiver|-R VERSION     Raspberry Pi version to build for."
+	echo "                         Can be either 0, 1, 2, 3, 4 or generic"
 	echo "                         0 and 1 are equivalent."
-	echo "                         Default: 1"
+	echo "                         generic runs on any Raspberry Pi 0-4."
+	echo "                         Default: generic"
 }
 
 # canonicalize basedir
 basedir="$(readlink -e "$basedir")"
 [ -n "$basedir" ] || die "Failed to canonicalize base directory."
 
+# Directory containing the PiLC bootstrapping files.
+basedir_pilc="$basedir/pilc"
+
 # Mountpoints. Will be umounted on cleanup.
 mp_shm=
 mp_proc=
@@ -1137,11 +984,12 @@ if [ -z "$__PILC_BOOTSTRAP_SECOND_STAGE__" ]; then
 	default_arch="armhf"
 	default_qemu="/usr/bin/qemu-arm-static"
 	default_imgsuffix="-$(date '+%Y%m%d')"
+	default_imgsize=4
 	default_img=1
 	default_zimg=1
 	default_writedev=
 	default_writeonly=0
-	default_rpiver=1
+	default_rpiver="generic"
 
 	opt_target_dir=
 	opt_branch="$default_branch"
@@ -1152,6 +1000,7 @@ if [ -z "$__PILC_BOOTSTRAP_SECOND_STAGE__" ]; then
 	opt_skip_debootstrap1=0
 	opt_skip_debootstrap2=0
 	opt_imgsuffix="$default_imgsuffix"
+	opt_imgsize="$default_imgsize"
 	opt_img="$default_img"
 	opt_zimg="$default_zimg"
 	opt_writedev="$default_writedev"
@@ -1193,10 +1042,15 @@ if [ -z "$__PILC_BOOTSTRAP_SECOND_STAGE__" ]; then
 		--skip-debootstrap2|-2)
 			opt_skip_debootstrap2=1
 			;;
-		--img-suffix|-s)
+		--img-suffix|-X)
 			shift
 			opt_imgsuffix="$1"
 			;;
+		--img-size|-S)
+			shift
+			opt_imgsize="$(expr "$1" \* 1)"
+			[ -n "$opt_imgsize" ] || die "--img-size|-S is invalid"
+			;;
 		--no-zimg|-Z)
 			opt_zimg=0
 			;;
@@ -1220,10 +1074,12 @@ if [ -z "$__PILC_BOOTSTRAP_SECOND_STAGE__" ]; then
 		--rpiver|-R)
 			shift
 			opt_rpiver="$1"
-			[ "$opt_rpiver" = "0" -o\
+			[ "$opt_rpiver" = "generic" -o\
+			  "$opt_rpiver" = "0" -o\
 			  "$opt_rpiver" = "1" -o\
 			  "$opt_rpiver" = "2" -o\
-			  "$opt_rpiver" = "3" ] || die "Invalid --rpiver|-R"
+			  "$opt_rpiver" = "3" -o\
+			  "$opt_rpiver" = "4" ] || die "Invalid --rpiver|-R"
 			;;
 		*)
 			opt_target_dir="$*"
@@ -1267,7 +1123,8 @@ if [ -z "$__PILC_BOOTSTRAP_SECOND_STAGE__" ]; then
 	export opt_writeonly
 	export opt_rpiver
 	export __PILC_BOOTSTRAP_SECOND_STAGE__=1
-	chroot "$opt_target_dir" "/pilc-bootstrap.sh" ||\
+	setarch linux32 \
+		chroot "$opt_target_dir" "/pilc-bootstrap.sh" ||\
 		die "Chroot failed."
 
 	# Run third stage.
