@@ -2,7 +2,7 @@
 #
 # PROFIBUS DP - Master
 #
-# Copyright (c) 2013-2016 Michael Buesch <m@bues.ch>
+# Copyright (c) 2013-2019 Michael Buesch <m@bues.ch>
 #
 # Licensed under the terms of the GNU General Public License version 2,
 # or (at your option) any later version.
@@ -40,13 +40,14 @@ class DpSlaveState(object):
 		STATE_DX	: "Data_Exchange",
 	}
 
-	defaultTimeLimits = {
+	# State timeouts in seconds
+	stateTimeLimits = {
 		STATE_INIT	: TimeLimit.UNLIMITED,
 		STATE_WDIAG	: 1.0,
 		STATE_WPRM	: 0.5,
 		STATE_WCFG	: 0.5,
 		STATE_WDXRDY	: 1.0,
-		STATE_DX	: 0.3,
+		STATE_DX	: 0.5,
 	}
 
 	def __init__(self, master, slaveDesc):
@@ -71,25 +72,36 @@ class DpSlaveState(object):
 		self.pendingReqTimeout = TimeLimit()
 		self.shortAckReceived = False
 
+		# Data_Exchange context
+		self.dxStartTime = 0.0
+
 		# Received telegrams
 		self.rxQueue = []
 
+		# In/Out user data
+		self.outData = None
+		self.inData = None
+
 	def getRxQueue(self):
 		rxQueue = self.rxQueue
 		self.rxQueue = []
 		return rxQueue
 
+	def flushRxQueue(self):
+		self.rxQueue = []
+
 	def getState(self):
 		return self.__state
 
 	def getNextState(self):
 		return self.__nextState
 
-	def setState(self, state, stateTimeout = TimeLimit.DEFAULT):
-		if stateTimeout == TimeLimit.DEFAULT:
-			stateTimeout = self.defaultTimeLimits[state]
+	def setState(self, state, stateTimeLimit=None):
+		if stateTimeLimit is None:
+			stateTimeLimit = self.stateTimeLimits[state]
 		self.__nextState = state
-		self.__stateTimeout.start(stateTimeout)
+		self.__stateTimeout.start(stateTimeLimit)
+		self.master.phy.clearTxQueueAddr(self.slaveDesc.slaveAddr)
 		self.master._releaseSlave(self)
 
 	def applyState(self):
@@ -97,8 +109,7 @@ class DpSlaveState(object):
 		self.__prevState, self.__state = self.__state, self.__nextState
 
 		# Handle state switch
-		if self.stateJustEntered() or\
-		   not self.pendingReq:
+		if self.stateJustEntered():
 			self.pendingReq = None
 
 	def stateJustEntered(self):
@@ -121,10 +132,12 @@ class DpSlaveDesc(object):
 	"""
 
 	def __init__(self,
-		     identNumber,
+		     gsd,
 		     slaveAddr):
-		self.identNumber = identNumber
+		self.gsd = gsd
 		self.slaveAddr = slaveAddr
+		self.identNumber = gsd.getIdentNumber() if gsd else 0
+		self.dpm = None
 
 		# Prepare a Set_Prm telegram.
 		self.setPrmTelegram = DpTelegram_SetPrm_Req(
@@ -141,15 +154,17 @@ class DpSlaveDesc(object):
 		"""Sets DpCfgDataElement()s from the specified list
 		in the Chk_Cfg telegram.
 		"""
-		self.chkCfgTelegram.clearCfgDataElements()
-		for cfgDataElement in cfgDataElements:
-			self.chkCfgTelegram.addCfgDataElement(cfgDataElement)
+		if cfgDataElements is not None:
+			self.chkCfgTelegram.clearCfgDataElements()
+			for cfgDataElement in cfgDataElements:
+				self.chkCfgTelegram.addCfgDataElement(cfgDataElement)
 
 	def setUserPrmData(self, userPrmData):
 		"""Sets the User_Prm_Data of the Set_Prm telegram.
 		"""
-		self.setPrmTelegram.clearUserPrmData()
-		self.setPrmTelegram.addUserPrmData(userPrmData)
+		if userPrmData is not None:
+			self.setPrmTelegram.clearUserPrmData()
+			self.setPrmTelegram.addUserPrmData(userPrmData)
 
 	def setSyncMode(self, enabled):
 		"""Enable/disable sync-mode.
@@ -199,6 +214,12 @@ class DpSlaveDesc(object):
 		self.setPrmTelegram.wdFact1 = fact1
 		self.setPrmTelegram.wdFact2 = fact2
 
+	def setOutData(self, outData):
+		self.dpm.setSlaveOutData(self, outData)
+
+	def getInData(self):
+		return self.dpm.getSlaveInData(self)
+
 	def __repr__(self):
 		return "DPSlaveDesc(identNumber=%s, slaveAddr=%d)" %\
 			(intToHex(self.identNumber), self.slaveAddr)
@@ -210,11 +231,12 @@ class DpMaster(object):
 		self.masterAddr = masterAddr
 		self.debug = debug
 
-		self.__activeSlave = None
+		# Create the transceivers
+		self.fdlTrans = FdlTransceiver(self.phy)
+		self.dpTrans = DpTransceiver(self.fdlTrans, thisIsMaster=True)
 
-		mcastSlaveDesc = DpSlaveDesc(
-			identNumber = 0,
-			slaveAddr = FdlTelegram.ADDRESS_MCAST)
+		mcastSlaveDesc = DpSlaveDesc(gsd=None,
+					     slaveAddr=FdlTelegram.ADDRESS_MCAST)
 		mcastSlave = DpSlaveState(self, mcastSlaveDesc)
 
 		self.__slaveDescs = {
@@ -223,14 +245,16 @@ class DpMaster(object):
 		self.__slaveStates = {
 			FdlTelegram.ADDRESS_MCAST : mcastSlave,
 		}
-
-		# Create the transceivers
-		self.fdlTrans = FdlTransceiver(self.phy)
-		self.dpTrans = DpTransceiver(self.fdlTrans, thisIsMaster = True)
+		self.__slaveDescsList = []
+		self.__runNextSlaveIndex = 0
 
 		# Do we have the token?
 		self.__haveToken = True
 
+		self.__slowDown = False
+		self.__slowDownUntil = monotonic_time()
+		self.__slowDownFact = 1
+
 	def __debugMsg(self, msg):
 		if self.debug:
 			print("DPM%d: %s" % (self.dpmClass, msg))
@@ -238,6 +262,15 @@ class DpMaster(object):
 	def __errorMsg(self, msg):
 		print("DPM%d:  >ERROR<  %s" % (self.dpmClass, msg))
 
+	def __masterSlowDown(self):
+		"""A severe communication error occurred.
+		Slow down the state machine a bit.
+		"""
+		self.__slowDown = True
+		self.__slowDownUntil = monotonic_time() + (0.01 * self.__slowDownFact)
+		self.__debugMsg("Slow down factor = %d" % self.__slowDownFact)
+		self.__slowDownFact = min(self.__slowDownFact + 1, 10)
+
 	def destroy(self):
 		if self.phy:
 			self.phy.close()
@@ -250,217 +283,228 @@ class DpMaster(object):
 		if slaveAddr in self.__slaveDescs or\
 		   slaveAddr in self.__slaveStates:
 			raise DpError("Slave %d is already registered." % slaveAddr)
+		slaveDesc.dpm = self
 		self.__slaveDescs[slaveAddr] = slaveDesc
 		self.__slaveStates[slaveAddr] = DpSlaveState(self, slaveDesc)
 
+		# Rebuild the slave desc list.
+		self.__slaveDescsList = [
+			desc
+			for addr, desc in sorted(self.__slaveDescs.items(),
+						 key=lambda x: x[0])
+			if addr != FdlTelegram.ADDRESS_MCAST
+		]
+
+		self.__runNextSlaveIndex = 0
+
 	def getSlaveList(self):
-		"""Get a list of registered DpSlaveDescs, sorted by address."""
+		"""Get a list of registered DpSlaveDescs, sorted by address.
+		"""
+		return self.__slaveDescsList
 
-		return [ desc for addr, desc in sorted(self.__slaveDescs.items(),
-						       key = lambda x: x[0])
-			 if addr != FdlTelegram.ADDRESS_MCAST ]
-
-	def __send(self, slave, telegram, timeout = 0.1):
+	def __send(self, slave, telegram, timeout):
 		"""Asynchronously send a telegram to a slave.
 		"""
 		slave.pendingReq = telegram
-		slave.pendingReqTimeout.start(timeout)
 		slave.shortAckReceived = False
 		try:
 			if FdlTelegram.checkType(telegram):
 				transceiver = self.fdlTrans
 			else:
 				transceiver = self.dpTrans
-			transceiver.send(fcb = slave.fcb,
-					 telegram = telegram)
+			transceiver.send(fcb=slave.fcb,
+					 telegram=telegram)
 		except ProfibusError as e:
 			slave.pendingReq = None
-			raise e
-		self.__activeSlave = slave
+			self.__masterSlowDown()
+			self.__debugMsg(str(e))
+			return False
+		self.__slowDownFact = 1
+		slave.pendingReqTimeout.start(timeout)
+		return True
 
 	def _releaseSlave(self, slave):
-		if slave is self.__activeSlave:
-			self.__activeSlave = None
-			self.phy.releaseBus()
+		self.phy.releaseBus()
 
-	def __runSlave_init(self, slave, dataExOutData):
-		if (not slave.pendingReq or\
-		    slave.pendingReqTimeout.exceed()):
+	def __runSlave_init(self, slave):
+		if slave.stateJustEntered():
 			self.__debugMsg("Trying to initialize slave %d..." % (
 				slave.slaveDesc.slaveAddr))
+			slave.flushRxQueue()
+		else:
+			for telegram in slave.getRxQueue():
+				if telegram.fc is not None:
+					slave.pendingReq = None
+					stype = telegram.fc & FdlTelegram.FC_STYPE_MASK
+					if telegram.fc & FdlTelegram.FC_REQ:
+						self.__debugMsg("Slave %d replied with "
+								"request bit set." %\
+								slave.slaveDesc.slaveAddr)
+					elif stype != FdlTelegram.FC_SLAVE:
+						self.__debugMsg("Device %d is not a slave. "
+								"Detected type: 0x%02X" % (
+								slave.slaveDesc.slaveAddr,
+								stype))
+					else:
+						slave.setState(slave.STATE_WDIAG)
+						return None
+				else:
+					self.__debugMsg("Slave %d replied with a "
+						"weird telegram:\n%s" % str(telegram))
 
+		if (not slave.pendingReq or
+		    slave.pendingReqTimeout.exceed()):
 			# Reset fault debounce counter.
 			slave.faultDeb.reset()
 
 			# Disable the FCB bit.
 			slave.fcb.enableFCB(False)
 
-			try:
-				self.__send(slave,
-					telegram = FdlTelegram_FdlStat_Req(
-						da = slave.slaveDesc.slaveAddr,
-						sa = self.masterAddr),
-					timeout = 0.2)
-			except ProfibusError as e:
-				self.__debugMsg("FdlStat_Req failed: %s" % str(e))
+			ok = self.__send(slave,
+					 telegram=FdlTelegram_FdlStat_Req(
+						da=slave.slaveDesc.slaveAddr,
+						sa=self.masterAddr),
+					 timeout=0.01)
+			if not ok:
+				self.__debugMsg("FdlStat_Req failed")
 				return None
-
-		for telegram in slave.getRxQueue():
-			if telegram.fc is not None:
-				slave.pendingReq = None
-				stype = telegram.fc & FdlTelegram.FC_STYPE_MASK
-				if telegram.fc & FdlTelegram.FC_REQ:
-					self.__debugMsg("Slave %d replied with "
-							"request bit set." %\
-							slave.slaveDesc.slaveAddr)
-				elif stype != FdlTelegram.FC_SLAVE:
-					self.__debugMsg("Device %d is not a slave. "
-							"Detected type: 0x%02X" % (
-							slave.slaveDesc.slaveAddr,
-							stype))
-				else:
-					slave.setState(slave.STATE_WDIAG)
-			else:
-				self.__debugMsg("Slave %d replied with a "
-					"weird telegram:\n%s" % str(telegram))
 		return None
 
-	def __runSlave_waitDiag(self, slave, dataExOutData):
-		if not slave.pendingReq:
+	def __runSlave_waitDiag(self, slave):
+		if slave.stateJustEntered():
 			self.__debugMsg("Requesting Slave_Diag from slave %d..." %\
 				slave.slaveDesc.slaveAddr)
+			slave.flushRxQueue()
+		else:
+			for telegram in slave.getRxQueue():
+				if DpTelegram_SlaveDiag_Con.checkType(telegram):
+					slave.setState(slave.STATE_WPRM)
+					return None
+				else:
+					self.__debugMsg("Received spurious "
+						"telegram:\n%s" % str(telegram))
 
+		if (not slave.pendingReq or
+		    slave.pendingReqTimeout.exceed()):
 			# Enable the FCB bit.
 			slave.fcb.enableFCB(True)
 
 			# Send a SlaveDiag request
-			try:
-				self.__send(slave,
-					telegram = DpTelegram_SlaveDiag_Req(
-						da = slave.slaveDesc.slaveAddr,
-						sa = self.masterAddr))
-			except ProfibusError as e:
-				self.__debugMsg("SlaveDiag_Req failed: %s" % str(e))
+			ok = self.__send(slave,
+					 telegram=DpTelegram_SlaveDiag_Req(
+						da=slave.slaveDesc.slaveAddr,
+						sa=self.masterAddr),
+					 timeout=0.05)
+			if not ok:
+				self.__debugMsg("SlaveDiag_Req failed")
 				return None
-
-
-		for telegram in slave.getRxQueue():
-			if DpTelegram_SlaveDiag_Con.checkType(telegram):
-				slave.setState(slave.STATE_WPRM)
-				break
-			else:
-				self.__debugMsg("Received spurious "
-					"telegram:\n%s" % str(telegram))
-		else:
-			if slave.pendingReqTimeout.exceed():
-				slave.setState(slave.STATE_INIT)
-
 		return None
 
-	def __runSlave_waitPrm(self, slave, dataExOutData):
-		if not slave.pendingReq:
+	def __runSlave_waitPrm(self, slave):
+		if slave.stateJustEntered():
 			self.__debugMsg("Sending Set_Prm to slave %d..." %\
 				slave.slaveDesc.slaveAddr)
-
-			# Send a Set_Prm request
-			try:
-				slave.slaveDesc.setPrmTelegram.sa = self.masterAddr
-				self.__send(slave,
-					telegram = slave.slaveDesc.setPrmTelegram)
-			except ProfibusError as e:
-				self.__debugMsg("Set_Prm failed: %s" % str(e))
+			slave.flushRxQueue()
+		else:
+			if slave.shortAckReceived:
+				slave.fcb.handleReply()
+				slave.setState(slave.STATE_WCFG)
 				return None
 
-		if slave.shortAckReceived:
-			slave.fcb.handleReply()
-			slave.setState(slave.STATE_WCFG)
-		elif slave.pendingReqTimeout.exceed():
-			slave.setState(slave.STATE_INIT)
-
+		if (not slave.pendingReq or
+		    slave.pendingReqTimeout.exceed()):
+			# Send a Set_Prm request
+			slave.slaveDesc.setPrmTelegram.sa = self.masterAddr
+			ok = self.__send(slave,
+					 telegram=slave.slaveDesc.setPrmTelegram,
+					 timeout=0.05)
+			if not ok:
+				self.__debugMsg("Set_Prm failed")
+				return None
 		return None
 
-	def __runSlave_waitCfg(self, slave, dataExOutData):
-		if not slave.pendingReq:
+	def __runSlave_waitCfg(self, slave):
+		if slave.stateJustEntered():
 			self.__debugMsg("Sending Chk_Cfg to slave %d..." %\
 				slave.slaveDesc.slaveAddr)
-
-			try:
-				slave.slaveDesc.chkCfgTelegram.sa = self.masterAddr
-				self.__send(slave,
-					telegram = slave.slaveDesc.chkCfgTelegram,
-					timeout = 0.3)
-			except ProfibusError as e:
-				self.__debugMsg("Chk_Cfg failed: %s" % str(e))
-				return None
-
-		if slave.shortAckReceived:
-			slave.fcb.handleReply()
-			slave.setState(slave.STATE_WDXRDY)
-		elif slave.pendingReqTimeout.exceed():
-			slave.setState(slave.STATE_INIT)
-
-		return None
-
-	def __runSlave_waitDxRdy(self, slave, dataExOutData):
-		if not slave.pendingReq:
-			self.__debugMsg("Requesting Slave_Diag from slave %d..." %\
-				slave.slaveDesc.slaveAddr)
-
-			try:
-				self.__send(slave,
-					telegram = DpTelegram_SlaveDiag_Req(
-						da = slave.slaveDesc.slaveAddr,
-						sa = self.masterAddr))
-			except ProfibusError as e:
-				self.__debugMsg("SlaveDiag_Req failed: %s" % str(e))
-				return None
-
-		for telegram in slave.getRxQueue():
-			if DpTelegram_SlaveDiag_Con.checkType(telegram):
-				if telegram.notExist():
-					self.__errorMsg("Slave %d is not reachable "
-						"via this line." %\
-						slave.slaveDesc.slaveAddr)
-				if telegram.cfgFault():
-					self.__errorMsg("Slave %d reports a faulty "
-						"configuration (Chk_Cfg)." %\
-						slave.slaveDesc.slaveAddr)
-				if telegram.prmFault():
-					self.__errorMsg("Slave %d reports a faulty "
-						"parameterization (Set_Prm)." %\
-						slave.slaveDesc.slaveAddr)
-				if telegram.isNotSupp():
-					self.__errorMsg("Slave %d replied with "
-						"\"function not supported\". "
-						"The parameters should be checked "
-						"(Set_Prm)." %\
-						slave.slaveDesc.slaveAddr)
-				if telegram.masterLock():
-					self.__errorMsg("Slave %d is already controlled "
-						"(locked to) another DP-master." %\
-						slave.slaveDesc.slaveAddr)
-				if not telegram.hasOnebit():
-					self.__debugMsg("Slave %d diagnostic "
-						"always-one-bit is zero." %\
-						slave.slaveDesc.slaveAddr)
-
-				if telegram.hasExtDiag():
-					pass#TODO turn on red DIAG-LED
-				if telegram.isReadyDataEx():
-					slave.setState(slave.STATE_DX)
-				elif telegram.needsNewPrmCfg():
-					slave.setState(slave.STATE_INIT)
-				break
-			else:
-				self.__debugMsg("Received spurious "
-					"telegram:\n%s" % str(telegram))
+			slave.flushRxQueue()
 		else:
-			if slave.pendingReqTimeout.exceed():
-				slave.setState(slave.STATE_INIT)
+			if slave.shortAckReceived:
+				slave.fcb.handleReply()
+				slave.setState(slave.STATE_WDXRDY)
 
+		if (not slave.pendingReq or
+		    slave.pendingReqTimeout.exceed()):
+			slave.slaveDesc.chkCfgTelegram.sa = self.masterAddr
+			ok = self.__send(slave,
+					 telegram=slave.slaveDesc.chkCfgTelegram,
+					 timeout=0.05)
+			if not ok:
+				self.__debugMsg("Chk_Cfg failed")
+				return None
 		return None
 
-	def __runSlave_dataExchange(self, slave, dataExOutData):
+	def __runSlave_waitDxRdy(self, slave):
+		if slave.stateJustEntered():
+			self.__debugMsg("Requesting Slave_Diag (WDXRDY) from slave %d..." %\
+				slave.slaveDesc.slaveAddr)
+			slave.flushRxQueue()
+		else:
+			for telegram in slave.getRxQueue():
+				if DpTelegram_SlaveDiag_Con.checkType(telegram):
+					if telegram.notExist():
+						self.__errorMsg("Slave %d is not reachable "
+							"via this line." %\
+							slave.slaveDesc.slaveAddr)
+					if telegram.cfgFault():
+						self.__errorMsg("Slave %d reports a faulty "
+							"configuration (Chk_Cfg)." %\
+							slave.slaveDesc.slaveAddr)
+					if telegram.prmFault():
+						self.__errorMsg("Slave %d reports a faulty "
+							"parameterization (Set_Prm)." %\
+							slave.slaveDesc.slaveAddr)
+					if telegram.isNotSupp():
+						self.__errorMsg("Slave %d replied with "
+							"\"function not supported\". "
+							"The parameters should be checked "
+							"(Set_Prm)." %\
+							slave.slaveDesc.slaveAddr)
+					if telegram.masterLock():
+						self.__errorMsg("Slave %d is already controlled "
+							"(locked to) another DP-master." %\
+							slave.slaveDesc.slaveAddr)
+					if not telegram.hasOnebit():
+						self.__debugMsg("Slave %d diagnostic "
+							"always-one-bit is zero." %\
+							slave.slaveDesc.slaveAddr)
+
+					if telegram.hasExtDiag():
+						pass#TODO turn on red DIAG-LED
+					if telegram.isReadyDataEx():
+						slave.setState(slave.STATE_DX)
+						return None
+					elif telegram.needsNewPrmCfg():
+						slave.setState(slave.STATE_INIT)
+						return None
+					break
+				else:
+					self.__debugMsg("Received spurious "
+						"telegram:\n%s" % str(telegram))
+
+		if (not slave.pendingReq or
+		    slave.pendingReqTimeout.exceed()):
+			ok = self.__send(slave,
+					 telegram=DpTelegram_SlaveDiag_Req(
+						da=slave.slaveDesc.slaveAddr,
+						sa=self.masterAddr),
+					 timeout=0.05)
+			if not ok:
+				self.__debugMsg("SlaveDiag_Req failed")
+				return None
+		return None
+
+	def __runSlave_dataExchange(self, slave):
 		#TODO: add support for in/out-only slaves
 		dataExInData = None
 
@@ -468,10 +512,8 @@ class DpMaster(object):
 			self.__debugMsg("Initialization finished. "
 				"Running Data_Exchange with slave %d..." %\
 				slave.slaveDesc.slaveAddr)
-
-		if slave.pendingReqTimeout.exceed():
-			slave.faultDeb.fault()
-			slave.pendingReq = None
+			slave.flushRxQueue()
+			slave.dxStartTime = monotonic_time()
 
 		if slave.pendingReq:
 			for telegram in slave.getRxQueue():
@@ -486,37 +528,50 @@ class DpMaster(object):
 					       FdlTelegram.FC_RDH}:
 					self.__debugMsg("Slave %d requested diagnostics." %\
 						slave.slaveDesc.slaveAddr)
-					slave.setState(slave.STATE_WDXRDY)
+					slave.setState(slave.STATE_WDXRDY, 0.2)
 				elif resFunc == FdlTelegram.FC_RS:
 					raise DpError("Service not active "
 						"on slave %d" % slave.slaveDesc.slaveAddr)
 				dataExInData = telegram.getDU()
-			if dataExInData is not None:
+
+			if dataExInData is None:
+				if slave.pendingReqTimeout.exceed():
+					self.__debugMsg("Data_Exchange timeout with slave %d" % (
+							slave.slaveDesc.slaveAddr))
+					slave.faultDeb.fault()
+					slave.pendingReq = None
+			else:
+				# We received some data.
 				slave.pendingReq = None
-				slave.faultDeb.faultless()
+				slave.faultDeb.ok()
 				slave.restartStateTimeout()
 				self._releaseSlave(slave)
 		else:
-			try:
-				self.__send(slave,
-					telegram = DpTelegram_DataExchange_Req(
-						da = slave.slaveDesc.slaveAddr,
-						sa = self.masterAddr,
-						du = dataExOutData))
-			except ProfibusError as e:
-				self.__debugMsg("DataExchange_Req failed: %s" % str(e))
-				return None
+			# Send the out data telegram, if any.
+			outData = slave.outData
+			if outData is not None:
+				ok = self.__send(slave,
+						 telegram=DpTelegram_DataExchange_Req(
+							da=slave.slaveDesc.slaveAddr,
+							sa=self.masterAddr,
+							du=outData),
+						 timeout=0.1)
+				if not ok:
+					self.__debugMsg("DataExchange_Req failed")
+					return None
+				# We sent it. Reset the data.
+				slave.outData = None
 
 		faultCount = slave.faultDeb.get()
 		if faultCount >= 5:
 			# communication lost
 			self.__debugMsg("Communication lost in Data_Exchange.")
 			slave.setState(slave.STATE_INIT)
-		elif faultCount >= 3:
+		elif faultCount >= 3 and monotonic_time() >= slave.dxStartTime + 0.2:
 			# Diagnose the slave
 			self.__debugMsg("Many errors in Data_Exchange. "
 				"Requesting diagnostic information...")
-			slave.setState(slave.STATE_WDXRDY)
+			slave.setState(slave.STATE_WDXRDY, 0.2)
 
 		return dataExInData
 
@@ -529,12 +584,11 @@ class DpMaster(object):
 		DpSlaveState.STATE_DX		: __runSlave_dataExchange,
 	}
 
-	def runSlave(self, slaveDesc, dataExOutData = None):
+	def __runSlave(self, slave):
 		self.__pollRx()
 		if not self.__haveToken:
 			return None
 
-		slave = self.__slaveStates[slaveDesc.slaveAddr]
 		if slave.stateHasTimeout():
 			self.__debugMsg("State machine timeout! "
 				"Trying to re-initializing slave %d..." %\
@@ -543,7 +597,7 @@ class DpMaster(object):
 			dataExInData = None
 		else:
 			handler = self.__slaveStateHandlers[slave.getState()]
-			dataExInData = handler(self, slave, dataExOutData)
+			dataExInData = handler(self, slave)
 
 			if slave.stateIsChanging():
 				self.__debugMsg("slave[%02X].state --> '%s'" % (
@@ -569,14 +623,14 @@ class DpMaster(object):
 			elif telegram.da == FdlTelegram.ADDRESS_MCAST:
 				self.__handleMcastTelegram(telegram)
 			elif telegram.da == self.masterAddr:
-				try:
+				if telegram.sa in self.__slaveStates:
 					slave = self.__slaveStates[telegram.sa]
-				except KeyError:
+					slave.rxQueue.append(telegram)
+					slave.fcb.handleReply()
+				else:
 					self.__debugMsg("Received telegram from "
 						"unknown station %d:\n%s" %(
 						telegram.sa, str(telegram)))
-				slave.rxQueue.append(telegram)
-				slave.fcb.handleReply()
 			else:
 				self.__debugMsg("Received telegram for "
 					"foreign station:\n%s" % str(telegram))
@@ -589,6 +643,46 @@ class DpMaster(object):
 		self.__debugMsg("Received multicast telegram:\n%s" % str(telegram))
 		pass#TODO
 
+	def run(self):
+		"""Run the DP-Master state machine.
+		"""
+		if self.__slowDown:
+			# Master slowdown is active.
+			# Do not run state machine until the end of the slowdown.
+			if monotonic_time() < self.__slowDownUntil:
+				return None
+			self.__slowDown = False
+
+		slaveDescsList = self.__slaveDescsList
+		runNextSlaveIndex = self.__runNextSlaveIndex
+
+		if not slaveDescsList:
+			return None
+
+		slaveDesc = slaveDescsList[runNextSlaveIndex]
+		self.__runNextSlaveIndex = (runNextSlaveIndex + 1) % len(slaveDescsList)
+
+		slave = self.__slaveStates[slaveDesc.slaveAddr]
+		slave.inData = self.__runSlave(slave)
+
+		return slaveDesc
+
+	def setSlaveOutData(self, slaveDesc, outData):
+		"""Set the out-data that will be sent the
+		next time we are able to send something to that slave.
+		"""
+		slave = self.__slaveStates[slaveDesc.slaveAddr]
+		slave.outData = outData
+
+	def getSlaveInData(self, slaveDesc):
+		"""Get the latest received in-data.
+		Returns None, if there was no received data.
+		"""
+		slave = self.__slaveStates[slaveDesc.slaveAddr]
+		inData = slave.inData
+		slave.inData = None
+		return inData
+
 	def initialize(self):
 		"""Initialize the DPM."""
 
