@@ -33,35 +33,33 @@ errormsg()
 
 die()
 {
-	errormsg "$*"
-
-	# We might be in a sub-job. So write to fail-file.
-	failfile_write "$*"
-
+	if [ -n "$*" ]; then
+		errormsg "$*"
+		# We might be in a sub-job. So write to fail-file.
+		failfile_write "$*"
+	fi
 	exit 1
 }
 
-# Create a temporary file. $1=name
+# Create a temporary file. $1=name, $2=subdir
 maketemp()
 {
 	local prefix="$1"
+	local subdir="$2"
 
-	mktemp --tmpdir="$tmp_dir" awlsim-test-${prefix}.XXXXXX
+	if [ -z "$subdir" ]; then
+		local subdir="."
+	else
+		mkdir -p "$tmp_dir/$subdir"
+	fi
+	mktemp --tmpdir="$tmp_dir" "${subdir}/awlsim-test-${prefix}.XXXXXX"
 }
 
 # $1=message
 test_failed()
 {
 	errormsg "=== TEST FAILED ==="
-
-	if [ $opt_softfail -eq 0 ]; then
-		die "$@"
-	else
-		failfile_write "$*"
-		errormsg "$*"
-		errormsg "^^^ TEST FAILED ^^^"
-		[ $global_retval -eq 0 ] && global_retval=1
-	fi
+	die "$@"
 }
 
 cleanup()
@@ -153,13 +151,28 @@ check_job_failure()
 	[ "0" != "$(du -s "$test_fail_file" | cut -f1)" ]
 }
 
+wait_for_all_background_jobs()
+{
+	is_parallel_run || return
+
+	infomsg "Waiting for background jobs..."
+	wait
+	# Print the fail information.
+	if check_job_failure; then
+		errormsg
+		errormsg "===== FAILURES in parallel run: ====="
+		cat "$test_fail_file" >&2
+		errormsg "====================================="
+		global_retval=1
+	fi
+}
+
 # $1=interpreter
 # Returns version on stdout as:  MAJOR MINOR PATCHLEVEL
 get_interpreter_version()
 {
 	local interpreter="$1"
 
-	[ "$interpreter" = "cython" -o "$interpreter" = "cython2" ] && local interpreter=python2
 	[ "$interpreter" = "cython3" ] && local interpreter=python3
 
 	"$interpreter" -c 'import sys; print("%d %d %d" % sys.version_info[0:3]);' 2>/dev/null
@@ -172,11 +185,7 @@ interpreter_is_gui_compat()
 	local interpreter="$1"
 
 	[ $opt_nogui -eq 0 ] &&\
-	[ "$interpreter" = "python" -o \
-	  "$interpreter" = "python2" -o \
-	  "$interpreter" = "python3" -o \
-	  "$interpreter" = "cython" -o \
-	  "$interpreter" = "cython2" -o \
+	[ "$interpreter" = "python3" -o \
 	  "$interpreter" = "cython3" ]
 }
 
@@ -218,34 +227,20 @@ check_dos_text_encoding()
 	fi
 }
 
-# $1=interpreter [$2=tested_file]
+# $1=interpreter $2=tested_file [$3=test_name]
 setup_test_environment()
 {
 	local interpreter="$1"
 	local tested_file="$2"
+	local test_name="$3"
 
+	[ -z "$test_name" ] && local test_name="$tested_file"
+	local test_name="$(realpath -m --no-symlinks --relative-base="$rootdir" "$test_name" | tr '/\\' _)"
+
+	# Check if we want to run on Cython3 and set up the environment.
 	local use_cython=0
-
-	# Check if we want to run on Cython2/3 and set the environment
-	# for Cython2/3.
-	if [ "$interpreter" = "cython" -o "$interpreter" = "cython2" ] ||\
-	   [ "$interpreter" = "python" -a "$AWLSIM_CYTHON" != "" ] ||\
-	   [ "$interpreter" = "python2" -a "$AWLSIM_CYTHON" != "" ]; then
-		# We want to run the test using Cython2
-
-		local use_cython=2
-
-		for i in "$rootdir"/build/lib.linux-*-2.*; do
-			export PYTHONPATH="$i"
-			break
-		done
-		# Enforce cython module usage
-		export AWLSIM_CYTHON=2
-		# The actual interpreter is Python
-		local interpreter=python2
-
-	elif [ "$interpreter" = "cython3" ] ||\
-	     [ "$interpreter" = "python3" -a "$AWLSIM_CYTHON" != "" ]; then
+	if [ "$interpreter" = "cython3" ] ||\
+	   [ "$interpreter" = "python3" -a "$AWLSIM_CYTHON" != "" ]; then
 		# We want to run the test using Cython3
 
 		local use_cython=3
@@ -259,8 +254,13 @@ setup_test_environment()
 		# The actual interpreter is Python
 		local interpreter=python3
 
+	elif [ "$interpreter" = "micropython" ]; then
+		# We want to run the test using Micropython
+
+		local interpreter="$rootdir/maintenance/micropython-wrapper.sh"
+
 	else
-		# Neither Cython2 nor Cython3
+		# Not Cython
 		export PYTHONPATH=
 		export AWLSIM_CYTHON=
 	fi
@@ -313,6 +313,16 @@ setup_test_environment()
 	# Disable CPU affinity
 	unset AWLSIM_AFFINITY
 
+	# Setup coverage tracing
+	if [ $coverage_enabled -eq 0 ]; then
+		unset AWLSIM_COVERAGE
+	else
+		local coverage_data_file="$(maketemp "coverage_${test_name}" "$coverage_data_subdir")"
+		rm "$coverage_data_file"
+
+		export AWLSIM_COVERAGE="$coverage_data_file"
+	fi
+
 	RET="$interpreter"
 }
 
@@ -340,73 +350,88 @@ run_awl_test()
 	local awl="$2"
 	shift; shift
 
-	setup_test_environment "$interpreter" "$awl"
-	local actual_interpreter="$RET"
-
 	# By default run once with all optimizers enabled.
 	local optimizer_runs="$(get_conf "$awl" optimizer_runs all)"
 
-	local first=1
+	local first_opti=1
 	for optimizers in $optimizer_runs; do
-		[ $first -eq 0 ] && infomsg -n " / "
-		local first=0
-
-		local test_time_file="$(maketemp time)"
+		[ $first_opti -eq 0 ] && infomsg -n " / "
+		local first_opti=0
 
 		local tries="$(get_conf "$awl" tries 1)"
 		[ $tries -lt 1 ] && local tries=1
+		local first_try=1
 
 		local ok=0
 		local exit_code=-1
 		local expected_exit_code=-2
 		while [ $tries -gt 0 -a $ok -eq 0 ]; do
-			local ok=1
 			local tries="$(expr "$tries" - 1)"
-			local loglevel="$(get_conf "$awl" loglevel "$opt_loglevel")"
-			local expected_exit_code="$(get_conf "$awl" exit_code 0)"
-			[ $expected_exit_code -eq 0 ] || local loglevel=0
-			local cycle_limit="$(get_conf "$awl" cycle_limit 60)"
-			local max_runtime="$(get_conf "$awl" max_runtime -1)"
-			local dump_opt=
-			[ $loglevel -ge 3 ] && local dump_opt="--no-cpu-dump"
 
-			command time -o "$test_time_file" -f '%E' --quiet \
-			"$actual_interpreter" "$rootdir/awlsim-test" \
-				--loglevel $loglevel \
-				--extended-insns \
-				--hardware debug:inputAddressBase=7:outputAddressBase=8:dummyParam=True \
-				--cycle-limit "$cycle_limit" \
-				--max-runtime "$max_runtime" \
-				--optimizers "$optimizers" \
-				$dump_opt \
-				"$@" \
-				"$awl"
-			local exit_code=$?
-			if [ $exit_code -ne $expected_exit_code ]; then
-				local ok=0
-				if [ $tries -gt 0 ]; then
-					infomsg "Test '$(basename "$awl")' FAILED, but retrying ($tries)..."
-					sleep 1
+			(
+				[ $first_try -ne 0 ] && adjust_niceness "$($SHELL -c 'echo $PPID')"
+				setup_test_environment "$interpreter" "$awl"
+				local actual_interpreter="$RET"
+
+				local loglevel="$(get_conf "$awl" loglevel "$opt_loglevel")"
+				local expected_exit_code="$(get_conf "$awl" exit_code 0)"
+				[ $expected_exit_code -eq 0 ] || local loglevel=0
+				local cycle_limit="$(get_conf "$awl" cycle_limit 60)"
+				local max_runtime="$(get_conf "$awl" max_runtime -1)"
+				local accus="$(get_conf "$awl" accus)"
+				if [ "$accus" = "2" ]; then
+					local accus=--twoaccu
+				elif [ "$accus" = "4" ]; then
+					local accus=--fouraccu
+				elif [ -n "$accus" ]; then
+					cleanup_test_environment
+					die "Invalid 'accus' value in .conf"
 				fi
-			fi
+				local dump_opt=
+				[ $loglevel -ge 3 ] && local dump_opt="--no-cpu-dump"
+
+				"$actual_interpreter" "$rootdir/awlsim-test" \
+					--loglevel $loglevel \
+					--extended-insns \
+					--hardware debug:inputAddressBase=7:outputAddressBase=8:dummyParam=True \
+					--cycle-limit "$cycle_limit" \
+					--max-runtime "$max_runtime" \
+					--optimizers "$optimizers" \
+					$accus \
+					$dump_opt \
+					"$@" \
+					"$awl"
+				local exit_code=$?
+				if [ $exit_code -ne $expected_exit_code ]; then
+					# Test failed
+					cleanup_test_environment
+					if [ $tries -gt 0 ]; then
+						infomsg "Test '$(basename "$awl")' FAILED, but retrying ($tries tries left)..."
+						sleep 1
+						die # Next try
+					else
+						test_failed "\nTest '$(basename "$awl")'   FAILED" \
+							"\nInterpreter        = $interpreter" \
+							"\nOptimizers         = $optimizers" \
+							"\nActual exit code   = $exit_code" \
+							"\nExpected exit code = $expected_exit_code"
+					fi
+				fi
+
+				cleanup_test_environment
+			) && local ok=1
+			local first_try=0
 		done
 		if [ $ok -eq 0 ]; then
-			test_failed "\nTest '$(basename "$awl")'   FAILED" \
-				"\nInterpreter        = $interpreter" \
-				"\nOptimizers         = $optimizers" \
-				"\nActual exit code   = $exit_code" \
-				"\nExpected exit code = $expected_exit_code"
+			die # Test failed
 		fi
 		if is_parallel_run; then
-			[ $ok -ne 0 ] && infomsg "$(basename "$awl"): O=$optimizers t=$(cat "$test_time_file") -> OK"
+			infomsg "$(basename "$awl"): O=$optimizers -> OK"
 		else
-			[ $ok -ne 0 ] && infomsg -n "O=$optimizers t=$(cat "$test_time_file") -> OK"
+			infomsg -n "O=$optimizers -> OK"
 		fi
-		rm "$test_time_file"
 	done
 	is_parallel_run || infomsg
-
-	cleanup_test_environment
 }
 
 # $1=interpreter $2=sh_file
@@ -420,18 +445,19 @@ run_sh_test()
 
 	[ "$(echo "$sh_file" | cut -c1)" = '/' ] || local sh_file="$(pwd)/$sh_file"
 
-	# Source the test file
-	. "$basedir/sh-test.defaults"
-	. "$sh_file"
-
 	# Run the test
 	(
-	 setup_test_environment "$interpreter" "$sh_file"
-	 local interpreter="$RET"
-	 local test_dir="$(dirname "$sh_file")"
-	 local test_name="$(basename "$sh_file" .sh)"
-	 sh_test "$interpreter" "$test_dir" "$test_name"
-	 cleanup_test_environment
+		# Source the test file
+		. "$basedir/sh-test.defaults"
+		. "$sh_file"
+
+		adjust_niceness "$($SHELL -c 'echo $PPID')"
+		setup_test_environment "$interpreter" "$sh_file"
+		local interpreter="$RET"
+		local test_dir="$(dirname "$sh_file")"
+		local test_name="$(basename "$sh_file" .sh)"
+		sh_test "$interpreter" "$test_dir" "$test_name"
+		cleanup_test_environment
 	)
 	local result=$?
 
@@ -444,62 +470,70 @@ run_sh_test()
 }
 
 # $1=interpreter $2=test_file
-run_nose_test()
+run_pyunit_test()
 {
 	local interpreter="$1"
 	local test_case="$2"
 	shift; shift
 
 	[ -z "$test_case" ] &&\
-		die "Nose test case is missing"
+		die "Python unittest test case is missing"
 	[ -d "$test_case" ] &&\
-		die "Nose test case '$test_case' must not be a directory"
+		die "Python unittest test case '$test_case' must not be a directory"
 	[ -x "$test_case" ] &&\
-		die "Nose test case '$test_case' must NOT be executable"
-
-	# Select nosetests or nosetests3
-	local interp_ver="$(get_interpreter_version "$interpreter")"
-	local interp_major="$(echo "$interp_ver" | cut -d' ' -f 1)"
-	if [ "$interp_major" = "2" ]; then
-		local nose="$nosetests"
-	else
-		local nose="$nosetests3"
-	fi
+		die "Python unittest test case '$test_case' must NOT be executable"
 
 	# Resolve relative path
 	[ "$(echo "$test_case" | cut -c1)" = '/' ] ||\
 		local test_case="$(pwd)/$test_case"
 
-	# Add nose libraries to pypy environment
-	if [ "$(basename "$interpreter")" = "pypy" ]; then
-		local p='import sys; print(":".join(p for p in sys.path if p.startswith("/usr/")))'
-		EXTRA_PYTHONPATH="$("$interpreter" -c "$p"):$(python2 -c "$p"):$EXTRA_PYTHONPATH"
-	fi
-	# Add awlsim_tstlib.py to PYTHONPATH
-	EXTRA_PYTHONPATH="$rootdir/tests:$EXTRA_PYTHONPATH"
+	(
+		# Add awlsim_tstlib.py to PYTHONPATH
+		EXTRA_PYTHONPATH="$rootdir:$rootdir/tests:$EXTRA_PYTHONPATH"
 
-	# Setup python environment
-	setup_test_environment "$interpreter" "$test_case"
-	local interpreter="$RET"
+		# Setup python environment
+		adjust_niceness "$($SHELL -c 'echo $PPID')"
+		local orig_interpreter="$interpreter"
+		setup_test_environment "$interpreter" "$test_case"
+		local interpreter="$RET"
 
-	# Run the nose test case
-	cd "$rootdir" || die "Failed to cd to rootdir."
-	local opts="--no-byte-compile --verbosity=2"
-	if [ -n "$AWLSIM_TEST_QUIET" ]; then
-		"$interpreter" "$nose" $opts "$test_case" >/dev/null 2>&1 ||\
-			die "Nose test case '$(basename "$test_case")' failed."
-	else
-		"$interpreter" "$nose" $opts "$test_case" ||\
-			die "Nose test case '$(basename "$test_case")' failed."
-	fi
+		export PYTHONDONTWRITEBYTECODE=1
 
-	infomsg "$(basename "$test_case"): OK"
+		if [ "$orig_interpreter" = "cython3" ] && ! [ -e "$(dirname "$test_case")/no_cython" ]; then
+			# Get the relative test case path starting in 'tests' directory.
+			local relpath="$(realpath -m --no-symlinks --relative-base="$rootdir/tests" "$test_case")"
+			# Patch the module name to Cython name (append _cython).
+			local patch_re='s/(tc[0-9][0-9][0-9]_[0-9a-zA-Z]*)/\1_cython/'
+			local relpath_cython="$(printf "%s" "$relpath" | sed -Ee "$patch_re")"
+			# Get the relative directory of the test case.
+			local reldir_cython="$(dirname "$relpath_cython")"
+			# Go to the unittest subdir to run the Cython unittest.
+			cd "$rootdir/tests/build/"lib.*-3.*"/$reldir_cython" || die "Failed to cd to test directory."
+		else
+			# Go to the unittest subdir to run the Python unittest.
+			cd "$(dirname "$test_case")" || die "Failed to cd to test directory."
+		fi
 
-	cleanup_test_environment
+		# Convert test name to module name (for python2)
+		local test_case="$(basename "$test_case" .py)"
+
+		# Run it.
+		if [ -n "$AWLSIM_TEST_QUIET" ]; then
+			"$interpreter" -m unittest "$test_case" >/dev/null 2>&1 ||\
+				die "Python unittest test case '$(basename "$test_case")' failed."
+		else
+			"$interpreter" -m unittest "$test_case" ||\
+				die "Python unittest test case '$(basename "$test_case")' failed."
+		fi
+
+		infomsg "$(basename "$test_case"): OK"
+
+		cleanup_test_environment
+	) || die "'$(basename "$test_case")' FAILED"
 }
 
 # $1=interpreter $2=testfile(.awl/.sh) ($3ff additional options to awlsim-test or testfile)
-__run_test()
+run_test()
 {
 	local interpreter="$1"
 	local testfile="$2"
@@ -507,8 +541,10 @@ __run_test()
 
 	# Don't run ourself
 	[ "$(basename "$testfile")" = "run.sh" ] && return
-	# Don't run the test helper library
+	# Don't run artifacts that aren't actual test cases.
 	[ "$(basename "$testfile")" = "awlsim_tstlib.py" ] && return
+	[ "$(basename "$testfile")" = "setup-cython-tests.py" ] && return
+	[ "$(basename "$testfile")" = "__init__.py" ] && return
 
 	local disabled="$(get_conf "$testfile" disabled)"
 	if [ -z "$disabled" ]; then
@@ -530,7 +566,7 @@ __run_test()
 		elif [ "$(echo -n "$testfile" | tail -c3)" = ".sh" ]; then
 			run_sh_test "$interpreter" "$testfile" "$@"
 		elif [ "$(echo -n "$testfile" | tail -c3)" = ".py" ]; then
-			run_nose_test "$interpreter" "$testfile" "$@"
+			run_pyunit_test "$interpreter" "$testfile" "$@"
 		else
 			die "Test file type of '$testfile' not recognized"
 		fi
@@ -541,15 +577,15 @@ __run_test()
 	fi
 }
 
-run_test()
+run_test_parallel()
 {
 	if is_parallel_run; then
 		# Run tests in parallel.
 		wait_for_free_job_slot
-		__run_test "$@" &
+		run_test "$@" &
 	else
 		# Run tests one-by-one.
-		__run_test "$@"
+		run_test "$@"
 	fi
 }
 
@@ -559,6 +595,8 @@ run_test_directory()
 	local interpreter="$1"
 	local directory="$2"
 
+	[ "$(basename "$directory")" = "build" ] && return
+
 	local prettydir="$(realpath -m --no-symlinks --relative-base="$rootdir" "$directory")/"
 
 	infomsg ">>> entering $prettydir"
@@ -568,7 +606,7 @@ run_test_directory()
 		[ "$(echo -n "$entry" | tail -c7)" = ".awlpro" ] || continue
 		[ -e "$(dirname "$entry")/$(basename "$entry" .awlpro).sh" ] && continue
 
-		run_test "$interpreter" "$entry"
+		run_test_parallel "$interpreter" "$entry"
 		check_job_failure && return
 	done
 	# run .awl tests
@@ -578,21 +616,22 @@ run_test_directory()
 		[ -e "$(dirname "$entry")/$(basename "$entry" .awl).awlpro" ] && continue
 		[ -e "$(dirname "$entry")/$(basename "$entry" .awl).sh" ] && continue
 
-		run_test "$interpreter" "$entry"
+		run_test_parallel "$interpreter" "$entry"
 		check_job_failure && return
 	done
 	# run .sh tests
 	for entry in "$directory"/*; do
 		[ -d "$entry" ] && continue
 		[ "$(echo -n "$entry" | tail -c3)" = ".sh" ] || continue
-		run_test "$interpreter" "$entry"
+		run_test_parallel "$interpreter" "$entry"
 		check_job_failure && return
 	done
-	# run nose tests
+	# run .py unittest tests
 	for entry in "$directory"/*; do
 		[ -d "$entry" ] && continue
 		[ "$(echo -n "$entry" | tail -c3)" = ".py" ] || continue
-		run_test "$interpreter" "$entry"
+		[ "$entry" = "__init__.py" ] && continue
+		run_test_parallel "$interpreter" "$entry"
 		check_job_failure && return
 	done
 	# Recurse into subdirectories
@@ -621,20 +660,28 @@ __build_cython()
 		warnmsg "=== WARNING: Cannot build $cython modules"
 		return 1
 	}
-	cd "$rootdir" || die "cd to $rootdir failed"
-	infomsg "=== Building awlsim $cython modules with $python"
-	CFLAGS="-O0" CPPFLAGS= CXXFLAGS="-O0" LDFLAGS= \
-		AWLSIM_CYTHON_BUILD=1 \
-		AWLSIM_CYTHON_PARALLEL=1 \
-		nice -n 5 \
-		"$python" ./setup.py build >/dev/null ||\
-		die "'$python ./setup.py build' failed"
-	return 0
-}
 
-build_cython2()
-{
-	__build_cython cython python2
+	(
+		infomsg "=== Building awlsim $cython modules with $python"
+		cd "$rootdir" || die "cd to $rootdir failed"
+		CFLAGS="-O0" CPPFLAGS= CXXFLAGS="-O0" LDFLAGS= \
+			AWLSIM_CYTHON_BUILD=1 \
+			AWLSIM_CYTHON_PARALLEL=1 \
+			nice -n 5 \
+			"$python" ./setup.py build >/dev/null ||\
+			die "'$python ./setup.py build' failed"
+	) || die
+
+	(
+		infomsg "=== Building awlsim $cython test cases with $python"
+		cd "$rootdir/tests" || die "cd to $rootdir/tests failed"
+		rm -rf build || die "Failed to clean test cases build"
+		nice -n 5 \
+			"$python" ./setup-cython-tests.py build >/dev/null ||\
+			die "'$python ./setup-cython-tests.py build' failed"
+	) || die
+
+	return 0
 }
 
 build_cython3()
@@ -645,14 +692,10 @@ build_cython3()
 # $@=testfiles
 do_tests()
 {
-	have_prog time ||\
-		die "ERROR: 'time' executable not available."\
-		    "Please install the 'time' package."
-
 	cleanup_test_environment
 
 	if [ $opt_quick -eq 0 ]; then
-		local all_interp="python3 python2 pypy3 pypy cython3"
+		local all_interp="python3 python2 cython3 pypy3"
 		if [ $opt_extended -ne 0 ]; then
 			local all_interp="$all_interp jython"
 		fi
@@ -663,53 +706,36 @@ do_tests()
 		fi
 	fi
 
-	find_executable nosetests
-	local nosetests="$RET"
-	find_executable nosetests3
-	local nosetests3="$RET"
-
-	local cython2_build_pid=
-	local cython3_build_pid=
-	if is_parallel_run; then
-		# Trigger the build jobs, if required.
-		local inter="$opt_interpreter $all_interp"
-		if printf '%s' "$inter" | grep -Eqwe 'cython|cython2'; then
-			wait_for_free_job_slot
-			build_cython2 &
-			local cython2_build_pid=$!
-		fi
-		if printf '%s' "$inter" | grep -qwe 'cython3'; then
-			wait_for_free_job_slot
-			build_cython3 &
-			local cython3_build_pid=$!
-		fi
-	fi
-
 	for interpreter in "$opt_interpreter" $all_interp; do
 		[ -z "$interpreter" ] && continue
 
 		cleanup_test_environment
 
-		if [ "$interpreter" = "cython" -o "$interpreter" = "cython2" ]; then
-			have_prog cython && have_prog python2 || {
-				warn_skipped "$interpreter"
-				[ -n "$opt_interpreter" ] && break || continue
-			}
-			if is_parallel_run; then
-				wait_for_job_pid $cython2_build_pid
-			else
-				build_cython2 || die "Cython2 build failed."
-			fi
-		elif [ "$interpreter" = "cython3" ]; then
+		# Create an interpreter name suitable as path component
+		local interpreter_name="$(printf '%s' "$interpreter" | tr '/\\' _)"
+
+		# Check if we should enable coverage tracing
+		coverage_enabled=$opt_coverage
+		if [ $coverage_enabled -ne 0 ] &&\
+		   [ "$interpreter" = "pypy" -o "$interpreter" = "pypy3" ]; then
+			# Performance impact of coverage on PyPy is too big.
+			# Disable coverage to avoid test failures.
+			warnmsg "Disabling code coverage tracing (-c|--coverage) on PyPy due to bad performace."
+			coverage_enabled=0
+		fi
+
+		# Prepare code coverage directory
+		coverage_data_subdir="coverage-$interpreter_name"
+		mkdir -p "$tmp_dir/$coverage_data_subdir" || die "Failed to create coverage data dir"
+
+		# Basic interpreter setup. Build Cython modules.
+		if [ "$interpreter" = "cython3" ]; then
 			have_prog cython3 && have_prog python3 || {
 				warn_skipped "$interpreter"
 				[ -n "$opt_interpreter" ] && break || continue
 			}
-			if is_parallel_run; then
-				wait_for_job_pid $cython3_build_pid
-			else
-				build_cython3 || die "Cython3 build failed."
-			fi
+			wait_for_all_background_jobs
+			build_cython3 || die "Cython3 build failed."
 		else
 			have_prog "$interpreter" || {
 				warn_skipped "$interpreter"
@@ -736,7 +762,7 @@ do_tests()
 				if [ -d "$opt" ]; then
 					run_test_directory "$interpreter" "$opt"
 				else
-					run_test "$interpreter" "$opt"
+					run_test_parallel "$interpreter" "$opt"
 				fi
 				check_job_failure && break
 			done
@@ -744,22 +770,29 @@ do_tests()
 		infomsg
 
 		check_job_failure && break
+
+		# Generate code coverage report
+		if [ $coverage_enabled -ne 0 ]; then
+			# Wait for background jobs to finish
+			wait_for_all_background_jobs
+
+			if [ $global_retval -eq 0 ]; then
+				infomsg "\nGenerating code coverage report..."
+				local reportbase="$rootdir/code-coverage-report"
+				local reportdir="$reportbase/awlsim-coverage-$interpreter_name"
+				rm -rf "$reportdir"
+				"$rootdir/awlsim-covreport" \
+					"$reportdir" \
+					"$tmp_dir/$coverage_data_subdir/" ||\
+					die "Failed to generate code coverage report."
+			fi
+		fi
+
 		[ -n "$opt_interpreter" ] && break
 	done
 
-	if is_parallel_run; then
-		# This is a parallel run. Wait for all jobs.
-		infomsg "Waiting for background jobs..."
-		wait
-		# Print the fail information.
-		if check_job_failure; then
-			errormsg
-			errormsg "===== FAILURES in parallel run: ====="
-			cat "$test_fail_file" >&2
-			errormsg "====================================="
-			global_retval=1
-		fi
-	fi
+	# Wait for background jobs to finish
+	wait_for_all_background_jobs
 
 	# Print summary
 	if [ $global_retval -eq 0 ]; then
@@ -792,12 +825,10 @@ show_help()
 	infomsg
 	infomsg "Options:"
 	infomsg " -i|--interpreter INTER        Use INTER as interpreter for the tests"
-	infomsg " -s|--softfail                 Do not abort on single test failures"
 	infomsg " -j|--jobs NR                  Set the number of jobs to run in parallel."
 	infomsg "                               0 means number-of-CPUs"
-	infomsg "                               Default: 1"
+	infomsg "                               Default: 0"
 	infomsg " -q|--quick                    Only run python2 and python3 tests"
-	infomsg " -qq                           Shortcut for: -q -j 0"
 	infomsg " -g|--no-gui                   Avoid tests that need GUI libraries"
 	infomsg " -x|--extended                 Run tests on additional interpreters"
 	infomsg " -n|--renice NICENESS          Renice by NICENESS. Defaults to 10."
@@ -806,6 +837,7 @@ show_help()
 	infomsg " -l|--loop COUNT               Number of test loops to execute."
 	infomsg "                               Default: 1"
 	infomsg "                               Set to 0 for infinite looping."
+	infomsg " -c|--coverage                 Enable code coverage tracing."
 }
 
 tmp_dir="/tmp/awlsim-test-$$"
@@ -822,14 +854,14 @@ touch "${port_alloc_file}.lock"
 echo 4096 > "$port_alloc_file" || die "Failed to initialize port file"
 
 opt_interpreter=
-opt_softfail=0
 opt_quick=0
 opt_nogui=0
 opt_extended=0
 opt_renice=
-opt_jobs=1
+opt_jobs=0
 opt_loglevel=2
 opt_loop=1
+opt_coverage=0
 
 while [ $# -ge 1 ]; do
 	[ "$(printf '%s' "$1" | cut -c1)" != "-" ] && break
@@ -845,9 +877,6 @@ while [ $# -ge 1 ]; do
 		have_prog "$opt_interpreter" ||\
 			die "Interpreter '${opt_interpreter}' not found"
 		;;
-	-s|--softfail)
-		opt_softfail=1
-		;;
 	-j|--jobs)
 		shift
 		opt_jobs="$1"
@@ -855,10 +884,6 @@ while [ $# -ge 1 ]; do
 	-q|--quick)
 		opt_quick=1
 		;;
-	-qq)
-		opt_quick=1
-		opt_jobs=0
-		;;
 	-g|--no-gui)
 		opt_nogui=1
 		;;
@@ -880,6 +905,9 @@ while [ $# -ge 1 ]; do
 		shift
 		opt_loop="$1"
 		;;
+	-c|--coverage)
+		opt_coverage=1
+		;;
 	*)
 		errormsg "Unknown option: $1"
 		exit 1
@@ -901,17 +929,26 @@ if [ -z "$opt_loop" -o -n "$(printf '%s' "$opt_loop" | tr -d '[0-9]')" ] || [ $o
 	opt_loop=infinite
 fi
 
+
 do_renice()
 {
-	renice "$1" "$$" >/dev/null
+	local niceness="$1"
+	local pid="$2"
+
+	renice "$niceness" "$pid" >/dev/null
 }
 
-if [ -n "$opt_renice" ]; then
-	do_renice "$opt_renice" || die "Failed to renice"
-else
-	# Try to renice. Ignore failure.
-	do_renice 10
-fi
+adjust_niceness()
+{
+	local pid="$1"
+
+	if [ -n "$opt_renice" ]; then
+		do_renice "$opt_renice" "$pid" || die "Failed to renice"
+	else
+		# Try to renice. Ignore failure.
+		do_renice 10 "$pid"
+	fi
+}
 
 
 # Run the tests
